import { TransactWriteCommandInput } from '@aws-sdk/lib-dynamodb';
import { ddb } from '@svc/config';
import Client from '@svc/lib/models/Client';
import { ClientContactMethod, ClientInterface, LowLevelClient, LowLevelClientContactMethod, LowLevelClientProfile } from '@improovy/types';
import _ from 'lodash';

const TableName = ddb.clientsTable;

export const buildLowLevelContactMethods = (client: Client): LowLevelClientContactMethod[] => {
  const { uuid } = client;
  const lowLevelContactMethods: LowLevelClientContactMethod[] = [];
  
  for (let contactMethod of client.contactMethods) {
    const { contactType, contactAddress, isDeleted } = contactMethod;
    const compositeKey = `Contact#${contactType}#${contactAddress}`;

    const lowLevelContactMethod: LowLevelClientContactMethod = {
      pk: uuid,
      sk: compositeKey,
      gsi1_pk: compositeKey,
      gsi1_sk: uuid,
      type: 'ClientContactMethod',
      contactAddress,
      contactType,
      isDeleted,
      clientUuid: uuid,
    };

    if (client.pdid) {
      lowLevelContactMethod.gsi2_pk = `PDID#${client.pdid}`;
      lowLevelContactMethod.gsi2_sk = compositeKey;
    }
    lowLevelContactMethods.push(lowLevelContactMethod);
  }

  return lowLevelContactMethods;
};

export const buildLowLevelClientProfile = (client: Client): LowLevelClientProfile => {
  const { uuid, isTest, changeSource, incomingContactMethod, isDeleted, createdAt, gclid, pdid, pipedriveName, lastUpdatedField } = client;

  const lowLevelProfile: LowLevelClientProfile = {
    pk: uuid,
    sk: 'Profile',
    type: 'ClientProfile',
    incomingContactMethod,
    isDeleted,
    uuid,
    createdAt,
    pipedriveName,
    changeSource,
  };
  
  if (lastUpdatedField) lowLevelProfile.lastUpdatedField = lastUpdatedField;
  if (gclid) lowLevelProfile.gclid = gclid;
  if (pdid) {
    lowLevelProfile.pdid = pdid;
    lowLevelProfile.gsi2_pk = `PDID#${pdid}`;
    lowLevelProfile.gsi2_sk = 'Profile';
  }
  if (isTest) lowLevelProfile.isTest = isTest;
  
  return lowLevelProfile;
};

export const buildLowLevelClient = (client: Client): LowLevelClient => {
  const lowLevelProfile = buildLowLevelClientProfile(client);
  const lowLevelContactMethods = buildLowLevelContactMethods(client);

  
  return {
    profile: lowLevelProfile,
    contactMethods: lowLevelContactMethods,
  };
};

export const assembleClientContactMethod = (contactMethodItem: LowLevelClientContactMethod): ClientContactMethod => {
  // @ts-ignore
  return _.pick(contactMethodItem, ['contactType', 'contactAddress', 'isDeleted']);
};

export const assembleClientProfile = (profile: LowLevelClientProfile, contactMethods: ClientContactMethod[]): ClientInterface => {
  const assembledProfile = { ..._.pick(profile, ['gclid', 'incomingContactMethod', 'isTest', 'isDeleted', 'uuid', 'createdAt', 'pdid', 'lastUpdatedField', 'isDeleted', 'pipedriveName', 'changeSource']), contactMethods };

  // @ts-ignore
  return assembledProfile;
};

export const buildClientPutTransaction = (client: LowLevelClient): TransactWriteCommandInput => {
  const contactMethodWrites = client.contactMethods.map(method => {
    return {
      Put: {
        TableName,
        Item: method,
      },
    };
  });
  
  const transaction: TransactWriteCommandInput = {
    TransactItems: [
      { Put: {
        TableName,
        Item: client.profile,
      },
      },
      ...contactMethodWrites,
    ],
  };

  return transaction;
};

export const assembleClientInterface = (lowLevelClient: LowLevelClient): ClientInterface => {
  const contactMethods = lowLevelClient.contactMethods.map(assembleClientContactMethod);

  return assembleClientProfile(lowLevelClient.profile, contactMethods);
};

export const assembleLowLevelClientInterface = (ddbItems: { [key: string]: any }[]): LowLevelClient => {
  const contactMethods: LowLevelClientContactMethod[] = [];
  let profile: LowLevelClientProfile;

  for (let item of ddbItems) {
    switch (item.type) {
      case 'ClientProfile':
        // @ts-expect-error
        profile = item;
        break;
      case 'ClientContactMethod':
        // @ts-expect-error
        contactMethods.push(item);
        break;
      default:
        throw new Error('Unkown item type in assembleClientInterface');
        
        break;
    }
  }

  // @ts-expect-error
  if (!profile) {
    console.log(JSON.stringify(ddbItems));
    throw new Error('item type ClientProfile not found');
  }

  return {
    profile,
    contactMethods,
  };
};

export const assembleClientInterfaceFromDDBItems = (ddbItems: { [key: string]: any }[]): ClientInterface => {
  const lowLevelClient = assembleLowLevelClientInterface(ddbItems);
  return assembleClientInterface(lowLevelClient);
};