import { ddb } from '@svc/config';
import { TransactWriteCommand } from '@aws-sdk/lib-dynamodb';
import Client from '../models/Client';
import { assembleClientInterfaceFromDDBItems, buildClientPutTransaction, buildLowLevelClient } from './clients-repo-utils';
import { queryByPK, removeProperty, updateItem, queryByGSI } from './repo-utils';
import { TransactWriteCommandInput } from '@aws-sdk/lib-dynamodb';
import { assembleProjectDetails, buildProjectGSI2_PK } from './projects-repo';
import { ContactType } from '@improovy/types';

const TableName = ddb.clientsTable;

export const fetchClientInterface = async (uuid: string) => {
  const clientItems = await queryByPK(TableName, uuid);
  if (!clientItems.length) return null;
  return assembleClientInterfaceFromDDBItems(clientItems);
};

export const fetchClientInterfaceByPdid = async (pdid: string) => {
  const clientItems = await queryByGSI(TableName, 2, `PDID#${pdid}`);
  if (!clientItems.length) return null;
  return assembleClientInterfaceFromDDBItems(clientItems);
};

export const fetchClientInterfaceByPhoneNumber = async (phoneNumber: string) => {
  const phoneNumberItems = await queryByGSI(TableName, 1, `Contact#${ContactType.PHONE}#${phoneNumber}`);
  if (!phoneNumberItems.length) return null;
  if (phoneNumberItems.length > 1) throw new Error("more than one phone number found");
  
  const clientUuid = phoneNumberItems[0].pk;
  return fetchClientInterface(clientUuid);
};

export const putClient = async (client: Client) => {
  const lowLevelClient = buildLowLevelClient(client);
  const transaction = buildClientPutTransaction(lowLevelClient);
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

// todo: add ability to delete more than 25 items at once/handle paginating
export const hardDeleteClient = async (client: Client) => {
  const clientItems = await queryByPK(TableName, client.uuid);
  const TransactItems = clientItems.map(item => ({
    Delete: {
      TableName,
      Key: {
        pk: item.pk,
        sk: item.sk,
      },
    },
  }));

  const transaction: TransactWriteCommandInput = {
    TransactItems,
  };

  if (!TransactItems.length) {
    return Promise.resolve();
  }
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

export const updateClient = async (client: Client) => {
  const { profile } = buildLowLevelClient(client);
  
  return updateItem(profile, TableName);
};

export const removeClientProperty = async (client: Client) => {
  const { profile } = buildLowLevelClient(client);

  if (profile.lastUpdatedField === undefined) throw new Error("lastUpdatedField needs to be set to determine which property to remove");

  // object reassignment needed so typescript recognizes lastUpdatedField as not undefined
  return removeProperty({...profile, lastUpdatedField: profile.lastUpdatedField}, TableName);
};

export const fetchClientProjectInterfaces = async (client: Client) => {
  const projectItems = await queryByGSI(ddb.projectsTable, 2, buildProjectGSI2_PK({ clientUuid: client.uuid }));

  // @ts-expect-error
  const projectInterfaces = projectItems.map(assembleProjectDetails);

  return projectInterfaces;
};