import { ddb } from '@svc/config';
import { DeleteCommand, DeleteCommandInput, PutCommandInput, PutCommand } from '@aws-sdk/lib-dynamodb';
import { getItem, queryByPK, updateItem } from './repo-utils';
import _ from 'lodash';
import { LowLevelGAConversion, GAConversionInterface, GAConversionType } from '@improovy/types';
import { ProjectInterface } from '@improovy/types';

const TableName = ddb.gaConversionsTable;

export const buildGAConversionSK = <Type extends {gclid: string, projectPdid: string, conversionType: GAConversionType}>(conversion: Type) => {
  return `${conversion.projectPdid}#${conversion.conversionType}`
}

export const buildLowLevelGAConversion = <Type extends GAConversionInterface>(gaConversion: Type): LowLevelGAConversion => {
  const {
    conversionType,
    gclid,
    conversionTime,
    projectPdid,
    isTest,
    lastUpdatedField,
    conversionValue,
    changeSource,
  } = gaConversion;

  const lowLevelGAConversion: LowLevelGAConversion = {
    pk: gclid,
    sk: buildGAConversionSK(gaConversion),
    type: 'GAConversion',
    conversionType,
    projectPdid,
    changeSource,
    gclid,
    conversionTime
  };

  if (isTest) lowLevelGAConversion.isTest = isTest;
  if (lastUpdatedField) lowLevelGAConversion.lastUpdatedField = lastUpdatedField;
  if (typeof conversionValue === 'number') lowLevelGAConversion.conversionValue = conversionValue;

  return lowLevelGAConversion;
};

export const assembleGAConversion = (gaConversion: LowLevelGAConversion): GAConversionInterface => {
  const assembledGAConversion = _.pick(gaConversion, [
    'gclid',
    'conversionType',
    'projectPdid',
    'conversionTime',
    'conversionValue',
    'changeSource',
    'isTest',
    'lastUpdatedField',
  ]) as GAConversionInterface;

  return assembledGAConversion;
};

// this method is basically useless but helps it match the pattern of other entities
export const assembleLowLevelGAConversionInterface = (ddbItem: { [key: string]: any }): LowLevelGAConversion => {
  return ddbItem as LowLevelGAConversion;
};

export const assembleGAConversionInterfaceFromDDBItem = (ddbItem: { [key: string]: any }): GAConversionInterface => {
  const lowLevelGAConversion = assembleLowLevelGAConversionInterface(ddbItem);
  return assembleGAConversion(lowLevelGAConversion);
};

// only really useful for testing
export const fetchGAConversionInterface = async (attributes: {gclid: string, projectPdid: string, conversionType: GAConversionType}) => {
  const gaConversionItem = await getItem(TableName, attributes.gclid, buildGAConversionSK(attributes));
  if (!gaConversionItem) return null;
  return assembleGAConversionInterfaceFromDDBItem(gaConversionItem);
};

export const fetchGAConversionInterfacesByGclid = async (gclid: string) => {
  const gaConversionItems: {[key: string]: any}[] = await queryByPK(TableName, gclid);
  if (!gaConversionItems.length) return null;
  return gaConversionItems.map(item => assembleGAConversionInterfaceFromDDBItem(item))
};

export const fetchGAConversionInterfacesByProject = async <Type extends ProjectInterface>(project: Type) => {
  if (!project.gclid || !project.pdid) return null
  const gaConversionItems: {[key: string]: any}[] = await queryByPK(TableName, project.gclid, project.pdid);
  if (!gaConversionItems.length) return null;
  return gaConversionItems.map(item => assembleGAConversionInterfaceFromDDBItem(item))
};

export const putGAConversion = async <Type extends GAConversionInterface>(gaConversion: Type) => {
  const lowLevelGAConversion = buildLowLevelGAConversion(gaConversion);
  const putCommand: PutCommandInput = {
    TableName,
    Item: lowLevelGAConversion
  }
  
  return ddb.documentClient.send(
    new PutCommand(putCommand),
  );
};

export const hardDeleteGAConversion = async <Conversion extends GAConversionInterface>(gaConversion: Conversion) => {
  const gaConversionItems = await getItem(TableName, gaConversion.gclid, buildGAConversionSK(gaConversion));
  
  const deleteCommand: DeleteCommandInput = {
    TableName,
    Key: {
      pk: gaConversion.gclid,
      sk: buildGAConversionSK(gaConversion)
    }
  }
  
  return ddb.documentClient.send(
    new DeleteCommand(deleteCommand),
  );
};

export const updateGAConversion = async <Conversion extends GAConversionInterface>(gaConversion: Conversion) => {
  const lowLevelGAConversion = buildLowLevelGAConversion(gaConversion);

  return updateItem(lowLevelGAConversion, TableName);
};