import { gaConversionsAPI } from "@svc/config";
import axios from "axios";
import { GAConversionInterface, GAConversionType } from '@improovy/types';
import log from '@dazn/lambda-powertools-logger';

const baseURL = `https://googleads.googleapis.com/v10/customers/${gaConversionsAPI.adsAccountID}`

const formatGoogleTime = (date: Date) => {
  return date.toISOString().replace('T', ' ').replace(/\..../,'').replace('Z', '+00:00')
}

class GoogleAdsApi {
  #accessToken?: string

  get accessToken() {
    return this.#accessToken
  }

  async refreshAccessToken() {
    const resp = await axios.post("https://oauth2.googleapis.com/token", {
      refresh_token: gaConversionsAPI.refreshToken,
      client_id: gaConversionsAPI.credentialsID,
      client_secret: gaConversionsAPI.credentialsSecret,
      grant_type: 'refresh_token'
    })

    this.#accessToken = await resp.data.access_token
  }

  get headers() {
    return {
      Authorization: `Bearer ${this.accessToken}`,
      'developer-token': gaConversionsAPI.apiKey,
    }
  }

  async uploadConversion<Conversion extends GAConversionInterface>(conversion: Conversion) {
    if (!this.accessToken) await this.refreshAccessToken();
    
    const conversions = this.#buildConversionsBody(conversion)

    try {
      const resp = await this.#uploadClickConversion(conversions) 
      if (resp.data.partialFailureError) {
        log.debug('partial failure error', {error: resp.data.partialFailureError})
      }
      return resp
    } catch (error: any) {
      log.debug('error received', {error})
      if (error.response.status === 401) {
        await this.refreshAccessToken()
        const resp = await this.#uploadClickConversion(conversions) 
        return resp
      }
      return error
    }
  }

  async uploadConversionAdjustment<Conversion extends GAConversionInterface>(conversion: Conversion, type: 'RESTATEMENT' | 'RETRACTION' = 'RESTATEMENT') {
    if (!this.accessToken) await this.refreshAccessToken();
    
    const conversions = this.#buildConversionsAdjustmentBody(conversion, type)

    try {
      const resp = await this.#uploadConversionAdjustment(conversions) 
      if (resp.data.partialFailureError) {
        log.debug('partial failure error', {error: resp.data.partialFailureError})
      }
      return resp
    } catch (error: any) {
      log.debug('error received', {error})
      if (error.response.status === 401) {
        await this.refreshAccessToken()
        const resp = await this.#uploadConversionAdjustment(conversions) 
        return resp
      }
      return error
    }
  }

  async retractConversion<Conversion extends GAConversionInterface>(conversion: Conversion) {
    return await this.uploadConversionAdjustment(conversion)
  }

  #buildConversionsBody<Conversion extends GAConversionInterface>(conversion: Conversion) {
    const resourceName = this.#getConversionRN(conversion)

    return {
      conversions: [
        {
          gclid: conversion.gclid,
          conversion_action: resourceName,
          conversion_date_time: formatGoogleTime(new Date(conversion.conversionTime)),
          conversion_value: conversion.conversionValue || 0,
        }
      ],
      partial_failure: true
    }
  }

  #buildConversionsAdjustmentBody<Conversion extends GAConversionInterface>(conversion: Conversion, adjustmentType: 'RESTATEMENT' | 'RETRACTION') {
    const resourceName = this.#getConversionRN(conversion)

    return {
      conversionAdjustments: [
        {
          gclidDateTimePair: {
            gclid: conversion.gclid,
            conversionDateTime: formatGoogleTime(new Date(conversion.conversionTime))
          },
          adjustmentType,
          restatementValue: {
            adjustedValue: conversion.conversionValue,
          },
          conversionAction: resourceName,
          adjustmentDateTime: formatGoogleTime(new Date())
        }
      ],
      partialFailure: true,
    }
  }

  #getConversionRN<Conversion extends GAConversionInterface>(conversion: Conversion) {
    switch (conversion.conversionType) {
      case GAConversionType.BOOKED:
        return gaConversionsAPI.bookedRN;
        break;
    
      case GAConversionType.QUOTE_REQUESTED:
        return gaConversionsAPI.quoteRequestedRN;
        break;

      case GAConversionType.QUOTE_RECEIVED:
        return gaConversionsAPI.quoteReceivedRN;
        break;

      default:
        throw new Error("conversion type not recognized");
    }
  }

  async #uploadClickConversion(conversion: any) {
    return axios.post(`${baseURL}:uploadClickConversions`, conversion, {headers: this.headers})
  }

  async #uploadConversionAdjustment(conversion: any) {
    return axios.post(`${baseURL}:uploadConversionAdjustments`, conversion, {headers: this.headers})
  }
}

export default new GoogleAdsApi();