import { ChangeSource } from '@improovy/types';
import { GAConversionInterface, GAConversionType, MutableGAConversionProperties } from '@improovy/types';
import { ProjectInterface, ProjectSalesStage } from '@improovy/types';
import { fetchGAConversionInterface, fetchGAConversionInterfacesByProject, hardDeleteGAConversion, putGAConversion, updateGAConversion } from '../ddbRepos/gaConversions-repo';

export default class GAConversion implements GAConversionInterface {
  #gclid: string;
  
  #conversionType: GAConversionType;

  #projectPdid: string;
  
  #conversionTime: number;

  #conversionValue?: number;

  changeSource: ChangeSource;

  #isTest?: boolean;

  #lastUpdatedField?: MutableGAConversionProperties;

  static async fetch(gclid: string, projectPdid: string, conversionType: GAConversionType) {
    const gaConversionInterface = await fetchGAConversionInterface({gclid, projectPdid, conversionType});
    if (!gaConversionInterface) return null;
    return new GAConversion(gaConversionInterface)
  }

  static async fetchByProject<ProjectType extends ProjectInterface>(project: ProjectType) {
    const gaConversionInterfaces: null | GAConversionInterface[] = await fetchGAConversionInterfacesByProject(project);
    if (!gaConversionInterfaces) return null;
    return gaConversionInterfaces.map(convInterface => new GAConversion(convInterface))
  }

  static fromProject<ProjectType extends ProjectInterface>(project: ProjectType): GAConversion {
    if (!project.pdid) throw new Error("project pdid required");
    if (!project.gclid) throw new Error("project gclid required");
    
    let conversionType;

    switch (project.salesStage) {
      case ProjectSalesStage.BOOKED:
        conversionType = GAConversionType.BOOKED
        break;
      
      case ProjectSalesStage.ESTIMATED:
        conversionType = GAConversionType.QUOTE_RECEIVED
        break;

      case ProjectSalesStage.UNATTEMPTED:
      case ProjectSalesStage.ATTEMPTED:
      case ProjectSalesStage.ESTABLISHED_CONTACT:
        conversionType = GAConversionType.QUOTE_REQUESTED
        break;
      default:
        throw new Error("projectSalesStage not found");
    }

    const gaConversionInterface: GAConversionInterface = {
      projectPdid: project.pdid,
      changeSource: ChangeSource.aws,
      gclid: project.gclid,
      conversionType,
      conversionTime: Date.now(),
    }

    if (typeof project.clientCost === 'number') gaConversionInterface.conversionValue = project.clientCost;

    return new GAConversion(gaConversionInterface)
  }

  constructor(gaConversionInterface: GAConversionInterface) {
    this.#gclid = gaConversionInterface.gclid;
    this.#conversionType = gaConversionInterface.conversionType;
    this.#projectPdid = gaConversionInterface.projectPdid;
    this.#conversionTime = gaConversionInterface.conversionTime;
    this.changeSource = gaConversionInterface.changeSource;
    if (gaConversionInterface.isTest) this.#isTest = gaConversionInterface.isTest;
    if (gaConversionInterface.lastUpdatedField) this.#lastUpdatedField = gaConversionInterface.lastUpdatedField;
    if (typeof gaConversionInterface.conversionValue === 'number') this.#conversionValue = gaConversionInterface.conversionValue;
  }

  async updateConversionValue(val: number) {
    const fieldsToUpdate = [MutableGAConversionProperties.conversionValue];
    this.#conversionValue = val;
    return this.updateFields(fieldsToUpdate);
  }

  async put() {
    return putGAConversion(this);
  }

  async hardDelete() {
    return hardDeleteGAConversion(this);
  }

  async updateFields(fields: MutableGAConversionProperties[]) {
    for (let field of fields) {
      this.#lastUpdatedField = field;
      await updateGAConversion(this);
    }
  }

  get conversionType() {
    return this.#conversionType;
  }
  
  get projectPdid() {
    return this.#projectPdid;
  }
  
  get conversionTime() {
    return this.#conversionTime;
  }

  get conversionValue() {
    return this.#conversionValue;
  }

  get lastUpdatedField() {
    return this.#lastUpdatedField;
  }

  get gclid() {
    return this.#gclid;
  }

  get isTest() {
    return this.#isTest || false;
  }
  
}