import { UpdateCommand, UpdateCommandInput } from '@aws-sdk/lib-dynamodb';
import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import { buildGAConversionSK } from '@svc/lib/ddbRepos/gaConversions-repo';
import { GAConversionType } from '@improovy/types';
import { ProjectSalesStage } from '@improovy/types';
import uuid from '@svc/lib/uuid';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { lambdaLogsWarmWait, waitForColdLambdaLogs, waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import _ from 'lodash';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBAddGAConversion`;

const lambdaLogs = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

const projectsTable = getTableName('PROJECTS')

const gaConversionTable = {
  region: AWS_REGION,
  table: getTableName('GA_CONVERSIONS')
}

describe('`addGAConversion` Lambda function', () => {
  describe('when a new project is added', () => {
    const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid'], fieldsToExclude: ['isTest']});
    const testConversion = testDataGenerator.getGAConversionInterfaceFromProject(testProject)
    it('is triggered if it has a gclid attribute', async () => {
      await testProject.put();
      const expectedLog = testProject.uuid;
      await expect({
        ...lambdaLogs,
        timeout: 50000,
      }).toHaveLog(expectedLog);
    });

    it('inserts a GAConversion event into the GAConversions table', async () => {
      const gclid = testProject.gclid!;
      const keys = {
        pk: gclid,
        sk: buildGAConversionSK(testConversion)
      }

      await waitForGSIUpdates()
      
      await expect(gaConversionTable).toHaveItem(keys);
    });

    it.concurrent('is not triggered if it does not have a gclid attribute', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.put();
      const unexpectedLog = testProject.uuid;

      await waitForColdLambdaLogs();

      await expect({
        ...lambdaLogs,
      }).not.toHaveLog(unexpectedLog);
    });
  })

  describe('when an existing project is updated', () => {
    describe('with a new gclid', () => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['isTest']});
      const gclid = uuid();
      const testConversion = testDataGenerator.getGAConversionInterfaceFromProject(testProject, {overrides: {gclid}})

      it('is triggered if it receives a new gclid', async () => {
        await testProject.put();

        await testProject.updateGclid(gclid)

        const expectedLog = testProject.uuid;

        await expect({
          ...lambdaLogs,
          timeout: lambdaLogsWarmWait,
        }).toHaveLog(expectedLog);
      });

      it('inserts a GAConversion event into the GAConversions table', async () => {
        const keys = {
          pk: gclid,
          sk: buildGAConversionSK(testConversion)
        }
        
        await waitForGSIUpdates()

        await expect(gaConversionTable).toHaveItem(keys);
      });
    })

    describe('when entering the estimated stage',() => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['isTest']});
      const gclid = uuid();
      const testConversion = testDataGenerator.getGAConversionInterfaceFromProject(testProject, {overrides: {gclid, conversionType: GAConversionType.QUOTE_RECEIVED}});
      it('is triggered if it has a gclid', async () => {
        await testProject.put();

        const update: UpdateCommandInput = {
          TableName: projectsTable,
          Key: {
            pk: testProject.uuid,
            sk: 'Details'
          },
          UpdateExpression: 'set gclid = :gclid',
          ExpressionAttributeValues: {
            ':gclid': gclid
          }
        }

        await ddb.documentClient.send(new UpdateCommand(update))

        await testProject.updateSalesStage(ProjectSalesStage.ESTIMATED)

        const expectedLog = testProject.uuid;

        await expect({
          ...lambdaLogs,
          timeout: 50000,
        }).toHaveLog(expectedLog);
      });

      it('inserts a GAConversion event into the GAConversions table', async () => {
        const keys = {
          pk: gclid,
          sk: buildGAConversionSK({gclid, projectPdid: testProject.pdid!, conversionType: GAConversionType.QUOTE_RECEIVED})
        }

        await waitForGSIUpdates();
        
        await expect(gaConversionTable).toHaveItem(keys);
      });
      
      it.concurrent('is not triggered if it does not have a gclid attribute', async () => {
        const testProject = testDataGenerator.getProject();
        await testProject.put();

        await testProject.updateSalesStage(ProjectSalesStage.ESTIMATED)
        const unexpectedLog = testProject.uuid;
  
        await waitForColdLambdaLogs();
  
        await expect({
          ...lambdaLogs,
        }).not.toHaveLog(unexpectedLog);
      });
    })

    describe('when entering the booked stage',() => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['isTest']});
      const gclid = uuid();
      const testConversion = testDataGenerator.getGAConversionInterfaceFromProject(testProject, {overrides: {gclid, conversionType: GAConversionType.BOOKED}});
      it('is triggered if it has a gclid', async () => {
        await testProject.put();

        const update: UpdateCommandInput = {
          TableName: projectsTable,
          Key: {
            pk: testProject.uuid,
            sk: 'Details'
          },
          UpdateExpression: 'set gclid = :gclid',
          ExpressionAttributeValues: {
            ':gclid': gclid
          }
        }

        await ddb.documentClient.send(new UpdateCommand(update))

        await testProject.updateSalesStage(ProjectSalesStage.BOOKED)

        const expectedLog = testProject.uuid;

        await expect({
          ...lambdaLogs,
          timeout: 50000,
        }).toHaveLog(expectedLog);
      });

      it('inserts a GAConversion event into the GAConversions table', async () => {
        const keys = {
          pk: gclid,
          sk: buildGAConversionSK({gclid, projectPdid: testProject.pdid!, conversionType: GAConversionType.BOOKED})
        }

        await waitForGSIUpdates()
        
        await expect(gaConversionTable).toHaveItem(keys);
      });
      
      it.concurrent('is not triggered if it does not have a gclid attribute', async () => {
        const testProject = testDataGenerator.getProject();
        await testProject.put();

        await testProject.updateSalesStage(ProjectSalesStage.BOOKED)
        const unexpectedLog = testProject.uuid;
  
        await waitForColdLambdaLogs();
  
        await expect({
          ...lambdaLogs,
        }).not.toHaveLog(unexpectedLog);
      });
    })

    describe('when entering a stage other than estimated or booked', () =>{
      it.concurrent('is not triggered if it does not have a gclid', async () => {
        const testProject = testDataGenerator.getProject();
        const gclid = uuid();
        await testProject.put();

        const update: UpdateCommandInput = {
          TableName: projectsTable,
          Key: {
            pk: testProject.uuid,
            sk: 'Details'
          },
          UpdateExpression: 'set gclid = :gclid',
          ExpressionAttributeValues: {
            ':gclid': gclid
          }
        }

        await ddb.documentClient.send(new UpdateCommand(update))

        await testProject.updateSalesStage(ProjectSalesStage.ATTEMPTED)

        const unexpectedLog = testProject.uuid;

        await waitForLambdaLogs()

        await expect({
          ...lambdaLogs,
          timeout: 50000,
        }).not.toHaveLog(unexpectedLog);
      });
    })
  })

  afterAll(async () => await testDataGenerator.deleteAll());
});