import { UpdateCommand } from '@aws-sdk/lib-dynamodb';
import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import Client from '@svc/lib/models/Client';
import { buildLowLevelClient } from '@svc/lib/ddbRepos/clients-repo-utils';
import uuid from '@svc/lib/uuid';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { lambdaLogsWarmWait, waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import { getTableName } from "@improovy/utils"
const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBAddGclidToClientProjects`;

const clientsTable = getTableName('CLIENTS')

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`addGclidToClientProjects` Lambda function', () => {
  it.concurrent('is triggered when a new client is inserted into the Clients DDB table with a gclid attribute', async () => {
    const testCLient = testDataGenerator.getClient({ overrides: { gclid: 'test' } });
    await testCLient.put();
    const expectedLog = testCLient.uuid;
    return expect({
      ...awsTestOptions,
      timeout: 50000,
    }).toHaveLog(expectedLog);

  });
  it.concurrent('is triggered when a gclid is added to a client without a gclid', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();
    const newGclid = uuid();
    await testClient.updateGclid(newGclid);
    return expect({ ...awsTestOptions, timeout: lambdaLogsWarmWait }).toHaveLog(newGclid);
  });

  it.concurrent('is triggered when the gclid on a client with a gclid already is changed', async () => {
    const testClient = testDataGenerator.getClient({ fieldsToGenerate: ['gclid'] });
    await testClient.put();
    const newGclid = uuid();
    await testClient.updateGclid(newGclid);
    return expect({ ...awsTestOptions, timeout: lambdaLogsWarmWait }).toHaveLog(newGclid);

  });
  
  it.concurrent('is not triggered when a new client is inserted into the Clients DDB table without a gclid attribute', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();
    const unexpectedLog = testClient.uuid;
    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });
  it.concurrent('is not triggered when a client is updated in the Clients DDB table without a new gclid attribute', async () => {
    const testClientInterface = testDataGenerator.getClientInterface();
    const testClient = new Client(testClientInterface);
    const testLowLevelClient = buildLowLevelClient(testClient);

    await testClient.put();
    
    // should not trigger because lastUpdatedField does not get set to 'gclid'
    const params = {
      TableName: clientsTable,
      Key: {
        pk: testLowLevelClient.profile.pk,
        sk: testLowLevelClient.profile.sk,
      },
      UpdateExpression: 'set gclid = :gclid',
      ExpressionAttributeValues: {
        ':gclid': 'test_gclid',
      },
    };

    await ddb.documentClient.send(new UpdateCommand(params));

    const unexpectedLog = testClient.uuid;
    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });
  it.concurrent('is not triggered when a gclid is removed from a client', async () => {
    const testClientInterface = testDataGenerator.getClientInterface();
    const testClient = new Client(testClientInterface);
    const testLowLevelClient = buildLowLevelClient(testClient);

    await testClient.put();
    
    const params = {
      TableName: clientsTable,
      Key: {
        pk: testLowLevelClient.profile.pk,
        sk: testLowLevelClient.profile.sk,
      },
      UpdateExpression: 'set gclid = :gclid',
      ExpressionAttributeValues: {
        ':gclid': 'test_gclid',
      },
    };

    await ddb.documentClient.send(new UpdateCommand(params));

    testClient.removeGclid();

    const unexpectedLog = testClient.uuid;
    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });

  it.concurrent('is not triggered when an entity of type other than client is added with a gclid', async () => {
    const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']});
    await testProject.put();

    const unexpectedLog = testProject.gclid!;

    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });
  
  it.concurrent('adds the gclid to all of projects belonging to the client', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();
    
    for (let i = 0; i < 2; i++) {
      const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } },);
      await testProject.put();
    }
    
    await waitForGSIUpdates();

    await testClient.updateGclid('test_gclid');
    
    await waitForGSIUpdates();

    const receivedProjects = await testClient.fetchProjects();

    expect(receivedProjects.length).toBeTruthy()

    for (let project of receivedProjects) {
      expect(project.gclid).toBe(testClient.gclid);
    }
  });

  it.concurrent('does not change the gclid on projects that already have a gclid', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();

    const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid, gclid: 'already_exists' } });
    const testProjectWithoutGclid = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
    await testProject.put();
    await testProjectWithoutGclid.put();

    await waitForGSIUpdates();

    await testClient.updateGclid('test_gclid');
    
    await waitForLambdaLogs();

    const receivedProjects = await testClient.fetchProjects();

    expect(receivedProjects).toHaveLength(2);
    for (let project of receivedProjects) {
      if (project.uuid === testProject.uuid) {
        expect(project.gclid).toBe(testProject.gclid)
      } else if (project.uuid === testProjectWithoutGclid.uuid) {
        expect (project.gclid).toBe(testClient.gclid)
      } else {
        expect(0).toBe(1)
      }
    }
  });

  afterAll(async () => testDataGenerator.deleteAll());
});