import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import CallRail from '@svc/lib/models/CallRail';
import Client from '@svc/lib/models/Client';
import { buildLowLevelCallRail } from '@svc/lib/ddbRepos/callRail-repo';
import { ContactType } from '@improovy/types';
import { ChangeSource } from '@improovy/types';
import generateNewPhoneNumber from '@tests/utils/generateNewPhoneNumber';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import _ from 'lodash';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBAttributeCallRail`;

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`attributeCallRail` Lambda function', () => {
  const phoneNumber = generateNewPhoneNumber();
  const testClient = testDataGenerator.getClient({overrides: {contactMethods: [{
    contactType: ContactType.PHONE,
    contactAddress: phoneNumber,
  }]}});
  const testCallRail = testDataGenerator.getCallRail({overrides: {phoneNumber: `+1${phoneNumber}`}})

  it.concurrent('is triggered when a new CallRail is inserted into the CallRail DDB table', async () => {
    const testCallRail = testDataGenerator.getCallRail();
    await testCallRail.put();
    const expectedLog = testCallRail.gclid;
    return expect({
      ...awsTestOptions,
      timeout: 50000,
    }).toHaveLog(expectedLog);

  });

  it.concurrent('returns without updating the callRail if no client matches the phone number', async () => {
    const testCallRail = testDataGenerator.getCallRail({overrides: { phoneNumber: 'nonexistentPhoneNumber0'}});
    await testCallRail.put();

    const lowLevel = buildLowLevelCallRail(testCallRail)

    await waitForLambdaLogs();

    return expect({region: AWS_REGION, table: getTableName('CALLRAIL')}).toHaveItem(_.pick(lowLevel, ['pk', 'sk']), lowLevel);

  });
  
  it('adds the gclid to the client with a matching phone number and sets its changeSource', async () => {
    await testClient.put();
    
    await waitForLambdaLogs(); // needed to allow checkForUnattributedCallRail to process

    await testCallRail.put();

    await waitForGSIUpdates();

    const receivedClient = await Client.fetch(testClient.uuid);

    if (!receivedClient) {
      return expect(receivedClient).toBeTruthy();
    }

    expect(receivedClient.gclid).toBe(testCallRail.gclid)
    expect(receivedClient.changeSource).toBe(ChangeSource.aws)
  });

  it('marks the callRail as attributed', async () => {
    const receivedCallRail = await CallRail.fetch(phoneNumber);

    if (!receivedCallRail) {
      return expect(receivedCallRail).toBeTruthy();
    }

    expect(receivedCallRail.isAttributed).toBe(true)
    expect(receivedCallRail.changeSource).toBe(ChangeSource.aws)
  });

  afterAll(async () => testDataGenerator.deleteAll());
});