import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import CallRail from '@svc/lib/models/CallRail';
import Client from '@svc/lib/models/Client';
import { buildLowLevelClientProfile } from '@svc/lib/ddbRepos/clients-repo-utils';
import { ClientIncomingContactMethod, ContactType } from '@improovy/types';
import { ChangeSource } from '@improovy/types';
import generateNewPhoneNumber from '@tests/utils/generateNewPhoneNumber';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForColdLambdaLogs, waitForLambdaLogs } from '@tests/utils/wait';
import _ from 'lodash';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBCheckForUnattrCallRail`;

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`checkForUnattrCallRail` Lambda function', () => {
  const phoneNumber = generateNewPhoneNumber();
  const testClient = testDataGenerator.getClient({overrides: {incomingContactMethod: ClientIncomingContactMethod.PHONE, contactMethods: [{
    contactType: ContactType.PHONE,
    contactAddress: phoneNumber,
  }]}});
  const testCallRail = testDataGenerator.getCallRail({overrides: {phoneNumber: `+1${phoneNumber}`}})
  
  it('is triggered when a new Client is inserted into the Clients DDB table with an incomingContactMethod of PHONE and no GCLID', async () => {
    await testCallRail.put();

    await waitForLambdaLogs(); // wait for attributeCallRail to process

    await testClient.put();

    const expectedLog = testClient.uuid;

    return expect({
      ...awsTestOptions,
      timeout: 50000,
    }).toHaveLog(expectedLog);
  });

  it.concurrent('is not triggered when a new Client is inserted into the Clients DDB table without an incomingContactMethod of PHONE and no gclid', async () => {
    const unexpectedClient = testDataGenerator.getClient();

    await unexpectedClient.put();

    await waitForColdLambdaLogs();

    const unexpectedLog = unexpectedClient.uuid;

    return expect({
      ...awsTestOptions,
    }).not.toHaveLog(unexpectedLog);

  });

  it.concurrent('is not triggered when a new Client is inserted into the Clients DDB table with an incomingContactMethod of PHONE and a gclid', async () => {
    const unexpectedClient = testDataGenerator.getClient({fieldsToGenerate: ['gclid'], overrides: { incomingContactMethod: ClientIncomingContactMethod.PHONE}});

    await unexpectedClient.put();

    const unexpectedLog = unexpectedClient.uuid;

    await waitForColdLambdaLogs();

    return expect({
      ...awsTestOptions,
    }).not.toHaveLog(unexpectedLog);

  });

  it.concurrent('returns without updating the client if no callrail record matches the phone number', async () => {
    const phoneNumber = generateNewPhoneNumber();
    const testClient = testDataGenerator.getClient({overrides: {incomingContactMethod: ClientIncomingContactMethod.PHONE, contactMethods: [{
      contactType: ContactType.PHONE,
      contactAddress: phoneNumber,
    }]}});

    await testClient.put();

    await waitForColdLambdaLogs();

    const lowLevel = buildLowLevelClientProfile(testClient)

    return expect({region: AWS_REGION, table: getTableName('CLIENTS')}).toHaveItem(_.pick(lowLevel, ['pk', 'sk']), lowLevel);

  });
  
  it('adds the gclid from the callrail with a matching phone number and sets the clients changeSource', async () => {
    const receivedClient = await Client.fetch(testClient.uuid);

    if (!receivedClient) {
      return expect(receivedClient).toBeTruthy();
    }

    expect(receivedClient.gclid).toBe(testCallRail.gclid)
    expect(receivedClient.changeSource).toBe(ChangeSource.aws)
  });

  it('marks the callRail as attributed', async () => {
    const receivedCallRail = await CallRail.fetch(phoneNumber);

    if (!receivedCallRail) {
      return expect(receivedCallRail).toBeTruthy();
    }

    expect(receivedCallRail.isAttributed).toBe(true)
    expect(receivedCallRail.changeSource).toBe(ChangeSource.aws)
  });

  afterAll(async () => testDataGenerator.deleteAll());
});