import { TransactWriteCommand, UpdateCommand } from '@aws-sdk/lib-dynamodb';
import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import Client from '@svc/lib/models/Client';
import Project from '@svc/lib/models/Project';
import { buildLowLevelClient } from '@svc/lib/ddbRepos/clients-repo-utils';
import { buildLowLevelProjectDetails, removeProjectProperty, buildProjectPutTransaction } from '@svc/lib/ddbRepos/projects-repo';
import { updateItem } from '@svc/lib/ddbRepos/repo-utils';
import { MutableProjectProperties } from '@improovy/types';
import uuid from '@svc/lib/uuid';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import wait, { lambdaLogsWarmWait, waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBCheckProjClientForGclid`;

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`checkProjectClientForGclid` Lambda function', () => {
  it.concurrent('is triggered when a new project is inserted into the Projects DDB table with a client UUID attribute', async () => {
    const testProject = testDataGenerator.getProject();
    await testProject.put();
    const expectedLog = testProject.uuid;
    return expect({
      ...awsTestOptions,
      timeout: 50000,
    }).toHaveLog(expectedLog);
  });

  it.concurrent('is triggered when the clientUuid on a project is updated', async () => {
    const testProject = testDataGenerator.getProject();
    await testProject.put();
    const newClientUuid = uuid();
    await testProject.updateClientUuid(newClientUuid);
    return expect({ ...awsTestOptions, timeout: lambdaLogsWarmWait }).toHaveLog(newClientUuid);
  });
  
  it.concurrent('is not triggered when a new project is inserted into the Projects DDB table without a client Uuid attribute', async () => {
    const testProject = testDataGenerator.getProject({ fieldsToExclude: ['clientUuid'] });
    
    const testLowLevelProject = buildLowLevelProjectDetails(testProject);

    await testProject.put();

    const unexpectedLog = testProject.uuid!;

    await waitForLambdaLogs();

    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });

  it.concurrent('is not triggered when a client is updated in the Clients DDB table without a new client Uuid attribute', async () => {
    const testProjectInterface = testDataGenerator.getProjectInterface({fieldsToExclude: ['clientUuid']});
    const testProject = new Project(testProjectInterface);

    await testProject.put();
    
    const newGclid = uuid();

    await testProject.updateGclid(newGclid);

    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(newGclid);
  });

  it.concurrent('is not triggered when a client Uuid is removed from a project', async () => {
    const testProjectInterface = testDataGenerator.getProjectInterface({fieldsToExclude: ['clientUuid']});
    const testProject = new Project(testProjectInterface);

    
    const testLowLevelProject = buildLowLevelProjectDetails(testProject);

    const transaction = buildProjectPutTransaction(testLowLevelProject);

    await ddb.documentClient.send(new TransactWriteCommand(transaction));

    const params = {
      TableName: getTableName('PROJECTS'),
      Key: {
        pk: testLowLevelProject.pk,
        sk: testLowLevelProject.sk,
      },
      UpdateExpression: 'set clientUuid = :clientUuid',
      ExpressionAttributeValues: {
        ':clientUuid': 'test_clientUuid',
      },
    };

    await ddb.documentClient.send(new UpdateCommand(params));

    testProjectInterface.lastUpdatedField = MutableProjectProperties.clientUuid;

    // @ts-expect-error
    await removeProjectProperty(testProjectInterface);
    
    const unexpectedLog = testProject.uuid!;

    await waitForLambdaLogs();

    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });

  it.concurrent('is not triggered when an entity of type other than client is added with a clientUuid', async () => {
    const testClientInterface = testDataGenerator.getClientInterface();
    const testClient = new Client(testClientInterface);

    await testClient.put();

    const { profile } = buildLowLevelClient(testClient);

    // @ts-expect-error
    profile.clientUuid = 'test-uuid';

    // @ts-expect-error
    profile.lastUpdatedField = MutableProjectProperties.clientUuid;

    await updateItem(profile, getTableName('CLIENTS'));

    const unexpectedLog = profile.uuid;
    await waitForLambdaLogs();
    return expect(awsTestOptions).not.toHaveLog(unexpectedLog);
  });
  
  it.concurrent('adds the clients` gclid to the project', async () => {
    const testClient = testDataGenerator.getClient({ fieldsToGenerate: ['gclid'] });
    const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
    await testClient.put();
    await wait(15000); // need to wait for addGclidToClientProjects to finish on this one
    await testProject.put();
    
    await waitForGSIUpdates();

    const receivedProject = await Project.fetch(testProject.uuid);

    expect(receivedProject.gclid).toBe(testClient.gclid);
  });

  it.concurrent('does not change the gclid on projects whose client does not have a gclid', async () => {
    const testClient = testDataGenerator.getClient();
    const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
    await testClient.put();
    await testProject.put();
    
    await waitForGSIUpdates();

    const receivedProject = await Project.fetch(testProject.uuid);

    expect(receivedProject.gclid).toBeUndefined;
  });

  it.concurrent('does not change the gclid on projects that already have a gclid', async () => {
    const testProject = testDataGenerator.getProject({ fieldsToGenerate: ['gclid'] });
    await testProject.put();
    
    await waitForGSIUpdates();

    const receivedProject = await Project.fetch(testProject.uuid);

    expect(receivedProject.gclid).toBeUndefined;
  });

  afterAll(async () => testDataGenerator.deleteAll());
});