import { AWS_REGION, ddb, lambdaConfig } from '@svc/config';
import GAConversion from '@svc/lib/models/GAConversion';
import { buildLowLevelGAConversion } from '@svc/lib/ddbRepos/gaConversions-repo';
import uuid from '@svc/lib/uuid';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import _ from 'lodash';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBRemoveGAConversion`;

const lambdaLogs = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

const gaConversionTable = {
  region: AWS_REGION,
  table: getTableName('GA_CONVERSIONS')
}

describe('`addGAConversion` Lambda function', () => {
  it('is triggered when a project with a gclid is flagged as deleted', async () => {
    const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
    await testProject.put();
    await testProject.softDelete();

    const expectedLog = testProject.uuid;
    await expect({
      ...lambdaLogs,
      timeout: 50000,
    }).toHaveLog(expectedLog);
  });

  it('is triggered when a project`s gclid is removed', async () => {
    const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
    await testProject.put();
    await testProject.removeGclid();
    
    const expectedLog = testProject.uuid;
    await expect({
      ...lambdaLogs,
      timeout: 50000,
    }).toHaveLog(expectedLog);
  });

  it('removes all conversions associated with the project the conversions table', async () => {
    const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
    const testConversionInterface = testDataGenerator.getGAConversionInterfaceFromProject(testProject) 
    const testConversion = new GAConversion(testConversionInterface)
    const lowLevelTestConversion = buildLowLevelGAConversion(testConversion)

    await testConversion.put();
    await testProject.put();
    await testProject.removeGclid();

    await waitForGSIUpdates();
    
    const expectedLog = testProject.uuid;

    await expect(gaConversionTable).not.toHaveItem(_.pick(lowLevelTestConversion, ['pk', 'sk']));
  });

  it('is not triggered if a deleted project does not have a gclid', async () => {
    const testProject = testDataGenerator.getProject();
    await testProject.put();
    const unexpectedLog = testProject.uuid;

    await testProject.softDelete();

    await waitForLambdaLogs();

    await expect({
      ...lambdaLogs,
    }).not.toHaveLog(unexpectedLog);
  });

  it('is not triggered if a project is updated with a gclid', async () => {
    const testProject = testDataGenerator.getProject();
    await testProject.put();
    const unexpectedLog = uuid();
    await testProject.updateGclid(unexpectedLog)

    await waitForLambdaLogs();

    await expect({
      ...lambdaLogs,
    }).not.toHaveLog(unexpectedLog);
  });

  afterAll(async () => testDataGenerator.deleteAll());
});