import { AWS_REGION, ddb } from '@svc/config';
import CallRail from '@svc/lib/models/CallRail';
import { buildLowLevelCallRail } from '@svc/lib/ddbRepos/callRail-repo';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import _ from 'lodash';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';

const testDataGenerator = new TestDataGenerator();

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.callRailTable,
};

describe('`CallRail`', () => {
  describe('`CallRail::fetch', () => {
    it('fetches a callRail via phoneNumber', async () => {
      const testCallRail = testDataGenerator.getCallRail();
      await testCallRail.put();
      const testPhoneNumber = testCallRail.phoneNumber;
  
      const returnedCallRail = await CallRail.fetch(testPhoneNumber);
  
      expect(returnedCallRail!.phoneNumber).toBe(testPhoneNumber);
      expect(returnedCallRail).toEqual(expect.objectContaining(testCallRail));
      expect(testCallRail).toEqual(expect.objectContaining(returnedCallRail));
    });

    it('returns null if the Callrail does not exist', async () => {
      const returnedCallRail = await CallRail.fetch('nonexistentPhoneNumber');
  
      expect(returnedCallRail).toBeNull();
    });
  });

  describe('`CallRail#put`', () => {
    it('adds the callRail to the callRail DynamoDB table', async () => {
      const testCallRail = testDataGenerator.getCallRail();
      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);

      testCallRail.put();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelCallRail, ['pk', 'sk']), {
        ...lowLevelCallRail,
      });
    });
  });

  describe('`CallRail#hardDelete`', () => {
    it('removes all items associated with the callRail', async () => {
      const testCallRail = testDataGenerator.getCallRail();
      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);

      await testCallRail.put();

      await testCallRail.hardDelete();

      await waitForGSIUpdates()

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelCallRail, ['pk', 'sk']));

      const fetchedItems = await queryByPK(ddb.callRailTable, testCallRail.phoneNumber);
      
      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`callRail#updateIsAttributed`', () => {
    it('updates the callRail item with the new isAttributed value and lastUpdatedField', async () => {
      const testCallRail = testDataGenerator.getCallRail({fieldsToGenerate: ['isAttributed']});

      const lowLevelCallRail = buildLowLevelCallRail(testCallRail)

      await testCallRail.put();

      await waitForGSIUpdates();
      
      await testCallRail.updateIsAttributed(true);

      const keys = _.pick(lowLevelCallRail, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelCallRail, lastUpdatedField: 'isAttributed', isAttributed: true });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});