import { AWS_REGION, ddb } from '@svc/config';
import Client from '@svc/lib/models/Client';
import { buildLowLevelClient } from '@svc/lib/ddbRepos/clients-repo-utils';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import { ContactType, MutableClientProperties } from '@improovy/types';
import generateNewPhoneNumber from '@tests/utils/generateNewPhoneNumber';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import _ from 'lodash';
import { getTableName } from '@improovy/utils';

const testDataGenerator = new TestDataGenerator();

const clientsTable = getTableName('CLIENTS')

const awsTestOptions = {
  region: AWS_REGION,
  table: clientsTable,
};

describe('`Client`', () => {
  describe('`Client#put`', () => {
    it('adds the client to the clients DynamoDB table', async () => {
      const testClient = testDataGenerator.getClient();
      const lowLevelClient = buildLowLevelClient(testClient);

      testClient.put();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelClient.profile, ['pk', 'sk']), {
        ...lowLevelClient.profile,
      });

      for (let contactMethod of lowLevelClient.contactMethods) {
        await expect(awsTestOptions).toHaveItem(_.pick(contactMethod, ['pk', 'sk']), { ...contactMethod });
      }
    });
  });

  describe('`Client::fetch', () => {
    it('fetches a client via uuid', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      const testUuid = testClient.uuid;
  
      const returnedClient = await Client.fetch(testUuid);
  
      expect(returnedClient!.uuid).toBe(testUuid);
      expect(returnedClient).toEqual(expect.objectContaining(testClient));
      expect(testClient).toEqual(expect.objectContaining(returnedClient));
    });

    it('returns null if the client does not exist', async () => {
      const returnedClient = await Client.fetch('nonexistent uuid');
  
      expect(returnedClient).toBeNull();
    });
  });

  describe('`Client::fetchByPdid', () => {
    it('fetches a client via pdid', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      const testPdid = testClient.pdid!;

      await waitForGSIUpdates();
  
      const returnedClient = await Client.fetchByPdid(testPdid);
      
      expect(returnedClient).toBeTruthy();
      if (!returnedClient) return;
      
      expect(returnedClient.pdid).toBe(testPdid);
      expect(returnedClient.uuid).toBe(testClient.uuid);
      expect(returnedClient).toEqual(expect.objectContaining(testClient));
      expect(testClient).toEqual(expect.objectContaining(returnedClient));
    });

    it('return null if the client does not exist', async () => {
      const returnedClient = await Client.fetchByPdid('non-existent PDID');
  
      expect(returnedClient).toBe(null);
    });
  });

  describe('`Client::fetchByPhoneNumber`', () => {
    it('fetches a client via phone number', async () => {
      const phoneNumber = generateNewPhoneNumber();
      const testClient = testDataGenerator.getClient({overrides: { contactMethods: [{
        contactType: ContactType.PHONE,
        contactAddress: phoneNumber,
      }]}});

      await testClient.put();

      await waitForGSIUpdates();
  
      const returnedClient = await Client.fetchByPhoneNumber(phoneNumber);
      
      expect(returnedClient).toBeTruthy();
      if (!returnedClient) return;
      
      expect(returnedClient.uuid).toBe(testClient.uuid);
      expect(returnedClient).toEqual(expect.objectContaining(testClient));
      expect(testClient).toEqual(expect.objectContaining(returnedClient));
    });

    it('returns null if the client does not exist', async () => {
      const returnedClient = await Client.fetchByPhoneNumber('nonexistentPhoneNumber');
  
      expect(returnedClient).toBe(null);
    });
  });

  describe('`Client#updateFields`', () => {
    it('updates the fields passed in and only those passed in', async () => {
      const testClientInterface1 = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClientInterface2 = { ...testClientInterface1, gclid: 'testing2', isDeleted: true, pdid: '5' };
      const testClient1 = new Client(testClientInterface1);
      const testClient2 = new Client(testClientInterface2);
      const lowLevelClient = buildLowLevelClient(testClient2);
      const expectedClient = new Client({ ...testClientInterface2, pdid: testClientInterface1.pdid, lastUpdatedField: MutableClientProperties.isDeleted });
      const expectedLowLevel = buildLowLevelClient(expectedClient);

      await testClient1.put();

      await testClient2.updateFields([MutableClientProperties.gclid, MutableClientProperties.isDeleted]);

      expect(awsTestOptions).toHaveItem(_.pick(lowLevelClient.profile, ['pk', 'sk']), expectedLowLevel.profile);
    });
  });

  describe('`Client#hardDelete`', () => {
    it('removes all items associated with the client', async () => {
      const testClient = testDataGenerator.getClient();
      const lowLevelClient = buildLowLevelClient(testClient);

      await testClient.put();

      await testClient.hardDelete();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelClient.profile, ['pk', 'sk']));

      for (let contactMethod of lowLevelClient.contactMethods) {
        await expect(awsTestOptions).not.toHaveItem(_.pick(contactMethod, ['pk', 'sk']));
      }

      const fetchedItems = await queryByPK(ddb.clientsTable, testClient.uuid);
      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`Client#updateGclid`', () => {
    it('updates the DDB client profile item with the new gclid and lastUpdatedField', async () => {
      const testClient = testDataGenerator.getClient();
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      await testClient.updateGclid('test_gclid');

      const keys = _.pick(profile, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...profile, lastUpdatedField: 'gclid', gclid: 'test_gclid' });
    });

    it('removes the gclid property if it receives a falsy value', async () => {
      const testClient = testDataGenerator.getClient({fieldsToGenerate: ['gclid']});
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      await testClient.updateGclid(undefined);

      const keys = _.pick(profile, ['pk', 'sk']);

      delete profile.gclid;

      await expect(awsTestOptions).toHaveItem(keys, { ...profile, lastUpdatedField: 'gclid' });
    });
  });

  describe('`Client#removeGclid`', () => {
    it('removes the gclid property from the ddb item and updates lastUpdatedField', async () => {
      const testClient = testDataGenerator.getClient({ fieldsToGenerate: ['gclid'] });
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      await testClient.removeGclid();

      const keys = _.pick(profile, ['pk', 'sk']);

      delete profile.gclid;

      await expect(awsTestOptions).toHaveItem(keys, { ...profile, lastUpdatedField: 'gclid' });
    });
  });

  describe('`Client#softDelete`', () => {
    it('updates the DDB client profile item with isDeleted set to true and lastUpdatedField', async () => {
      const testClient = testDataGenerator.getClient();
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      await testClient.softDelete();

      const keys = _.pick(profile, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...profile, lastUpdatedField: 'isDeleted', isDeleted: true });
    });
  });

  describe('`Client#fetchProjects`', () => {
    it('fetches all of the clients` projects', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      
      const testProjects = [];
      for (let i = 0; i < 2; i++) {
        const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
        await testProject.put();
        testProjects.push(testProject);
      }
      
      await waitForGSIUpdates();

      const receivedProjects = await testClient.fetchProjects();

      expect(receivedProjects).toHaveLength(testProjects.length);
      for (let project of testProjects) {
        expect(receivedProjects.some(({ uuid }) => uuid === project.uuid)).toBe(true);
      }
    });
  });

  describe('`Client#forceFetchProjects`', () => {
    it('fetches all of the clients` projects', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      
      const testProjects = [];
      for (let i = 0; i < 2; i++) {
        const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
        await testProject.put();
        testProjects.push(testProject);
      }
      
      await waitForGSIUpdates();

      const receivedProjects = await testClient.fetchProjects();

      expect(receivedProjects).toHaveLength(testProjects.length);
      for (let project of testProjects) {
        expect(receivedProjects.some(({ uuid }) => uuid === project.uuid)).toBe(true);
      }
    });
    it('fetches all of the clients` projects even if one was added after initializing Client#projects', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      
      const testProjects = [];
      for (let i = 0; i < 3; i++) {
        const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
        await testProject.put();
        testProjects.push(testProject);
        if (i === 1) {
          await testClient.fetchProjects();
        }
      }
      
      await waitForGSIUpdates();

      const receivedProjects = await testClient.forceFetchProjects();

      expect(receivedProjects).toHaveLength(testProjects.length);
      for (let project of testProjects) {
        expect(receivedProjects.some(({ uuid }) => uuid === project.uuid)).toBe(true);
      }
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});