import { AWS_REGION, ddb } from '@svc/config';
import GAConversion from '@svc/lib/models/GAConversion';
import _ from 'lodash';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import { GAConversionType, MutableGAConversionProperties } from '@improovy/types';
import { buildLowLevelGAConversion, putGAConversion } from '@svc/lib/ddbRepos/gaConversions-repo';

const testDataGenerator = new TestDataGenerator();

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.gaConversionsTable,
};

describe('`GAConversion`', () => {
  describe('`GAConversion::fetch', () => {
    it('fetches a gaConversion via gclid, projectPdid, and conversion type', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      await putGAConversion(testGAConversion)
  
      const returnedGAConversion = await GAConversion.fetch(testGAConversion.gclid, testGAConversion.projectPdid, testGAConversion.conversionType);
  
      expect(returnedGAConversion).toStrictEqual(testGAConversion);
    });

    it('returns null if the gaConversion is not found', async () => {
      const returnedGAConversion = await GAConversion.fetch('blah', 'blah', GAConversionType.BOOKED)
  
      expect(returnedGAConversion).toBe(null);
    });
  });

  describe('`GAConversion::fetchByProject', () => {
    it('fetches the gaConversions matching the project', async () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
      const testGclid = testProject.gclid
      const overrides = {
        gclid: testGclid,
        projectPdid: testProject.pdid,
        conversionType: GAConversionType.QUOTE_REQUESTED
      }
      const testGAConversion1 = testDataGenerator.getGAConversion({overrides});
      const testGAConversion2 = testDataGenerator.getGAConversion({overrides: {...overrides, conversionType: GAConversionType.QUOTE_RECEIVED}, fieldsToGenerate: ['conversionValue']});
      const testGAConversion3 = testDataGenerator.getGAConversion({overrides: {...overrides, conversionType: GAConversionType.BOOKED}, fieldsToGenerate: ['conversionValue']});
  
      await putGAConversion(testGAConversion1);
      await putGAConversion(testGAConversion2);
      await putGAConversion(testGAConversion3);
  
      await waitForGSIUpdates();
  
      const receivedInterfaces = await GAConversion.fetchByProject(testProject);
  
      expect(receivedInterfaces).toHaveLength(3);
      expect(receivedInterfaces).toContainEqual(testGAConversion1)
      expect(receivedInterfaces).toContainEqual(testGAConversion2)
      expect(receivedInterfaces).toContainEqual(testGAConversion3)
    });

    it('returns null if the gaConversion is not found', async () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
      const returnedGAConversion = await GAConversion.fetchByProject(testProject)
  
      expect(returnedGAConversion).toBe(null);
    });
  });

  describe('`GAConversion#put`', () => {
    it('adds the gaConversion to the gaConversion DynamoDB table', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      testGAConversion.put();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelGAConversion, ['pk', 'sk']), {
        ...lowLevelGAConversion,
      });
    });
  });

  describe('`GAConversion#hardDelete`', () => {
    it('removes all items associated with the gaConversion', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      await testGAConversion.put();

      await waitForGSIUpdates();

      await testGAConversion.hardDelete();

      await waitForGSIUpdates();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelGAConversion, ['pk', 'sk']));
    });
  });

  describe('`GAConversion#updateFields`', () => {
    it('updates the fields passed in and only those passed in', async () => {
      const testGAConversionInterface1 = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });
      const testGAConversionInterface2 = { ...testGAConversionInterface1, conversionValue: 1001 };
      const testGAConversion1 = new GAConversion(testGAConversionInterface1);
      const testGAConversion2 = new GAConversion(testGAConversionInterface2);
      const lowLevelAcivity = buildLowLevelGAConversion(testGAConversion2);
      const expectedGAConversion = new GAConversion({ ...testGAConversionInterface1, conversionValue: 1001, lastUpdatedField: MutableGAConversionProperties.conversionValue });
      const expectedLowLevel = buildLowLevelGAConversion(expectedGAConversion);

      await testGAConversion1.put();

      await waitForGSIUpdates();

      await testGAConversion2.updateFields([MutableGAConversionProperties.conversionValue]);

      await waitForGSIUpdates();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelAcivity, ['pk', 'sk']), expectedLowLevel);
    });
  });
  
  describe('`GAConversion#updateBody`', () => {
    it('updates the DDB gaConversion item with the new value and lastUpdatedField', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      const lowLevelAcivity = buildLowLevelGAConversion(testGAConversion)

      await testGAConversion.put();

      await waitForGSIUpdates();
      
      await testGAConversion.updateConversionValue(1001);

      const keys = _.pick(lowLevelAcivity, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelAcivity, lastUpdatedField: 'conversionValue', conversionValue: 1001 });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});