import { AWS_REGION, ddb } from '@svc/config';
import {
  fetchClientInterface,
  fetchClientInterfaceByPdid,
  fetchClientInterfaceByPhoneNumber,
  fetchClientProjectInterfaces,
  hardDeleteClient,
  putClient,
  removeClientProperty,
  updateClient,
} from '@svc/lib/ddbRepos/clients-repo';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import _ from 'lodash';
import { buildLowLevelClient } from '@svc/lib/ddbRepos/clients-repo-utils';
import Client from '@svc/lib/models/Client';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import { ContactType, MutableClientProperties } from '@improovy/types';
import { assembleProjectDetails, buildLowLevelProjectDetails } from '@svc/lib/ddbRepos/projects-repo';
import { waitForGSIUpdates } from '@tests/utils/wait';
import generateNewPhoneNumber from '@tests/utils/generateNewPhoneNumber';
import { getTableName } from '@improovy/utils';

const clientsTable = getTableName('CLIENTS')

const awsTestOptions = {
  region: AWS_REGION,
  table: clientsTable,
};

const testDataGenerator = new TestDataGenerator();

describe('`Client` repo', () => {
  describe('`putClient`', () => {
    it('adds the client instance to the clients DynamoDB table', async () => {
      const testClient = testDataGenerator.getClient();
      const lowLevelClient = buildLowLevelClient(testClient);

      await putClient(testClient);

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelClient.profile, ['pk', 'sk']), {
        ...lowLevelClient.profile,
      });

      for (let contactMethod of lowLevelClient.contactMethods) {
        await expect(awsTestOptions).toHaveItem(_.pick(contactMethod, ['pk', 'sk']), { ...contactMethod });
      }
    });
  });

  describe('`fetchClientInterface`', () => {
    it('returns a highlevel client interface of the requested client', async () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = new Client(testClientInterface);

      await testClient.put();
      await waitForGSIUpdates();
      const receivedInterface = await fetchClientInterface(testClient.uuid);

      expect(receivedInterface).toEqual(expect.objectContaining(testClientInterface));
      expect(testClientInterface).toEqual(expect.objectContaining(receivedInterface));
    });

    it('returns null if the client does not exist', async () => {
      const receivedInterface = await fetchClientInterface('non-existest uuid');

      expect(receivedInterface).toBe(null);
    });
  });

  describe('`fetchClientInterfaceByPdid`', () => {
    it('returns a highlevel client interface of the requested client', async () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = new Client(testClientInterface);

      await testClient.put();
      
      await waitForGSIUpdates();

      const receivedInterface = await fetchClientInterfaceByPdid(testClient.pdid!);

      expect(receivedInterface).toEqual(expect.objectContaining(testClientInterface));
      expect(testClientInterface).toEqual(expect.objectContaining(receivedInterface));
    });

    it('returns null if the client does not exist', async () => {
      const receivedInterface = await fetchClientInterfaceByPdid('non-existest pdid');

      expect(receivedInterface).toBe(null);
    });
  });

  describe('`fetchClientInterfaceByPhoneNumber`', () => {
    it('returns a highlevel client interface of the requested client', async () => {
      const phoneNumber = generateNewPhoneNumber();

      const testClientInterface = testDataGenerator.getClientInterface({overrides: { contactMethods: [{
        contactType: ContactType.PHONE,
        contactAddress: phoneNumber,
      }]}});

      const testClient = new Client(testClientInterface);

      await testClient.put();
      
      await waitForGSIUpdates();

      const receivedInterface = await fetchClientInterfaceByPhoneNumber(phoneNumber);

      expect(receivedInterface).toEqual(expect.objectContaining(testClientInterface));
      expect(testClientInterface).toEqual(expect.objectContaining(receivedInterface));
    });

    it('returns null if the client does not exist', async () => {
      const receivedInterface = await fetchClientInterfaceByPhoneNumber('non-existentPhoneNumber');

      expect(receivedInterface).toBe(null);
    });
  });

  describe('`hardDeleteClient`', () => {
    it('removes all items associated with the client', async () => {
      const testClient = testDataGenerator.getClient();
      const lowLevelClient = buildLowLevelClient(testClient);

      await putClient(testClient);

      await hardDeleteClient(testClient);

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelClient.profile, ['pk', 'sk']));

      for (let contactMethod of lowLevelClient.contactMethods) {
        await expect(awsTestOptions).not.toHaveItem(_.pick(contactMethod, ['pk', 'sk']));
      }

      const fetchedItems = await queryByPK(clientsTable, testClient.uuid);
      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`updateClient`', () => {
    it('updates the client lastUpdatedField and field referenced by lastUpdatedField', async () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = new Client(testClientInterface);
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      testClientInterface.lastUpdatedField = MutableClientProperties.gclid;
      testClientInterface.gclid = 'test_gclid';

      // @ts-expect-error
      await updateClient(testClientInterface);

      const keys = _.pick(profile, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...profile, lastUpdatedField: 'gclid', gclid: 'test_gclid' });
    });
  });

  describe('`removeClientProperty`', () => {
    it('updates the client lastUpdatedField and removes the property referenced by lastUpdatedField', async () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = new Client(testClientInterface);
      const { profile } = buildLowLevelClient(testClient);

      await testClient.put();
      
      testClientInterface.lastUpdatedField = MutableClientProperties.gclid;
      
      // @ts-expect-error
      await removeClientProperty(testClientInterface);
      
      const keys = _.pick(profile, ['pk', 'sk']);
      await expect(awsTestOptions).not.toHaveItem(keys, { ...profile });
      
      delete profile.gclid;
      profile.lastUpdatedField = MutableClientProperties.gclid;
      await expect(awsTestOptions).toHaveItem(keys, { ...profile });
    });
  });

  describe('fetchClientProjectInterfaces', () => {
    it('fetches all the projectDetails for projects belonging to a client', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      
      const testProjects = [];
      for (let i = 0; i < 2; i++) {
        const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
        await testProject.put();
        testProjects.push(assembleProjectDetails(buildLowLevelProjectDetails(testProject)));
      }
      
      await waitForGSIUpdates();

      const receivedProjects = await fetchClientProjectInterfaces(testClient);

      expect(receivedProjects).toHaveLength(testProjects.length);
      expect(receivedProjects).toEqual(expect.arrayContaining(testProjects));
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});