import { AWS_REGION, ddb } from '@svc/config';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import _ from 'lodash';
import {
  buildLowLevelGAConversion,
  fetchGAConversionInterface,
  fetchGAConversionInterfacesByGclid,
  fetchGAConversionInterfacesByProject,
  hardDeleteGAConversion,
  putGAConversion,
  updateGAConversion,
} from '@svc/lib/ddbRepos/gaConversions-repo';
import GAConversion from '@svc/lib/models/GAConversion';
import { waitForGSIUpdates } from '@tests/utils/wait';
import { GAConversionType, MutableGAConversionProperties } from '@improovy/types';
import uuid from '@svc/lib/uuid';

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.gaConversionsTable,
};

const testDataGenerator = new TestDataGenerator();

describe('`GAConversion` repo', () => {
  describe('`putGAConversion`', () => {
    it('adds the gaConversion instance to the gaConversion DynamoDB table', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      await putGAConversion(testGAConversion);

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelGAConversion, ['pk', 'sk']), {
        ...lowLevelGAConversion,
      });
    });
  });

  describe('`fetchGAConversionInterface`', () => {
    it('returns the high level interface of the gaConversion matching the gaConversion pdid', async () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all']});
      const testGAConversion = new GAConversion(testGAConversionInterface);
  
      await putGAConversion(testGAConversion);
  
      await waitForGSIUpdates();
  
      const receivedInterface = await fetchGAConversionInterface(testGAConversion);
  
      expect(receivedInterface).toStrictEqual(testGAConversionInterface);
    });
    
    it('returns null if no results are found', async () => {
      const testInterface = testDataGenerator.getGAConversionInterface();
      const testConversion = new GAConversion(testInterface)
      const receivedInterface = await fetchGAConversionInterface(testConversion);
  
      expect(receivedInterface).toBeNull();
    });
  });

  describe('`fetchGAConversionInterfaceByGclid`', () => {
    it('returns the high level interface of the gaConversions matching the gclid', async () => {
      const testGclid = uuid()
      const testGclidOverride = { overrides: { gclid: testGclid}}
      const testGAConversionInterface1 = testDataGenerator.getGAConversionInterface(testGclidOverride);
      const testGAConversionInterface2 = testDataGenerator.getGAConversionInterface(testGclidOverride);
      const testGAConversionInterface3 = testDataGenerator.getGAConversionInterface(testGclidOverride);
  
      await putGAConversion(testGAConversionInterface1);
      await putGAConversion(testGAConversionInterface2);
      await putGAConversion(testGAConversionInterface3);
  
      await waitForGSIUpdates();
  
      const receivedInterfaces = await fetchGAConversionInterfacesByGclid(testGclid);
  
      expect(receivedInterfaces).toHaveLength(3);
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface1)
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface2)
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface3)
    });
    
    it('returns null if no results are found', async () => {
      const receivedInterface = await fetchGAConversionInterfacesByGclid('nonexistentgclid');
  
      expect(receivedInterface).toBeNull();
    });
  });

  describe('`fetchGAConversionInterfaceByProject`', () => {
    it('returns the high level interface of the gaConversions belonging to the project', async () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
      const testGclid = testProject.gclid
      const overrides = {
        gclid: testGclid,
        projectPdid: testProject.pdid,
        conversionType: GAConversionType.QUOTE_REQUESTED
      }
      const testGAConversionInterface1 = testDataGenerator.getGAConversionInterface({overrides});
      const testGAConversionInterface2 = testDataGenerator.getGAConversionInterface({overrides: {...overrides, conversionType: GAConversionType.QUOTE_RECEIVED}, fieldsToGenerate: ['conversionValue']});
      const testGAConversionInterface3 = testDataGenerator.getGAConversionInterface({overrides: {...overrides, conversionType: GAConversionType.BOOKED}, fieldsToGenerate: ['conversionValue']});
  
      await putGAConversion(testGAConversionInterface1);
      await putGAConversion(testGAConversionInterface2);
      await putGAConversion(testGAConversionInterface3);
  
      await waitForGSIUpdates();
  
      const receivedInterfaces = await fetchGAConversionInterfacesByProject(testProject);
  
      expect(receivedInterfaces).toHaveLength(3);
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface1)
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface2)
      expect(receivedInterfaces).toContainEqual(testGAConversionInterface3)
    });
    
    it('returns null if no results are found', async () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
      const receivedInterface = await fetchGAConversionInterfacesByProject(testProject);
  
      expect(receivedInterface).toBeNull();
    });
  });

  describe('`hardDeleteGAConversion`', () => {
    it('removes all items associated with the gaConversion', async () => {
      const testGAConversion = testDataGenerator.getGAConversion();
      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      await putGAConversion(testGAConversion);

      await waitForGSIUpdates();

      await hardDeleteGAConversion(testGAConversion);

      await waitForGSIUpdates();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelGAConversion, ['pk', 'sk']));
    });
  });

  describe('`updateGAConversion`', () => {
    it('updates the gaConversion lastUpdatedField, field referenced by lastUpdatedField, and changeSource', async () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['conversionValue'] });
      const testGAConversion = new GAConversion(testGAConversionInterface);
      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      await putGAConversion(testGAConversionInterface)

      testGAConversionInterface.lastUpdatedField = MutableGAConversionProperties.conversionValue;
      testGAConversionInterface.conversionValue = 1001;

      await updateGAConversion(testGAConversionInterface);

      const keys = _.pick(lowLevelGAConversion, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelGAConversion, lastUpdatedField: MutableGAConversionProperties.conversionValue, conversionValue: 1001 });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});