import { AWS_REGION, ddb } from '@svc/config';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import _ from 'lodash';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import {
  buildLowLevelProjectDetails,
  putProject,
  hardDeleteProject,
  fetchProjectInterface,
  updateProject,
  removeProjectProperty,
  fetchProjectInterfaceByPdid,
} from '@svc/lib/ddbRepos/projects-repo';
import Project from '@svc/lib/models/Project';
import { MutableProjectProperties } from '@improovy/types';
import { waitForDDBDeletes, waitForGSIUpdates } from '@tests/utils/wait';

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.projectsTable,
};

const testDataGenerator = new TestDataGenerator();

describe('`Projects` repo', () => {
  describe('`putProject`', () => {
    it('adds the project instance to the projects DynamoDB table', async () => {
      const testProject = testDataGenerator.getProject();
      const lowLevelProject = buildLowLevelProjectDetails(testProject);

      await putProject(testProject);

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelProject, ['pk', 'sk']), {
        ...lowLevelProject,
      });
    });
  });

  describe('`fetchProjectInterface`', () => {
    it('returns a highlevel project interface of the requested project', async () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = new Project(testProjectInterface);

      await testProject.put();

      await waitForGSIUpdates();

      const receivedInterface = await fetchProjectInterface(testProject.uuid);

      expect(receivedInterface).toEqual(expect.objectContaining(testProjectInterface));
      expect(testProjectInterface).toEqual(expect.objectContaining(receivedInterface));
    });
  });

  describe('`fetchProjectInterfaceByPdid`', () => {
    it('returns a highlevel project interface of the requested project', async () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = new Project(testProjectInterface);

      await testProject.put();
      
      await waitForGSIUpdates();

      const receivedInterface = await fetchProjectInterfaceByPdid(testProject.pdid!);

      expect(receivedInterface).toEqual(expect.objectContaining(testProjectInterface));
      expect(testProjectInterface).toEqual(expect.objectContaining(receivedInterface));
    });
  });

  describe('`hardDeleteProject`', () => {
    it('removes all items associated with the project', async () => {
      const testProject = testDataGenerator.getProject();
      const lowLevelProject = buildLowLevelProjectDetails(testProject);

      await putProject(testProject);

      await hardDeleteProject(testProject);

      await waitForDDBDeletes();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelProject, ['pk', 'sk']));

      const fetchedItems = await queryByPK(ddb.projectsTable, testProject.uuid);

      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`updateProject`', () => {
    it('updates the project lastUpdatedField and field referenced by lastUpdatedField', async () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = new Project(testProjectInterface);
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      testProjectInterface.lastUpdatedField = MutableProjectProperties.gclid;
      testProjectInterface.gclid = 'test_gclid';

      // @ts-expect-error
      await updateProject(testProjectInterface);

      const keys = _.pick(details, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'gclid', gclid: 'test_gclid' });
    });
  });

  describe('`removeProjectProperty`', () => {
    it('updates the project lastUpdatedField and removes the property referenced by lastUpdatedField', async () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = new Project(testProjectInterface);
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      testProjectInterface.lastUpdatedField = MutableProjectProperties.gclid;
      
      // @ts-expect-error
      await removeProjectProperty(testProjectInterface);
      
      const keys = _.pick(details, ['pk', 'sk']);
      await expect({ region: AWS_REGION, table: ddb.projectsTable }).not.toHaveItem(keys, { ...details });
      
      delete details.gclid;
      details.lastUpdatedField = MutableProjectProperties.gclid;
      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});