import Client from '@svc/lib/models/Client';
import { generateValidatedUuid, isValidUuid } from '@svc/lib/models/modelUtils';
import {
  fetchClientInterface,
  fetchClientProjectInterfaces,
  putClient,
  removeClientProperty,
  updateClient,
} from '@svc/lib/ddbRepos/clients-repo';
import { ClientIncomingContactMethod, ClientInterface, ContactType, MutableClientProperties } from '@improovy/types';
import { ChangeSource } from '@improovy/types';
import { PipedrivePersonFieldKeys } from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/lib/ddbRepos/clients-repo', () => {
  return {
    putClient: jest.fn(() => {}),
    fetchClientInterface: jest.fn(),
    updateClient: jest.fn(),
    removeClientProperty: jest.fn(),
    fetchClientProjectInterfaces: jest.fn(() => []),
  };
});


describe('`Client` model', () => {
  it('throws an error if the supplied UUID starts with `PDID`', () => {
    expect(() => testDataGenerator.getClient({ overrides: { uuid: 'PDID1234' } })).toThrow();
  });

  it('correctly maps the fields when created with a ClientInterface', () => {
    const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
    // if this assertion fails, it's likely because the clientInterface was changed.
    // Update the expected number of properties and add tests to make sure they are
    // mapped correctly
    expect(Object.keys(testClientInterface)).toHaveLength(11);

    const testClient = new Client(testClientInterface);

    for (let key in testClientInterface) {
      expect(testClient[key as keyof Client]).toStrictEqual(testClientInterface[key as keyof ClientInterface]);
    }
  });

  it('correctly assigns default values when values are missing', () => {
    const testClientInterface = testDataGenerator.getClient({ overrides: { uuid: undefined } });

    const testClient = new Client(testClientInterface);


    expect(testClient.isDeleted).toBe(false);
    expect(isValidUuid(testClient.uuid)).toBe(true);
    for (let contact of testClient.contactMethods) {
      expect(contact.isDeleted).toBe(false);
    }
  });

  it('does not assign default values to properties that should not have default values', () => {
    const testClient = testDataGenerator.getClient({fieldsToExclude: ['pdid', 'isDeleted']});

    expect(testClient.gclid).toBeUndefined();
    expect(testClient.pdid).toBeUndefined();
    expect(testClient.lastUpdatedField).toBeUndefined();
  });

  it('removes duplicate contact methods', () => {
    const testClientInterface = testDataGenerator.getClientInterface({overrides: {
      contactMethods: [
        {
          contactType: ContactType.PHONE,
          contactAddress: '+12625551234'
        },
        {
          contactType: ContactType.PHONE,
          contactAddress: '2625551234'
        },
        {
          contactType: ContactType.PHONE,
          contactAddress: '(262) 555-1234'
        },
      ]
    }});

    const testClient = new Client(testClientInterface);

    expect(testClient.contactMethods).toHaveLength(1)
  })

  describe('`Client#put`', () => {
    it('passes itself to the lowlevel function putClient', () => {
      const testClient = testDataGenerator.getClient();
      expect(putClient).not.toHaveBeenCalled();
      testClient.put();
      expect(putClient).toHaveBeenCalledTimes(1);
      expect(putClient).toHaveBeenCalledWith(testClient);
    });
  });

  describe('`Client::fetch`', () => {
    it('returns the client returned by the lowlevel function fetchClient', async () => {
      const uuid = generateValidatedUuid();
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'], overrides: { uuid } });
      
      // @ts-expect-error
      fetchClientInterface.mockReturnValueOnce(testClientInterface);

      expect(fetchClientInterface).not.toHaveBeenCalled();

      const testClient = await Client.fetch(uuid);
      expect(fetchClientInterface).toHaveBeenCalledTimes(1);
      expect(fetchClientInterface).toHaveBeenCalledWith(uuid);
      expect(testClient!.uuid).toBe(uuid);
      expect(testClient).toBeInstanceOf(Client);
    });
  });

  describe('`Client::differingFields`', () => {
    it('returns the fields that are different between two clients', async () => {
      const testClientInterface1 = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClientInterface2 = { ...testClientInterface1, gclid: 'testing2', isDeleted: true };
      const testClient1 = new Client(testClientInterface1);
      const testClient2 = new Client(testClientInterface2);

      const diff = Client.differingFields(testClient1, testClient2);

      expect(diff).toHaveLength(2);
      expect(diff).toEqual(expect.arrayContaining(['gclid', 'isDeleted']));
    });

    it('does not return `lastUpdatedField` even if it is different', async () => {
      const testClientInterface1 = testDataGenerator.getClientInterface({ overrides: { lastUpdatedField: MutableClientProperties.gclid} });
      const testClientInterface2 = { ...testClientInterface1, lastUpdatedField: MutableClientProperties.incomingContactMethod };
      const testClient1 = new Client(testClientInterface1);
      const testClient2 = new Client(testClientInterface2);

      const diff = Client.differingFields(testClient1, testClient2);

      expect(diff).toHaveLength(0);
    });

    it('does not return `changeSource` even if it is different', async () => {
      const testClientInterface1 = testDataGenerator.getClientInterface();
      const testClientInterface2 = { ...testClientInterface1, changeSource: ChangeSource.aws };
      const testClient1 = new Client(testClientInterface1);
      const testClient2 = new Client(testClientInterface2);

      const diff = Client.differingFields(testClient1, testClient2);

      expect(diff).toHaveLength(0);
    });
  });

  describe('`Client#updateGclid`', () => {
    it('sets the gclid to the parameter', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.updateGclid('test_gclid');
      expect(testClient.gclid).toBe('test_gclid');
    });
    it('sets the lastUpdatedField to `gclid`', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.updateGclid('test_gclid');
      expect(testClient.lastUpdatedField).toBe('gclid');
    });
    it('invokes the updateClient function', async () => {
      jest.clearAllMocks();
      const testClient = testDataGenerator.getClient();
      expect(updateClient).not.toHaveBeenCalled();
      await testClient.updateGclid('test_gclid');
      expect(updateClient).toHaveBeenCalled();
    });
  });

  describe('`Client#removeGclid`', () => {
    it('removes the gclid from the client instance', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.removeGclid();
      expect(testClient.gclid).toBeFalsy();
    });
    it('sets the lastUpdatedField to `gclid`', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.removeGclid();
      expect(testClient.lastUpdatedField).toBe('gclid');
    });
    it('invokes the removeClientProperty function', async () => {
      const testClient = testDataGenerator.getClient();
      jest.clearAllMocks();
      
      expect(removeClientProperty).not.toHaveBeenCalled();
      await testClient.removeGclid();
      expect(removeClientProperty).toHaveBeenCalled();
    });
  });

  describe('`Client#softDelete`', () => {
    it('sets the isDeleted feild to true', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.softDelete();
      expect(testClient.isDeleted).toBe(true);
    });
    it('sets the lastUpdatedField to `isDeleted`', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.softDelete();
      expect(testClient.lastUpdatedField).toBe('isDeleted');
    });
    it('invokes the updateClient function', async () => {
      jest.clearAllMocks();
      const testClient = testDataGenerator.getClient();
      expect(updateClient).not.toHaveBeenCalled();
      await testClient.softDelete();
      expect(updateClient).toHaveBeenCalled();
    });
  });

  describe('`Client#fetchProjects`', () => {
    it('only calls fetchClientProjectInterfaces once despite multiple calls', async () => {
      jest.clearAllMocks();
      const testClient = testDataGenerator.getClient();

      expect(fetchClientProjectInterfaces).not.toHaveBeenCalled();
      await testClient.fetchProjects();
      
      expect(fetchClientProjectInterfaces).toHaveBeenCalledTimes(1);
      await testClient.fetchProjects();
      expect(fetchClientProjectInterfaces).toHaveBeenCalledTimes(1);
    });
  });

  describe('`Client::fromPipedrive`', () => {
    it('correctly maps in a person from the pipeDrive object', () => {
      const pdid = 1234;
      const gclid = 'test-gclid';
      const date = '2022-04-20 22:27:36';
      const createdAt = 1650493656000; // convert "date" to milliseconds since epoch
      const incomingContactMethod = ClientIncomingContactMethod.PHONE;
      const callVal = '238';
      const unformattedPhoneNumber = '+1 (608).555-1234';
      const formattedPhoneNumber = '6085551234';
      const email = 'test@improovy.com';
      
      const person = testDataGenerator.getPipedrivePerson({
        [PipedrivePersonFieldKeys.pdid]: pdid,
        [PipedrivePersonFieldKeys.gclid]: gclid,
        [PipedrivePersonFieldKeys.createdAt]: date,
        [PipedrivePersonFieldKeys.incomingContactMethod]: callVal,
        [PipedrivePersonFieldKeys.phones]: [
          {
            value: unformattedPhoneNumber,
          },
          {
            value: '',
          },
        ],
        [PipedrivePersonFieldKeys.emails]: [
          {
            value: email,
          },
          {
            value: '',
          },
        ],
      });

      const client = Client.fromPipedrive(person);

      expect(client.pdid).toBe(`${pdid}`); // pdid is converted to a string
      expect(client.gclid).toBe(gclid);
      expect(client.uuid).toBe(person[PipedrivePersonFieldKeys.uuid]);
      expect(client.createdAt).toBe(createdAt);
      expect(client.incomingContactMethod).toBe(incomingContactMethod);
      expect(client.contactMethods).toHaveLength(2);
      expect(client.contactMethods[0].contactType).toBe(ClientIncomingContactMethod.PHONE);
      expect(client.contactMethods[0].contactAddress).toBe(formattedPhoneNumber);
      expect(client.contactMethods[1].contactType).toBe(ClientIncomingContactMethod.EMAIL);
      expect(client.contactMethods[1].contactAddress).toBe(email);
    });
  });
});