import GAConversion from '@svc/lib/models/GAConversion';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { GAConversionInterface, GAConversionType } from '@improovy/types';
import { ChangeSource } from '@improovy/types';
import { ProjectSalesStage } from '@improovy/types';

const testDataGenerator = new TestDataGenerator(false);

describe('`GAConversion` model', () => {
  it('correctly maps the fields when created with a GAConversionInterface', () => {
    const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });
    // if this assertion fails, it's likely because the GAConversionInterface was changed.
    // Update the expected number of properties and add tests to make sure they are
    // mapped correctly
    expect(Object.keys(testGAConversionInterface)).toHaveLength(8);

    const testGAConversion = new GAConversion(testGAConversionInterface);

    for (let key in testGAConversionInterface) {
      expect(testGAConversion[key as keyof GAConversion]).toBe(testGAConversionInterface[key as keyof GAConversionInterface]);
    }
  });

  describe('`GAConversion::fromProject`', () => {
    it('correctly maps in a gaConversion from a Project that is not in an estimated/booked stage', () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']})
      const receivedConversion = GAConversion.fromProject(testProject)
      
      expect(receivedConversion.projectPdid).toBe(testProject.pdid)
      expect(receivedConversion.changeSource).toBe(ChangeSource.aws)
      expect(receivedConversion.gclid).toBe(testProject.gclid)
      expect(receivedConversion.conversionType).toBe(GAConversionType.QUOTE_REQUESTED)
      expect(receivedConversion.conversionValue).toBeUndefined()
      const now = Date.now()
      expect(receivedConversion.conversionTime).toBeLessThanOrEqual(now)
      expect(receivedConversion.conversionTime).toBeGreaterThanOrEqual(now - 5000)
    })

    it('correctly maps in a gaConversion from a Project that is in an estimated stage', () => {
      const testProject = testDataGenerator.getProject({overrides: {salesStage: ProjectSalesStage.ESTIMATED}, fieldsToGenerate: ['gclid', 'clientCost']})
      const receivedConversion = GAConversion.fromProject(testProject)
      
      expect(receivedConversion.projectPdid).toBe(testProject.pdid)
      expect(receivedConversion.changeSource).toBe(ChangeSource.aws)
      expect(receivedConversion.gclid).toBe(testProject.gclid)
      expect(receivedConversion.conversionType).toBe(GAConversionType.QUOTE_RECEIVED)
      expect(receivedConversion.conversionValue).toBe(testProject.clientCost)
      const now = Date.now()
      expect(receivedConversion.conversionTime).toBeLessThanOrEqual(now)
      expect(receivedConversion.conversionTime).toBeGreaterThanOrEqual(now - 5000)
    })

    it('correctly maps in a gaConversion from a Project that is in a booked stage', () => {
      const testProject = testDataGenerator.getProject({overrides: {salesStage: ProjectSalesStage.BOOKED}, fieldsToGenerate: ['gclid', 'clientCost']})
      const receivedConversion = GAConversion.fromProject(testProject)
      
      expect(receivedConversion.projectPdid).toBe(testProject.pdid)
      expect(receivedConversion.changeSource).toBe(ChangeSource.aws)
      expect(receivedConversion.gclid).toBe(testProject.gclid)
      expect(receivedConversion.conversionType).toBe(GAConversionType.BOOKED)
      expect(receivedConversion.conversionValue).toBe(testProject.clientCost)
      const now = Date.now()
      expect(receivedConversion.conversionTime).toBeLessThanOrEqual(now)
      expect(receivedConversion.conversionTime).toBeGreaterThanOrEqual(now - 5000)
    })

    it('throws an error if the project does not have a gclid', () => {
      const testProject = testDataGenerator.getProject();
      expect(() => GAConversion.fromProject(testProject)).toThrowError();
    })

    it('throws an error if the project does not have a pdid', () => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['pdid'], fieldsToGenerate: ['gclid']});
      expect(() => GAConversion.fromProject(testProject)).toThrowError();
    })
  })
});