import Project from '@svc/lib/models/Project';
import { generateValidatedUuid, isValidUuid } from '@svc/lib/models/modelUtils';
import { fetchProjectInterface, removeProjectProperty, updateProject } from '@svc/lib/ddbRepos/projects-repo';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { MutableProjectProperties, ProjectInterface } from '@improovy/types';
import { PipedriveDealFieldKeys } from '@improovy/types';
import { ChangeSource } from '@improovy/types';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/lib/ddbRepos/projects-repo', () => {
  const original = jest.requireActual('@svc/lib/ddbRepos/projects-repo');
  return {
    __esModule: true,
    ...original,
    putProject: jest.fn(() => {}),
    fetchProjectInterface: jest.fn(),
    updateProject: jest.fn(),
    removeProjectProperty: jest.fn(),
  };
});


describe('`Project` model', () => {
  // test needed in case using uuid as a sort key
  it('throws an error if the supplied UUID starts with `PDID`', () => {
    expect(() => testDataGenerator.getProject({ overrides: { uuid: 'PDID1234' } })).toThrow();
  });

  it('correctly maps the fields when created with a ProjectInterface', () => {
    const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
    // if this assertion fails, it's likely because the projectInterface was changed.
    // Update the expected number of properties and add tests to make sure they are
    // mapped correctly
    expect(Object.keys(testProjectInterface)).toHaveLength(13);

    const testProject = new Project(testProjectInterface);

    for (let key in testProjectInterface) {
      expect(testProject[key as keyof Project]).toBe(testProjectInterface[key as keyof ProjectInterface]);
    }
  });

  it('correctly assigns default values when values are missing', () => {
    const testProject = testDataGenerator.getProject({fieldsToExclude: ['uuid']});

    expect(testProject.isDeleted).toBe(false);
    expect(isValidUuid(testProject.uuid)).toBe(true);
  });

  it('does not assign default values to properties that should not have default values', () => {
    const testProject = testDataGenerator.getProject({fieldsToExclude: ['pdid', 'clientUuid']});

    expect(testProject.gclid).toBeUndefined();
    expect(testProject.gaTransactionID).toBeUndefined();
    expect(testProject.clientCost).toBeUndefined();
    expect(testProject.pdid).toBeUndefined();
    expect(testProject.lastUpdatedField).toBeUndefined();
    expect(testProject.clientUuid).toBeUndefined();
  });

  describe('`Project::differingFields`', () => {
    it('returns the fields that are different between two projects', async () => {
      const testProjectInterface1 = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProjectInterface2 = { ...testProjectInterface1, gclid: 'testing2', isDeleted: true };
      const testProject1 = new Project(testProjectInterface1);
      const testProject2 = new Project(testProjectInterface2);

      const diff = Project.differingFields(testProject1, testProject2);

      expect(diff).toHaveLength(2);
      expect(diff).toEqual(expect.arrayContaining(['gclid', 'isDeleted']));
    });

    it('does not return `lastUpdatedField` even if it is different', async () => {
      const testProjectInterface1 = testDataGenerator.getProjectInterface({ overrides: { lastUpdatedField: MutableProjectProperties.clientCost} });
      const testProjectInterface2 = { ...testProjectInterface1, lastUpdatedField: MutableProjectProperties.gaTransactionID };
      const testProject1 = new Project(testProjectInterface1);
      const testProject2 = new Project(testProjectInterface2);

      const diff = Project.differingFields(testProject1, testProject2);

      expect(diff).toHaveLength(0);
    });

    it('does not return `changeSource` even if it is different', async () => {
      const testProjectInterface1 = testDataGenerator.getProjectInterface();
      const testProjectInterface2 = { ...testProjectInterface1, changeSource: ChangeSource.aws };
      const testProject1 = new Project(testProjectInterface1);
      const testProject2 = new Project(testProjectInterface2);

      const diff = Project.differingFields(testProject1, testProject2);

      expect(diff).toHaveLength(0);
    });
  });

  describe('`Project::fetch`', () => {
    it('returns the project returned by the lowlevel function fetchProject', async () => {
      const uuid = generateValidatedUuid();
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'], overrides: { uuid } });
      
      // @ts-expect-error
      fetchProjectInterface.mockReturnValueOnce(testProjectInterface);

      expect(fetchProjectInterface).not.toHaveBeenCalled();

      const testProject = await Project.fetch(uuid);
      expect(fetchProjectInterface).toHaveBeenCalledTimes(1);
      expect(fetchProjectInterface).toHaveBeenCalledWith(uuid);
      expect(testProject.uuid).toBe(uuid);
      expect(testProject).toBeInstanceOf(Project);
    });
  });

  describe('`Project#updateGclid`', () => {
    it('sets the gclid to the parameter', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.updateGclid('test_gclid');
      expect(testProject.gclid).toBe('test_gclid');
    });

    it('sets the lastUpdatedField to `gclid`', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.updateGclid('test_gclid');
      expect(testProject.lastUpdatedField).toBe('gclid');
    });

    it('invokes the updateProject function', async () => {
      jest.clearAllMocks();
      const testProject = testDataGenerator.getProject();
      expect(updateProject).not.toHaveBeenCalled();
      await testProject.updateGclid('test_gclid');
      expect(updateProject).toHaveBeenCalled();
    });

    it('throws an error if the project already has a gclid', async () => {
      const testProject = testDataGenerator.getProject({fieldsToGenerate: ['gclid']});
      await expect(testProject.updateGclid('test_gclid')).rejects.toEqual(new Error('project already has a gclid'));
    });
  });

  describe('`Project#removeGclid`', () => {
    it('removes the gclid from the project instance', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.removeGclid();
      expect(testProject.gclid).toBeFalsy();
    });
    it('sets the lastUpdatedField to `gclid`', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.removeGclid();
      expect(testProject.lastUpdatedField).toBe('gclid');
    });
    it('invokes the removeProjectProperty function', async () => {
      const testProject = testDataGenerator.getProject();
      jest.clearAllMocks();
      
      expect(removeProjectProperty).not.toHaveBeenCalled();
      await testProject.removeGclid();
      expect(removeProjectProperty).toHaveBeenCalled();
    });
  });

  describe('`Project#softDelete`', () => {
    it('sets the isDeleted feild to true', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.softDelete();
      expect(testProject.isDeleted).toBe(true);
    });
    it('sets the lastUpdatedField to `isDeleted`', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.softDelete();
      expect(testProject.lastUpdatedField).toBe('isDeleted');
    });
    it('invokes the updateProject function', async () => {
      jest.clearAllMocks();
      const testProject = testDataGenerator.getProject();
      expect(updateProject).not.toHaveBeenCalled();
      await testProject.softDelete();
      expect(updateProject).toHaveBeenCalled();
    });
  });

  describe('`Project#dealIsProject`', () => {
    it('returns true if the pipeline id matches the pipeline id for projects', async () => {
      const testDeal = testDataGenerator.getPipedriveDeal()
      
      expect(Project.dealIsProject(testDeal)).toBe(true);
    });
    it('returns false if the pipeline id does not match the pipeline id for porjects', async () => {
      const testDeal = testDataGenerator.getPipedriveDeal({[PipedriveDealFieldKeys.pipeline]: 29});
      
      expect(Project.dealIsProject(testDeal)).toBe(false);
    });
  });
});