import GAConversion from '@svc/lib/models/GAConversion';
import {
  assembleGAConversion,
  assembleGAConversionInterfaceFromDDBItem,
  assembleLowLevelGAConversionInterface,
  buildGAConversionSK,
  buildLowLevelGAConversion
} from '@svc/lib/ddbRepos/gaConversions-repo';
import { LowLevelGAConversion, GAConversionInterface } from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/config', () => {
  const original = jest.requireActual('@svc/config');

  return {
    __esModule: true,
    ddb: {
      ...original.ddb,
      documentClient: {
        send: jest.fn(),
      },
    },
  };
});

jest.mock('@svc/lib/ddbRepos/repo-utils', () => {
  return {
    __esModule: true,
    queryByPK: jest.fn().mockImplementation(() => ([])),
  };
});

describe('`GAConversion` repo', () => {
  beforeEach(() => jest.clearAllMocks());

  describe('`buildGAConversionSK`', () => {
    it('builds the sk from a conversion', () => {
      const testConversion = testDataGenerator.getGAConversionInterface();
      const expectedSK = `${testConversion.projectPdid}#${testConversion.conversionType}`
      const receivedSK = buildGAConversionSK(testConversion);

      expect(receivedSK).toBe(expectedSK)
    })
  })

  describe('`buildLowLevelGAConversion`', () => {
    it('has 8 keys when supplied with no optional fields', () => {
      const testGAConversion = testDataGenerator.getGAConversion({fieldsToExclude: ['isTest']});
      const lowLevel = buildLowLevelGAConversion(testGAConversion);
      expect(Object.keys(lowLevel)).toHaveLength(8);
    });
    
    it('has 11 keys when supplied with all optional fields', () => {
      const testGAConversion = testDataGenerator.getGAConversion({ fieldsToGenerate: ['all'] });
      const lowLevel = buildLowLevelGAConversion(testGAConversion);
      expect(Object.keys(lowLevel)).toHaveLength(11);
    });
    
    it('correctly inherits the properties from the high-level interface', () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });
      const testGAConversion = testDataGenerator.getGAConversion({ overrides: testGAConversionInterface });
      const details = buildLowLevelGAConversion(testGAConversion);

      for (let key in testGAConversionInterface) {
        expect(details[key as keyof LowLevelGAConversion]).toBe(testGAConversionInterface[key as keyof GAConversionInterface]);
      }
    });

    it('correctly assigns the new properties specific to the low-level interface', () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });
      const testGAConversion = testDataGenerator.getGAConversion({ overrides: testGAConversionInterface });
      const lowLevel = buildLowLevelGAConversion(testGAConversion);

      expect(lowLevel.pk).toBe(testGAConversion.gclid);
      expect(lowLevel.sk).toBe(buildGAConversionSK(testGAConversion));
      expect(lowLevel.type).toBe('GAConversion');
    });
  });

  describe('`assembleGAConversion`', () => {
    it('creates a high level gaConversion profile from a low level one', () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });
      const testGAConversion = new GAConversion(testGAConversionInterface);

      const lowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);
      const receivedResult = assembleGAConversion(lowLevelGAConversion);

      expect(receivedResult).toEqual(expect.objectContaining(testGAConversionInterface));
      expect(testGAConversionInterface).toEqual(expect.objectContaining(receivedResult));
    });
  });

  describe('`assembleLowLevelGAConversionInterface`', () => {
    it('creates a LowLevelGAConversion object from its respective DDB item', () => {
      const testGAConversion = testDataGenerator.getGAConversion({ fieldsToGenerate: ['all'] });
      const testLowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      const resultLowLevelGAConversion = assembleLowLevelGAConversionInterface(testLowLevelGAConversion);

      expect(resultLowLevelGAConversion).toEqual(expect.objectContaining(testLowLevelGAConversion));
      expect(testLowLevelGAConversion).toEqual(expect.objectContaining(resultLowLevelGAConversion));
    });
  });

  describe('`assembleGAConversionInterfaceFromDDBItems`', () => {
    it('creates a LowLevelGAConversion object from its respective DDB item', () => {
      const testGAConversionInterface = testDataGenerator.getGAConversionInterface({ fieldsToGenerate: ['all'] });

      const testGAConversion = new GAConversion(testGAConversionInterface)

      const testLowLevelGAConversion = buildLowLevelGAConversion(testGAConversion);

      const resultGAConversion = assembleGAConversionInterfaceFromDDBItem(testLowLevelGAConversion as { [key: string] : any });

      expect(resultGAConversion).toEqual(expect.objectContaining(testGAConversionInterface));
      expect(testGAConversionInterface).toEqual(expect.objectContaining(resultGAConversion));
    });
  });
});