import Client from '@svc/lib/models/Client';
import Project from '@svc/lib/models/Project';
import { generateValidatedUuid } from '@svc/lib/models/modelUtils';
import {
  ClientIncomingContactMethod,
  ClientInterface,
  ContactType,
  GenerateTestClientInput,
  MutableClientProperties,
} from '@improovy/types';
import {
  GenerateTestProjectInput,
  MutableProjectProperties,
  ProjectInterface,
  ProjectSalesStage,
  ProjectStatus,
} from '@improovy/types';
import uuid from '@svc/lib/uuid';
import { CallRailFieldKeys, CallRailInterface, CallRailRequest, CallRailRequestBody, GenerateTestCallRailInput, MutableCallRailProperties } from '@improovy/types';
import CallRail from '@svc/lib/models/CallRail';
import {
  PipedrivePersonFieldKeys,
  PipedrivePerson,
  PipedriveDealFieldKeys,
  PipedrivePersonRequest,
  PipedriveDeal,
  PipedriveDealRequest,
  PipedriveDealStatus,
  GetTestPipedriveActivityInput,
  PipedriveEventActions,
  PipedriveEventEntities,
  PipedriveActivity,
  PipedriveActivityFieldKeys,
  PipedriveActivityRequest,
} from '@improovy/types';
import _ from 'lodash';
import { ActivityInterface, ActivityType, GenerateTestActivityInput, MutableActivityProperties } from '@improovy/types';
// import Activity from '@svc/lib/models/Activity';
import { ChangeSource } from '@improovy/types';
import getCallRailSignature from '@svc/lib/getCallRailSignature';
import { CALLRAIL_SIGNING_KEY } from '@svc/config';
import generateNewPhoneNumber from './generateNewPhoneNumber';
import { GAConversionInterface, GAConversionType, GenerateTestGAConversionInput, MutableGAConversionProperties } from '@improovy/types';
import GAConversion from '@svc/lib/models/GAConversion';
// import { buildGAConversionSK } from '@svc/lib/ddbRepos/gaConversions-repo';

const generateNewGclid = () => {
  return uuid();
};

export const generateCallSid = () => {
  const str = `${uuid()}`
  return str.replace(/-/g, '')
}

export const generateCallSidBody = (callSid = generateCallSid()) => {
  return `Call SId : ${callSid}`
}

const generateNewPdid = () => {
  return uuid();
};

const generateNewGATransactionID = () => {
  return uuid();
};

const generateNumericalPdid = () => {
  return Math.floor((Math.random() * 9000000000) + 1000000000);
};

export const generateTestEmail = () => {
  return `test_${uuid()}@improovy.com`;
};

export default class TestDataGenerator {
  #createdEntities: {hardDelete: Function, [key: string]: any}[];

  #isSaved: boolean;

  constructor(isSaved = true) {
    this.#isSaved = isSaved;
    this.#createdEntities = [];
  }

  #addCreatedEntity <Entity extends {hardDelete: Function, [key: string]: any}>(entity: Entity) {
    if (this.#isSaved) this.#createdEntities.push(entity);
  }

  async deleteAll(): Promise<unknown> {
    const promises = [];
    const seen = new Set();

    for (let entity of this.#createdEntities) {
      const entityID = entity?.uuid || entity.phoneNumber || entity.pdid // || buildGAConversionSK(entity);
      if (seen.has(entityID)) continue;

      promises.push(entity.hardDelete());
      seen.add(entityID);
    }
    return Promise.all(promises);
  }
  
  getClientInterface = ({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestClientInput = {}): ClientInterface => {
    const newOverrides: Partial<ClientInterface> = {};
  
    if (fieldsToGenerate[0]?.toUpperCase() === 'ALL') {
      fieldsToGenerate = [
        'gclid',
        'lastUpdatedField',
      ];
    }
  
    for (let field of fieldsToGenerate) {
  
      switch (field) {
        case 'gclid':
          newOverrides.gclid = generateNewGclid(); 
          break;
        case 'lastUpdatedField':
          newOverrides.lastUpdatedField = MutableClientProperties.gclid;
          break;
        default:
          throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
      }
    }
  
    Object.assign(newOverrides, overrides);

    const clientInterface: ClientInterface = {
      incomingContactMethod: ClientIncomingContactMethod.EMAIL,
      createdAt: Date.now(),
      isDeleted: false,
      isTest: true,
      contactMethods: [{ contactType: ContactType.EMAIL, contactAddress: generateTestEmail() }],
      uuid: generateValidatedUuid(),
      pipedriveName: uuid(),
      pdid: generateNewPdid(),
      changeSource: ChangeSource.pipedrive,
      ...newOverrides,
    };

    for (let field of fieldsToExclude) {
      switch (field) {
        case 'pdid':
          delete clientInterface.pdid;
          break;
        case 'isDeleted':
          delete clientInterface.isDeleted;
          break;
        case 'isTest':
          delete clientInterface.isTest;
          break;
        default:
          throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
      }
    }

    this.#addCreatedEntity(new Client(clientInterface));

    return clientInterface;
  };

  getClient = ({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestClientInput = {}): Client => {
    const clientInterface = this.getClientInterface({ fieldsToGenerate, overrides, fieldsToExclude });
    
    const client = new Client(clientInterface);

    this.#addCreatedEntity(client); // may double add clients but is needed for when
    // testing clients with undefined uuids

    return client;
  };

  getProjectInterface = ({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestProjectInput = {}): ProjectInterface => {
    const newOverrides: Partial<ProjectInterface> = {};
  
    if (fieldsToGenerate[0]?.toUpperCase() === 'ALL') {
      fieldsToGenerate = [
        'gclid',
        'gaTransactionID',
        'lastUpdatedField',
        'clientCost',
      ];
    }
  
    for (let field of fieldsToGenerate) {
  
      switch (field) {
        case 'gclid':
          newOverrides.gclid = generateNewGclid(); 
          break;
        case 'gaTransactionID':
          newOverrides.gaTransactionID = generateNewGATransactionID();
          break;
        case 'lastUpdatedField':
          newOverrides.lastUpdatedField = MutableProjectProperties.gclid;
          break;
        case 'clientCost':
          newOverrides.clientCost = 0;
          break;
        default:
          throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
      }

    }
  
    Object.assign(newOverrides, overrides);

    const projectInterface: ProjectInterface = {
      createdAt: Date.now(),
      pdid: generateNewPdid(),
      isDeleted: false,
      uuid: generateValidatedUuid(),
      clientUuid: generateValidatedUuid(),
      projectStatus: ProjectStatus.OPEN,
      salesStage: ProjectSalesStage.UNATTEMPTED,
      isTest: true,
      changeSource: ChangeSource.pipedrive,
      ...newOverrides,
    };

    for (let field of fieldsToExclude) {
  
      switch (field) {
        case 'pdid':
          delete projectInterface.pdid;
          break;
        case 'isDeleted':
          delete projectInterface.isDeleted;
          break;
        case 'clientCost':
          delete projectInterface.clientCost;
          break;
        case 'uuid':
          delete projectInterface.uuid;
          break;
        case 'clientUuid':
          delete projectInterface.clientUuid;
          break;
        case 'isTest':
          delete projectInterface.isTest;
          break;
        default:
          throw new Error('Field to exclude Not Found, check testDataGenerator file for supported fields and add support if needed');
      }
    }


    this.#addCreatedEntity(new Project(projectInterface));

    return projectInterface;
  };

  getProject({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestProjectInput = {}): Project {
    const projectInterface = this.getProjectInterface({ fieldsToGenerate, overrides, fieldsToExclude });
    
    const project = new Project(projectInterface);

    this.#addCreatedEntity(project); // may double add projects but is needed for when
    // testing projects with undefined uuids

    return project;
  }

  getCallRailInterface = ({ fieldsToGenerate = [], overrides = {} }: GenerateTestCallRailInput = {}): CallRailInterface => {
    const newOverrides: Partial<CallRailInterface> = {};
  
    if (fieldsToGenerate[0]?.toUpperCase() === 'ALL') {
      fieldsToGenerate = [
        'isAttributed',
        'lastUpdatedField'
      ];
    }
  
    for (let field of fieldsToGenerate) {
  
      switch (field) {
        case 'isAttributed':
          newOverrides.isAttributed = false;
          break;
        case 'lastUpdatedField':
          newOverrides.lastUpdatedField = MutableCallRailProperties.isAttributed;
          break;
        default:
          throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
      }

    }
  
    Object.assign(newOverrides, overrides);

    const callRailInterface = {
      createdAt: Date.now(),
      gclid: generateNewGclid(),
      phoneNumber: `${generateNewPhoneNumber()}`,
      changeSource: ChangeSource.callRail,
      isTest: true,
      ...newOverrides,
    };

    this.#addCreatedEntity(new CallRail(callRailInterface));

    return callRailInterface;
  };

  getCallRail({ fieldsToGenerate = [], overrides = {} }: GenerateTestCallRailInput = {}): CallRail {
    const callRailInterface = this.getCallRailInterface({ fieldsToGenerate, overrides });
    
    const callRail = new CallRail(callRailInterface);

    this.#addCreatedEntity(callRail); // may double add callRails but is needed for when
    // testing callRails with undefined uuids

    return callRail;
  }

  getCallRailRequestBody(overrides?: Partial<CallRailRequestBody>): CallRailRequestBody {
    const callRail = {
      [CallRailFieldKeys.phoneNumber]: `+1${generateNewPhoneNumber()}`,
      [CallRailFieldKeys.gclid]: generateNewGclid(),
      [CallRailFieldKeys.timestamp]: (new Date(Date.now())).toISOString(),
      isTest: true,
    };

    
    const req = { ...callRail, ...overrides}
    this.#addCreatedEntity(CallRail.fromCallRailRequest(req))
    return req;
  }

  getCallRailRequest(overrides?: Partial<CallRailRequestBody>): CallRailRequest {
    const reqBody = this.getCallRailRequestBody(overrides);
    const req: CallRailRequest = {
      body: reqBody,
      headers: {
        Signature: getCallRailSignature(reqBody, CALLRAIL_SIGNING_KEY)
      }
    }

    return req;
  }

  getPipedrivePerson(overrides?: Partial<PipedrivePerson>): PipedrivePerson {
    const person = {
      [PipedrivePersonFieldKeys.pdid]: generateNumericalPdid(),
      [PipedrivePersonFieldKeys.uuid]: generateValidatedUuid(),
      [PipedrivePersonFieldKeys.createdAt]: '2022-04-21 18:13:49',
      [PipedrivePersonFieldKeys.incomingContactMethod]: null,
      [PipedrivePersonFieldKeys.pipedriveName]: uuid(),
      isTest: true,
      [PipedrivePersonFieldKeys.phones]: [
        {
          value: generateNewPhoneNumber(),
        },
      ],
      [PipedrivePersonFieldKeys.emails]: [
        {
          value: generateTestEmail(),
        },
      ],
    };

    return { ...person, ...overrides };
  }

  // getPipedrivePersonRequestBody(overrides?: Partial<PipedrivePerson>): PipedrivePersonRequest {
  //   const person = this.getPipedrivePerson(overrides); 
    
  //   // ensures it's deleted if it's added to database
  //   this.getClient({ overrides: { uuid: person[PipedrivePersonFieldKeys.uuid] } });
    
  //   const personReq = {
  //     meta: {
  //       id: person[PipedrivePersonFieldKeys.pdid],
  //       action: PipedriveEventActions.updated,
  //       object: PipedriveEventEntities.person,
  //       timestamp_micro: Date.now(),
  //       isTest: true,
  //     },
  //     current: person,
  //   } as const;

  //   return personReq
  // }

  getPipedriveDeal(overrides?: Partial<PipedriveDeal>): PipedriveDeal {
    const deal = {
      [PipedriveDealFieldKeys.pdid]: generateNumericalPdid(),
      [PipedriveDealFieldKeys.uuid]: generateValidatedUuid(),
      [PipedriveDealFieldKeys.createdAt]: '2022-04-21 18:13:49',
      [PipedriveDealFieldKeys.personPdid]: generateNumericalPdid(),
      [PipedriveDealFieldKeys.pipeline]: 1,
      [PipedriveDealFieldKeys.stage]: 1,
      [PipedriveDealFieldKeys.projectStatus]: PipedriveDealStatus.open,
      isTest: true,
    };

    return { ...deal, ...overrides };
  }

  // getPipedriveDealRequestBody(overrides?: Partial<PipedriveDeal>): PipedriveDealRequest {
  //   const deal = this.getPipedriveDeal(overrides);

  //   // ensures it's deleted if it's added to database
  //   this.getProject({ overrides: { uuid: deal[PipedriveDealFieldKeys.uuid] } });

  //   const dealReq = {
  //     current: deal,
  //     meta: {
  //       id: deal[PipedriveDealFieldKeys.pdid],
  //       object: PipedriveEventEntities.deal,
  //       action: PipedriveEventActions.updated,
  //       timestamp_micro: Date.now(),
  //       isTest: true,
  //     },
  //   };

  //   return dealReq;
  // }

  // getPipedriveActivity({overrides = {}, fieldsToGenerate = [], bodyOptions}: GetTestPipedriveActivityInput = {}): PipedriveActivity {
  //   if (fieldsToGenerate.length) throw new Error("fields to generate for getPipedriveActivity not yet implemented");
    
  //   const activity: PipedriveActivity = {
  //     [PipedriveActivityFieldKeys.pdid]: generateNewPdid(),
  //     [PipedriveActivityFieldKeys.body]: uuid(),
  //     [PipedriveActivityFieldKeys.projectPdid]: generateNewPdid(),
  //     [PipedriveActivityFieldKeys.activityType]: ActivityType.INCOMING_CALL,
  //     isTest: true,
  //   }

  //   if (bodyOptions) {
  //     switch (bodyOptions.type) {
  //       case 'callSid':
  //         activity[PipedriveActivityFieldKeys.body] = generateCallSidBody(bodyOptions.override)
  //         break;
      
  //       default:
  //         throw new Error("bodyOption type not yet supported, add in if needed");
          
  //         break;
  //     }
  //   }

  //   return {...activity, ...overrides};
  // }

  // getActivityInterface = ({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestActivityInput = {}): ActivityInterface => {
  //   const newOverrides: Partial<ActivityInterface> = {};
  
  //   if (fieldsToGenerate[0]?.toUpperCase() === 'ALL') {
  //     fieldsToGenerate = [
  //       'callSid',
  //     ];
  //   }
  
  //   for (let field of fieldsToGenerate) {
  
  //     switch (field) {
  //       case 'callSid':
  //         newOverrides.callSid = overrides.callSid || generateCallSid();
  //         newOverrides.body = generateCallSidBody(newOverrides.callSid)
  //         break;
  //       default:
  //         throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
  //     }

  //   }
  
  //   Object.assign(newOverrides, overrides);

  //   const activityInterface: ActivityInterface = {
  //     pdid: generateNewPdid(),
  //     isTest: true,
  //     activityType: ActivityType.INCOMING_CALL,
  //     body: uuid(),
  //     projectPdid: generateNewPdid(),
  //     lastUpdatedField: MutableActivityProperties.body,
  //     isDeleted: false,
  //     changeSource: ChangeSource.pipedrive,
  //     ...newOverrides,
  //   };

  //   for (let field of fieldsToExclude) {
  
  //     switch (field) {
  //       case 'body':
  //         delete activityInterface.body;
  //         break;
  //       case 'isTest':
  //         delete activityInterface.isTest;
  //         break;
  //       case 'isDeleted':
  //         delete activityInterface.isDeleted;
  //         break;
  //       case 'lastUpdatedField':
  //         delete activityInterface.lastUpdatedField;
  //         break;
  //       default:
  //         throw new Error('Field to exclude Not Found, check testDataGenerator file for supported fields and add support if needed');
  //     }
  //   }


  //   this.#addCreatedEntity(new Activity(activityInterface));

  //   return activityInterface;
  // };

  // getActivity({ fieldsToGenerate = [], overrides = {}, fieldsToExclude = [] }: GenerateTestActivityInput = {}): Activity {
  //   const activityInterface = this.getActivityInterface({ fieldsToGenerate, overrides, fieldsToExclude });
    
  //   const activity = new Activity(activityInterface);

  //   this.#addCreatedEntity(activity);

  //   return activity;
  // }

  // getPipedriveActivityRequestBody({overrides = {}, fieldsToGenerate = [], bodyOptions}: GetTestPipedriveActivityInput = {}): PipedriveActivityRequest {
  //   const activity = this.getPipedriveActivity({overrides, fieldsToGenerate, bodyOptions})

  //   this.getActivity({overrides: {pdid: `${activity[PipedriveActivityFieldKeys.pdid]}`}})

  //   const activityReq = {
  //     current: activity,
  //     meta: {
  //       id: activity[PipedriveActivityFieldKeys.pdid],
  //       object: PipedriveEventEntities.activity,
  //       action: PipedriveEventActions.updated,
  //       timestamp_micro: Date.now(),
  //       isTest: true,
  //     },
  //   };

  //   return activityReq;
  // }

  getGAConversionInterface = ({overrides = {}, fieldsToExclude = [], fieldsToGenerate = [] }: GenerateTestGAConversionInput = {}): GAConversionInterface => {
    const newOverrides: Partial<GAConversionInterface> = {};
  
    if (fieldsToGenerate[0]?.toUpperCase() === 'ALL') {
      fieldsToGenerate = [
        'conversionValue',
        'lastUpdatedField',
      ];
    }
  
    for (let field of fieldsToGenerate) {
  
      switch (field) {
        case 'conversionValue':
          newOverrides.conversionValue = 1000;
          break;
        case 'lastUpdatedField':
          newOverrides.lastUpdatedField = MutableGAConversionProperties.conversionValue;
          break;
        default:
          throw new Error('Optional Field Not Found, check testDataGenerator file for supported fields and add support if needed');
      }

    }

    const gaConversionInterface: GAConversionInterface = {
      conversionType: GAConversionType.QUOTE_REQUESTED,
      projectPdid: generateNewPdid(),
      conversionTime: Date.now(),
      changeSource: ChangeSource.aws,
      isTest: true,
      gclid: generateNewGclid(),
      ...newOverrides,
      ...overrides,
    };

    for (let field of fieldsToExclude) {
  
      switch (field) {
        case 'isTest':
          delete gaConversionInterface.isTest;
          break;
        default:
          throw new Error('Field to exclude Not Found, check testDataGenerator file for supported fields and add support if needed');
      }
    }


    this.#addCreatedEntity(new GAConversion(gaConversionInterface));

    return gaConversionInterface;
  };

  getGAConversionInterfaceFromProject = <Project extends ProjectInterface>(project: Project, {overrides = {}, fieldsToExclude = [], fieldsToGenerate = [] }: GenerateTestGAConversionInput = {}): GAConversionInterface => {
    const newOverrides: Partial<GAConversionInterface> = {};
    
    switch (project.salesStage) {
      case ProjectSalesStage.UNATTEMPTED:
      case ProjectSalesStage.ATTEMPTED:
      case ProjectSalesStage.ESTABLISHED_CONTACT:
        newOverrides.conversionType = GAConversionType.QUOTE_REQUESTED
        break
      case ProjectSalesStage.ESTIMATED:
        newOverrides.conversionType = GAConversionType.QUOTE_RECEIVED
        break
      case ProjectSalesStage.BOOKED:
        newOverrides.conversionType = GAConversionType.BOOKED
        break
      default:
        throw new Error("unknown sales stage");
    }

    newOverrides.conversionValue = project.clientCost || 0;
    if (project.pdid) newOverrides.projectPdid = project.pdid;
    if (project.gclid) newOverrides.gclid = project.gclid;

    return this.getGAConversionInterface({overrides: {...newOverrides, ...overrides}, fieldsToExclude, fieldsToGenerate})
  };

  getGAConversion({overrides = {}, fieldsToExclude = [], fieldsToGenerate = [] }: GenerateTestGAConversionInput = {}): GAConversion {
    const gaConversionInterface = this.getGAConversionInterface({ overrides, fieldsToExclude, fieldsToGenerate });
    
    const gaConversion = new GAConversion(gaConversionInterface);

    this.#addCreatedEntity(gaConversion);

    return gaConversion;
  }
}