import log from '@dazn/lambda-powertools-logger';
import { OK } from '@svc/lib/handlerUtils';
import Project from '@svc/lib/models/Project';
import { addSeenPipedriveRequest, wasPipedriveRequestSeen } from '@svc/lib/ddbRepos/idempotency-repo';
import { APIGatewayProxyEvent, Context } from 'aws-lambda';
import { PersonsApi } from 'pipedrive';
import { PipedriveDealFieldKeys, PipedriveEventActions, PipedriveEventEntities, PipedrivePersonFieldKeys, PipedrivePersonRequest } from '@improovy/types';
import { lambdaConfig } from '@svc/config';
import { convertPipedriveTimeToMS } from '@svc/lib/models/modelUtils';
import wait from '@svc/lib/wait';

const pipedriveAPI = new PersonsApi();

export const handler = async (event: APIGatewayProxyEvent, ctx: Context) => {
  log.debug('Received event', { event });
  const dealReq = JSON.parse(event.body!);
  log.debug('received deal request', { dealReq: dealReq });

  const wasSeen = await wasPipedriveRequestSeen(dealReq, ctx.functionName);

  if (!wasSeen && dealReq.current.pipeline_id === 1) {
    let project = await Project.fromPipedrive(dealReq.current);
    
    if (!project) {
      const personResp = await pipedriveAPI.getPerson(dealReq.current[PipedriveDealFieldKeys.clientPdid]);
      const person = personResp.data;
      const personReq: PipedrivePersonRequest = {
        meta: {
          object: PipedriveEventEntities.person,
          action: PipedriveEventActions.added,
          id: person[PipedrivePersonFieldKeys.pdid],
          timestamp_micro: convertPipedriveTimeToMS(person[PipedrivePersonFieldKeys.createdAt]),
          isTest: dealReq.meta.isTest,
        },
        current: person,
      }; 
      
      await lambdaConfig.invokeSync('APIGWPipedriveAddPerson', personReq);
    }
    
    for (let tries = 0; tries < 3 && !project; tries++) {
      await wait(500)

      project = await Project.fromPipedrive(dealReq.current);
    }
  
    await project!.put();
    await addSeenPipedriveRequest(dealReq, ctx.functionName);
  }
  
  return OK;
};

const getPipedrivePerson = async (pdid:string | number, tries = 0) => {
  try {
    await pipedriveAPI.getPerson(pdid)
  } catch (error) {
    if (tries > 2) throw error;
    
    await getPipedrivePerson(pdid, tries++)
  }
}