import log from '@dazn/lambda-powertools-logger';
import Project from '@svc/lib/models/Project';
import { addSeenPipedriveRequest, wasPipedriveRequestSeen } from '@svc/lib/ddbRepos/idempotency-repo';
import { APIGatewayProxyEvent, Context } from 'aws-lambda';
import { lambdaConfig } from '@svc/config';

const OK = {
  statusCode: 200,
  body: JSON.stringify({}),
};

export const handler = async (event: APIGatewayProxyEvent, ctx: Context) => {
  log.debug('Received event', { event });
  const dealReq = JSON.parse(event.body!);
  log.debug('received deal request', { dealReq });

  const requestWasSeen = await wasPipedriveRequestSeen(dealReq, ctx.functionName);
  const deal = dealReq.current;
  const oldDeal = dealReq.previous;

  // 'if the deal was in the project pipeline but moved to a different pipeline, e.g. because it's a subcontractor'
  if (oldDeal && Project.dealIsProject(oldDeal) && !Project.dealIsProject(deal)) {
    const project = await Project.fromPipedrive(oldDeal);
    if (project) {
      await project.softDelete();
    }
  }
  
  if (!requestWasSeen && Project.dealIsProject(deal)) {

    const newProject = await Project.fromPipedrive(deal);
    let oldProject: Project | null | undefined;

    if (newProject) {
      oldProject = await Project.fetchByPdid(newProject.pdid!);
    }

    // newProject is null if the client was not yet added to the database, which also means
    // the project was not yet added. Adding the project will add the client
    if (!newProject || !oldProject) {
      const resp = await lambdaConfig.invokeSync('APIGWPipedriveAddDeal', event.body!);
      
      return {
        statusCode: resp.StatusCode!,
        body: JSON.stringify({}),
      };
    }

    const diffs = Project.differingFields(newProject, oldProject);

    await newProject.updateFields(diffs);
    
    await addSeenPipedriveRequest(dealReq, ctx.functionName);
  }


  
  return OK;
};