import { ddb } from '@svc/config';
import { TransactWriteCommand } from '@aws-sdk/lib-dynamodb';
import { queryByGSI, queryByPK, updateItem } from './repo-utils';
import { TransactWriteCommandInput } from '@aws-sdk/lib-dynamodb';
import _ from 'lodash';
import Activity from '../models/Activity';
import { LowLevelActivity, ActivityInterface, MutableActivityProperties } from '@improovy/types';

const TableName = ddb.activitiesTable;

export const buildActivityIndex1Keys = (callSid: string) => {
  return `Sid#${callSid}`
}

export const buildLowLevelActivity = (activity: Activity): LowLevelActivity => {
  const {
    activityType,
    body,
    callSid,
    projectPdid,
    pdid,
    isTest,
    isDeleted,
    lastUpdatedField,
    changeSource,
  } = activity;

  const lowLevelActivity: LowLevelActivity = {
    pk: pdid,
    sk: pdid,
    type: 'Activity',
    activityType,
    projectPdid,
    pdid,
    isDeleted,
    changeSource,
  };

  if (body) lowLevelActivity.body = body;
  if (callSid) {
    lowLevelActivity.callSid = callSid;
    const key = buildActivityIndex1Keys(callSid);
    lowLevelActivity.gsi1_pk = key
    lowLevelActivity.gsi1_sk = key
  }
  if (isTest) lowLevelActivity.isTest = isTest;
  if (lastUpdatedField) lowLevelActivity.lastUpdatedField = lastUpdatedField;
  
  return lowLevelActivity;
};

export const buildActivityPutTransaction = (activity: LowLevelActivity): TransactWriteCommandInput => {
  const transaction: TransactWriteCommandInput = {
    TransactItems: [
      { Put: {
        TableName,
        Item: activity,
      },
      },
    ],
  };

  return transaction;
};

export const assembleActivity = (activity: LowLevelActivity): ActivityInterface => {
  const assembledActivity = _.pick(activity, [
    'activityType',
    'body',
    'callSid',
    'projectPdid',
    'pdid',
    'isTest',
    'lastUpdatedField',
    'isDeleted',
    'changeSource',
  ]) as ActivityInterface;

  return assembledActivity;
};

// this method is basically useless but helps it match the pattern of other entities
export const assembleLowLevelActivityInterface = (ddbItem: { [key: string]: any }): LowLevelActivity => {
  return ddbItem as LowLevelActivity;
};

export const assembleActivityInterfaceFromDDBItems = (ddbItems: { [key: string]: any }[]): ActivityInterface => {
  if (ddbItems.length > 1) throw new Error("Expected only one item");
  
  const lowLevelActivity = assembleLowLevelActivityInterface(ddbItems[0]);
  return assembleActivity(lowLevelActivity);
};

export const fetchActivityInterface = async (pdid: string) => {
  const activityItems = await queryByPK(TableName, pdid);
  if (!activityItems.length) return null;
  return assembleActivityInterfaceFromDDBItems(activityItems);
};

export const fetchActivityInterfaceByCallSid = async (callSid: string) => {
  const activityItems = await queryByGSI(TableName, 1, buildActivityIndex1Keys(callSid));
  if (!activityItems.length) return null;
  return assembleActivityInterfaceFromDDBItems(activityItems);
};

export const putActivity = async (activity: Activity) => {
  const lowLevelActivity = buildLowLevelActivity(activity);
  const transaction = buildActivityPutTransaction(lowLevelActivity);
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

export const hardDeleteActivity = async (activity: Activity) => {
  const activityItems = await queryByPK(TableName, activity.pdid);
  const TransactItems = activityItems.map(item => ({
    Delete: {
      TableName,
      Key: {
        pk: item.pk,
        sk: item.sk,
      },
    },
  }));

  const transaction: TransactWriteCommandInput = {
    TransactItems,
  };

  if (!TransactItems.length) {
    return Promise.resolve();
  }
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

export const updateActivity = async (activity: Activity) => {
  const lowLevelActivity = buildLowLevelActivity(activity);
  let secondaryIndices;

  if (lowLevelActivity.lastUpdatedField === MutableActivityProperties.callSid) {
    const keys = buildActivityIndex1Keys(lowLevelActivity.callSid!)
    secondaryIndices = {
      indexNum: 1,
      pk: keys,
      sk: keys,
    }
  }
  
  return updateItem(lowLevelActivity, TableName, secondaryIndices);
};