import { ddb } from '@svc/config';
import { TransactWriteCommand } from '@aws-sdk/lib-dynamodb';
import { queryByPK, updateItem } from './repo-utils';
import { TransactWriteCommandInput } from '@aws-sdk/lib-dynamodb';
import { LowLevelCallRail, CallRailInterface } from '@improovy/types';
import _ from 'lodash';
import CallRail from '../models/CallRail';

const TableName = ddb.callRailTable;

export const buildCallRailPutTransaction = (callRail: LowLevelCallRail): TransactWriteCommandInput => {
  const transaction: TransactWriteCommandInput = {
    TransactItems: [
      { Put: {
        TableName,
        Item: callRail,
      },
      },
    ],
  };

  return transaction;
};

export const fetchCallRailInterface = async (phoneNumber: string) => {
  const callRailItems = await queryByPK(TableName, phoneNumber);
  if (!callRailItems.length) return null;
  const lowLevelInterface = assembleLowLevelCallRailInterface(callRailItems[0]);
  return assembleCallRail(lowLevelInterface);
};

export const buildLowLevelCallRail = (callRail: CallRail): LowLevelCallRail => {
  const {
    phoneNumber,
    isAttributed,
    createdAt,
    gclid,
    changeSource,
    isTest,
    lastUpdatedField,
  } = callRail;

  const lowLevelProfile: LowLevelCallRail = {
    pk: phoneNumber,
    sk: `CallRail#gclid#${gclid}`,
    type: 'CallRailGclid',
    createdAt,
    isAttributed,
    gclid,
    phoneNumber,
    changeSource,
    isTest,
  };

  if (lastUpdatedField) lowLevelProfile.lastUpdatedField = lastUpdatedField;
  
  return lowLevelProfile;
};

export const assembleCallRail = (profile: LowLevelCallRail): CallRailInterface => {
  const assembledCallRail = _.pick(profile, [
    'gclid',
    'createdAt',
    'isAttributed',
    'phoneNumber',
    'changeSource',
    'isTest',
    'lastUpdatedField',
  ]) as CallRailInterface;

  return assembledCallRail;
};

// this method is basically useless but helps it match the pattern of other entities
export const assembleLowLevelCallRailInterface = (ddbItem: { [key: string]: any }): LowLevelCallRail => {
  // @ts-expect-error
  return ddbItem;
};

export const putCallRail = async (callRail: CallRail) => {
  const lowLevelCallRail = buildLowLevelCallRail(callRail);
  const transaction = buildCallRailPutTransaction(lowLevelCallRail);
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

// todo: add ability to delete more than 25 items at once/handle paginating
export const hardDeleteCallRail = async (callRail: CallRail) => {
  const callRailItems = await queryByPK(TableName, callRail.phoneNumber);
  const TransactItems = callRailItems.map(item => ({
    Delete: {
      TableName,
      Key: {
        pk: item.pk,
        sk: item.sk,
      },
    },
  }));

  const transaction: TransactWriteCommandInput = {
    TransactItems,
  };

  if (!TransactItems.length) {
    return Promise.resolve();
  }
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

export const updateCallRail = async (callRail: CallRail) => {
  const lowLevelCallRail = buildLowLevelCallRail(callRail);

  return updateItem(lowLevelCallRail, TableName);
};