import { GetCommand, GetCommandInput, PutCommand, PutCommandInput } from '@aws-sdk/lib-dynamodb';
import { ddb } from '@svc/config';
import { DynamoDBRecord } from 'aws-lambda';
import { PipedriveRequest } from '@improovy/types';

const TableName = ddb.idempotencyTable;

export const addSeenToken = (token: string, seenBy = '', daysToExpiry: number | null = 3) => {
  const Item: {pk: string, TTL?: number} = {
    pk: buildSeenByToken(token, seenBy)
  }
  
  if (daysToExpiry !== null) {
    const nowInSeconds = Math.ceil(Date.now() / 1000000) * 1000; // rounds up to nearest 1000 seconds
    const additionalSeconds = daysToExpiry * 24 * 60 * 60;
    Item.TTL = nowInSeconds + additionalSeconds;
  }

  const put: PutCommandInput = {
    TableName,
    Item,
  };

  return ddb.documentClient.send(new PutCommand(put));
};

const buildSeenByToken = (token: string, seenBy: string) => `${token}#${seenBy}`

export const wasTokenSeen = async (token: string, seenBy = '') => {
  const get: GetCommandInput = {
    TableName,
    Key: {
      pk: buildSeenByToken(token, seenBy)
    },
  };

  const resp = await ddb.documentClient.send(new GetCommand(get));

  return !!resp.Item;
};

export const wasDDBRecordSeen = (record: DynamoDBRecord, seenBy?: string) => {
  return wasTokenSeen(getDDBRecordToken(record), seenBy);
};

export const addSeenDDBRecord = (record: DynamoDBRecord, seenBy?: string) => {
  return addSeenToken(getDDBRecordToken(record), seenBy);
};

const getDDBRecordToken = (record: DynamoDBRecord) => `DDB#${record.eventID!}`

// const getS3EventRecordToken = (record: S3EventRecord) => {
//   return `${record.s3.bucket.name}#${record.eventName}#${record.s3.object.key}`
// }

// export const wasS3RecordSeen = (record: S3EventRecord, seenBy?: string) => {
//   return wasTokenSeen(getS3EventRecordToken(record), seenBy);
// };

// export const addSeenS3Record = (record: S3EventRecord, seenBy?: string) => {
//   return addSeenToken(getS3EventRecordToken(record), seenBy);
// };

export const addSeenPipedriveRequest = (request: PipedriveRequest, seenBy?: string) => {
  return addSeenToken(buildPipedriveRequestUid(request), seenBy);
};

export const wasPipedriveRequestSeen = (request: PipedriveRequest, seenBy?: string) => {
  return !request.meta.isTest && wasTokenSeen(buildPipedriveRequestUid(request), seenBy);
};

export const buildPipedriveRequestUid = (request: PipedriveRequest) => {
  const eventEntity = request.meta.object;
  const eventAction = request.meta.action;
  const eventId = request.meta.id;
  const eventTimestamp = eventAction === 'added' ? 'NA' : request.meta.timestamp_micro;

  return `${eventEntity}#${eventId}#${eventAction}#${eventTimestamp}`;

};

// export const addSeenCallSid = (sid: string, seenBy?: string) => {
//   return addSeenToken(buildCallSidUid(sid), seenBy);
// };

// export const wasCallSidSeen = (sid: string, seenBy?: string) => {
//   return wasTokenSeen(buildCallSidUid(sid), seenBy); 
// };

// export const buildCallSidUid = (sid: string) => {
//   return `callSid#${sid}`;
// };

export const addSeenCallRailSignature = (signature: string, seenBy?: string) => {
  return addSeenToken(buildCallRailSignatureUid(signature), seenBy);
};

export const wasCallRailSignatureSeen = (signature: string, seenBy?: string) => {
  return wasTokenSeen(buildCallRailSignatureUid(signature), seenBy); 
};

export const buildCallRailSignatureUid = (signature: string) => {
  return `callRail#reqSignature#${signature}`;
};

export const addSeenGclid = (gclid: string, seenBy?: string) => {
  return addSeenToken(buildGclidUid(gclid), seenBy);
};

export const wasGclidSeen = (gclid: string, seenBy?: string) => {
  return wasTokenSeen(buildGclidUid(gclid), seenBy); 
};

export const buildGclidUid = (gclid: string) => {
  return `gclid#${gclid}`;
};