import { ddb } from '@svc/config';
import { TransactWriteCommand } from '@aws-sdk/lib-dynamodb';
import Project from '../models/Project';
import { queryByGSI, queryByPK, removeProperty, updateItem } from './repo-utils';
import { TransactWriteCommandInput } from '@aws-sdk/lib-dynamodb';
import { LowLevelProjectDetails, ProjectInterface } from '@improovy/types';
import _ from 'lodash';

const TableName = ddb.projectsTable;

export const buildProjectPutTransaction = (project: LowLevelProjectDetails): TransactWriteCommandInput => {
  const transaction: TransactWriteCommandInput = {
    TransactItems: [
      { Put: {
        TableName,
        Item: project,
      },
      },
    ],
  };

  return transaction;
};

export const buildLowLevelProjectDetails = (project: Project): LowLevelProjectDetails => {
  const {
    uuid,
    pdid,
    clientUuid,
    createdAt,
    isDeleted,
    projectStatus,
    salesStage,
    clientCost,
    lastUpdatedField,
    gclid,
    gaTransactionID,
    isTest,
    changeSource,
  } = project;

  const lowLevelProfile: LowLevelProjectDetails = {
    pk: uuid,
    sk: 'Details',
    type: 'ProjectDetails',
    changeSource,
    createdAt,
    uuid,
    isDeleted,
    projectStatus,
    salesStage,
  };
  
  if (clientUuid) {
    lowLevelProfile.gsi2_pk = buildProjectGSI2_PK(project as { clientUuid: string})
    lowLevelProfile.gsi2_sk = uuid;
    lowLevelProfile.clientUuid = clientUuid;
  }
  if (lastUpdatedField) lowLevelProfile.lastUpdatedField = lastUpdatedField;
  if (gclid) lowLevelProfile.gclid = gclid;
  if (clientCost !== undefined) lowLevelProfile.clientCost = clientCost;
  if (gaTransactionID) lowLevelProfile.gaTransactionID = gaTransactionID;
  if (isTest) lowLevelProfile.isTest = isTest;
  if (pdid) {
    lowLevelProfile.pdid = pdid;
    lowLevelProfile.gsi1_pk = `PDID#${pdid}`;
    lowLevelProfile.gsi1_sk = `PDID#${pdid}`;
  }
  
  return lowLevelProfile;
};

export const assembleProjectDetails = (profile: LowLevelProjectDetails): ProjectInterface => {
  const assembledProfile = _.pick(profile, [
    'gclid',
    'createdAt',
    'clientUuid',
    'gaTransactionID',
    'isDeleted',
    'uuid',
    'projectStatus',
    'salesStage',
    'clientCost',
    'pdid',
    'lastUpdatedField',
    'isTest',
    'changeSource',
  ]);

  return assembledProfile;
};

export const assembleLowLevelProjectInterface = (ddbItems: { [key: string]: any }[]): LowLevelProjectDetails => {
  let details: LowLevelProjectDetails;

  for (let item of ddbItems) {
    switch (item.type) {
      case 'ProjectDetails':
        // @ts-expect-error
        details = item;
        break;
      default:
        throw new Error('Unkown item type in assembleProjectInterface');
        
        break;
    }
  }

  // @ts-expect-error
  if (!details) throw new Error('item type ProjectDetails not found');

  return details;
};

export const fetchProjectInterface = async (uuid: string) => {
  const projectItems = await queryByPK(TableName, uuid);
  // const lowLevelProject = assembleLowLevelProjectInterface(projectItems);
  return assembleProjectInterfaceFromDDBItems(projectItems);
};

export const fetchProjectInterfaceByPdid = async (pdid: string) => {
  const projectItems = await queryByGSI(TableName, 1, `PDID#${pdid}`);
  if (!projectItems.length) return null;
  return assembleProjectInterfaceFromDDBItems(projectItems);
};

export const putProject = async (project: Project) => {
  const lowLevelProject = buildLowLevelProjectDetails(project);
  const transaction = buildProjectPutTransaction(lowLevelProject);
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

// // todo: add ability to delete more than 25 items at once/handle paginating
export const hardDeleteProject = async (project: Project) => {
  const projectItems = await queryByPK(TableName, project.uuid);
  const TransactItems = projectItems.map(item => ({
    Delete: {
      TableName,
      Key: {
        pk: item.pk,
        sk: item.sk,
      },
    },
  }));

  const transaction: TransactWriteCommandInput = {
    TransactItems,
  };

  if (!TransactItems.length) {
    return Promise.resolve();
  }
  
  return ddb.documentClient.send(
    new TransactWriteCommand(transaction),
  );
};

export const updateProject = async (project: Project) => {
  const details = buildLowLevelProjectDetails(project);
  
  return updateItem(details, TableName);
};

export const removeProjectProperty = async (project: Project) => {
  const details = buildLowLevelProjectDetails(project);

  if (details.lastUpdatedField === undefined) throw new Error("lastUpdatedField needs to be set to determine which property to remove");
  
  // object reassignment needed so typescript recognizes lastUpdatedField as not undefined
  return removeProperty({...details, lastUpdatedField: details.lastUpdatedField}, TableName);
};

export const buildProjectGSI2_PK = <Type extends { clientUuid: string }>(project: Type) => `Project#clientUuid#${project.clientUuid}`;

export const assembleProjectInterfaceFromDDBItems = (ddbItems: { [key: string]: any }[]): ProjectInterface => {
  const lowLevelClient = assembleLowLevelProjectInterface(ddbItems);
  return assembleProjectDetails(lowLevelClient);
};