import parseCallSid from '@svc/lib/parseCallSid';
import { ChangeSource } from '@improovy/types';
import { PipedriveActivity, PipedriveActivityFieldKeys } from '@improovy/types';
import { ActivityInterface, ActivityType, MutableActivityProperties } from '../types/activities';
import { assembleActivity, buildLowLevelActivity, fetchActivityInterface, fetchActivityInterfaceByCallSid, hardDeleteActivity, putActivity, updateActivity } from '../ddbRepos/activities-repo';

export default class Activity implements ActivityInterface {
  #activityType: ActivityType;

  #body?: string;

  #callSid?: string;

  #projectPdid: string;
  
  #pdid: string;

  #isTest?: boolean;

  #isDeleted?: boolean;

  #lastUpdatedField?: MutableActivityProperties;

  changeSource: ChangeSource;

  static async fetch(pdid: string) {
    const activityInterface = await fetchActivityInterface(pdid);
    if (!activityInterface) return null;
    return new Activity(activityInterface)
  }

  static async fetchByCallSid(callSid: string) {
    const activityInterface = await fetchActivityInterfaceByCallSid(callSid);
    if (!activityInterface) return null;
    return new Activity(activityInterface)
  }

  static differingFields(activity1: Activity, activity2: Activity) {
    const diff: MutableActivityProperties[] = [];

    const llActivityInterface1 = buildLowLevelActivity(activity1);
    const llAactivityInterface2 = buildLowLevelActivity(activity2);
    
    const activityInterface1 = assembleActivity(llActivityInterface1);
    const activityInterface2 = assembleActivity(llAactivityInterface2);

    for (let key of Object.values(MutableActivityProperties)) {
      if (activityInterface1[key] !== activityInterface2[key] && key !== MutableActivityProperties.lastUpdatedField && key !== MutableActivityProperties.changeSource) {
        diff.push(key);
      }
    }

    return diff;
  }

  static fromPipedrive(pipedriveActivity: PipedriveActivity): Activity {
    const activityInterface: ActivityInterface = {
      pdid: `${pipedriveActivity[PipedriveActivityFieldKeys.pdid]}`,
      activityType: pipedriveActivity[PipedriveActivityFieldKeys.activityType],
      projectPdid: `${pipedriveActivity[PipedriveActivityFieldKeys.projectPdid]}`,
      changeSource: ChangeSource.pipedrive
    }

    const body = pipedriveActivity[PipedriveActivityFieldKeys.body];
    const isTest = pipedriveActivity.isTest;

    if (body) activityInterface.body = body;
    if (isTest) activityInterface.isTest = isTest;

    return new Activity(activityInterface)
  }

  constructor(activityInterface: ActivityInterface) {
    this.#activityType = activityInterface.activityType;
    this.#isDeleted = activityInterface.isDeleted || false;
    this.changeSource = activityInterface.changeSource;
    if (activityInterface.callSid) this.#callSid = activityInterface.callSid;
    if (activityInterface.body) {
      this.#body = activityInterface.body;
      const callSid = parseCallSid(this.#body);
      if (callSid) this.#callSid = callSid;
    };
    this.#projectPdid = activityInterface.projectPdid;
    this.#pdid = activityInterface.pdid;
    if (activityInterface.isTest) this.#isTest = activityInterface.isTest;
    if (activityInterface.lastUpdatedField) this.#lastUpdatedField = activityInterface.lastUpdatedField;
  }

  async updateBody(body: string) {
    const fieldsToUpdate = [MutableActivityProperties.body];
    this.#body = body;
    const newCallSid = parseCallSid(body);
    if (newCallSid !== this.#callSid) {
      fieldsToUpdate.push(MutableActivityProperties.callSid);
      this.#callSid = newCallSid || undefined;
    }
    return this.updateFields(fieldsToUpdate);
  }

  async put() {
    return putActivity(this);
  }

  async hardDelete() {
    return hardDeleteActivity(this);
  }

  async updateFields(fields: MutableActivityProperties[]) {
    for (let field of fields) {
      this.#lastUpdatedField = field;
      await updateActivity(this);
    }
  }

  get activityType() {
    return this.#activityType;
  }

  get lastUpdatedField() {
    return this.#lastUpdatedField;
  }

  get body() {
    return this.#body;
  }

  get callSid() {
    return this.#callSid;
  }

  get projectPdid() {
    return this.#projectPdid;
  }

  get pdid() {
    return this.#pdid;
  }

  get isTest() {
    return this.#isTest || false;
  }

  get isDeleted() {
    return this.#isDeleted || false;
  }
  
}