import {
  fetchClientInterface,
  fetchClientInterfaceByPdid,
  fetchClientInterfaceByPhoneNumber,
  fetchClientProjectInterfaces,
  hardDeleteClient,
  putClient,
  removeClientProperty,
  updateClient,
} from '../ddbRepos/clients-repo';
import { assembleClientInterface, buildLowLevelClient } from '../ddbRepos/clients-repo-utils';
import {
  ClientContactMethod,
  ClientIncomingContactMethod,
  ClientInterface,
  ContactType,
  LowLevelClientProfile,
  MutableClientProperties,
} from '@improovy/types';
import {
  PipedrivePersonFieldKeys,
  PipedrivePerson,
  inboundIncomingContactMethodMap,
} from '@improovy/types';
import { convertPipedriveTimeToMS, generateValidatedUuid, isValidUuid } from './modelUtils';
import Project from './Project';
import { ChangeSource } from '@improovy/types';
import formatPhoneNumber from '@svc/lib/formatPhoneNumber';

const addContactMethodDefaults = (method: ClientContactMethod) => {
  method.isDeleted = method.isDeleted || false;
};

const buildContactMethodsFromPipedrive = (person: PipedrivePerson): ClientContactMethod[] => {
  const phones = person.phone;
  const emails = person.email;

  const contacts: ClientContactMethod[] = [];
  for (let phone of phones) {
    if (!phone.value) continue;
    contacts.push({
      contactType: ContactType.PHONE,
      contactAddress: formatPhoneNumber(phone.value),
    });
  }

  for (let email of emails) {
    if (!email.value) continue;
    contacts.push({
      contactType: ContactType.EMAIL,
      contactAddress: email.value,
    });
  }

  return contacts;
};

export default class Client implements ClientInterface {
  #gclid?: string;

  #isDeleted: boolean;

  #uuid!: string;

  #pdid?: string;

  #createdAt: number;

  #incomingContactMethod: ClientIncomingContactMethod;

  #contactMethods: ClientContactMethod[];

  #lastUpdatedField?: MutableClientProperties;

  #projects? : Project[];

  #pipedriveName: string;

  #isTest?: boolean

  changeSource: ChangeSource;

  static differingFields(client1: Client, client2: Client) {
    const diff: MutableClientProperties[] = [];
    const clientInterface1 = assembleClientInterface(buildLowLevelClient(client1));
    const clientInterface2 = assembleClientInterface(buildLowLevelClient(client2));

    for (let key of Object.values(MutableClientProperties)) {
      if (clientInterface1[key] !== clientInterface2[key] && key !== MutableClientProperties.lastUpdatedField && key !== MutableClientProperties.changeSource) {
        diff.push(key);
      }
    }

    return diff;
  }

  static async fetch(uuid:string) {
    const clientInterface = await fetchClientInterface(uuid);
    if (!clientInterface) return null;
    return new Client(clientInterface);
  }

  static async fetchByPdid(pdid:string) {
    const clientInterface = await fetchClientInterfaceByPdid(pdid);
    if (!clientInterface) return null;
    const client = new Client(clientInterface);
    return client;
  }

  static async fetchByPhoneNumber(phoneNumber: string) {
    // todo: add unique constraints for phone numbers
    const clientInterface = await fetchClientInterfaceByPhoneNumber(phoneNumber)
    if (!clientInterface) return null;
    return new Client(clientInterface);
  }

  static fromPipedrive(pipedrivePerson: PipedrivePerson) {
    const pdIncomingContactMethodID = pipedrivePerson[PipedrivePersonFieldKeys.incomingContactMethod];

    const incomingContactMethod = pdIncomingContactMethodID ? inboundIncomingContactMethodMap[pdIncomingContactMethodID] : ClientIncomingContactMethod.UNKNOWN;

    const clientInterface: ClientInterface = {
      pdid: `${pipedrivePerson.id}`,
      gclid: pipedrivePerson[PipedrivePersonFieldKeys.gclid],
      uuid: pipedrivePerson[PipedrivePersonFieldKeys.uuid],
      createdAt: convertPipedriveTimeToMS(pipedrivePerson.add_time),
      incomingContactMethod,
      contactMethods: buildContactMethodsFromPipedrive(pipedrivePerson),
      pipedriveName: pipedrivePerson[PipedrivePersonFieldKeys.pipedriveName],
      changeSource: ChangeSource.pipedrive,
    };

    if (pipedrivePerson.isTest) clientInterface.isTest = pipedrivePerson.isTest;
    
    return new Client(clientInterface);
  }

  constructor(clientInterface: ClientInterface | LowLevelClientProfile) {
    if (clientInterface.gclid) this.#gclid = clientInterface.gclid;
    if (clientInterface.isTest) this.#isTest = clientInterface.isTest;
    this.#isDeleted = clientInterface.isDeleted || false;
    if (clientInterface.pdid) this.#pdid = clientInterface.pdid;
    if (clientInterface.lastUpdatedField) this.#lastUpdatedField = clientInterface.lastUpdatedField;

    this.#assignAndValidateUUID(clientInterface.uuid);

    this.#createdAt = clientInterface.createdAt;
    this.#incomingContactMethod = clientInterface.incomingContactMethod;
    this.#pipedriveName = clientInterface.pipedriveName;
    this.changeSource = clientInterface.changeSource

    // @ts-expect-error
    if (clientInterface.contactMethods) {
      // @ts-expect-error
      clientInterface.contactMethods.forEach(addContactMethodDefaults);
    }
    
    // @ts-expect-error
    this.#contactMethods = clientInterface.contactMethods || [];
    this.formatContactMethods()
  }

  formatContactMethods() {
    const newMethods: ClientContactMethod[] = [];
    const seen = new Set();
    const oldMethods = this.#contactMethods;
    for (let method of oldMethods) {
      const newAddress = method.contactType === ContactType.PHONE ? formatPhoneNumber(method.contactAddress) : method.contactAddress
      const uuid = `${method.contactType}${newAddress}`
      if (seen.has(uuid)) continue;
      newMethods.push({...method, contactAddress: newAddress})
      seen.add(uuid)
    }

    this.#contactMethods = newMethods;
  }

  async put() {
    return putClient(this);
  }

  async hardDelete() {
    return hardDeleteClient(this);
  }

  async updateFields(fields: MutableClientProperties[]) {
    for (let field of fields) {
      this.#lastUpdatedField = field;
      await updateClient(this);
    }
  }

  async updateGclid(gclid?: string | undefined | null) {
    if (!gclid) return this.removeGclid();

    this.#gclid = gclid;
    this.#lastUpdatedField = MutableClientProperties.gclid;
    return updateClient(this);
  }

  async removeGclid() {
    this.#gclid = undefined;
    this.#lastUpdatedField = MutableClientProperties.gclid;
    return removeClientProperty(this);
  }

  async fetchProjects() {
    if (!this.#projects) {
      const projectInterfaces = await fetchClientProjectInterfaces(this);
      this.#projects = projectInterfaces.map(proj => new Project(proj));
    }
    return this.#projects;
  }

  async forceFetchProjects() {
    const projectInterfaces = await fetchClientProjectInterfaces(this);
    this.#projects = projectInterfaces.map(proj => new Project(proj));
    return this.#projects;
  }

  async softDelete() {
    this.#isDeleted = true;
    this.#lastUpdatedField = MutableClientProperties.isDeleted;
    return updateClient(this);
  }

  get uuid() {
    return this.#uuid;
  }

  get isTest() {
    return this.#isTest || false
  }

  get pipedriveName() {
    return this.#pipedriveName;
  }

  get lastUpdatedField() {
    return this.#lastUpdatedField;
  }

  get incomingContactMethod() {
    return this.#incomingContactMethod;
  }

  get createdAt() {
    return this.#createdAt;
  }

  get contactMethods() {
    return this.#contactMethods;
  }

  get pdid() {
    return this.#pdid;
  }

  get gclid() {
    return this.#gclid;
  }

  get isDeleted() {
    return this.#isDeleted;
  }
  
  #assignAndValidateUUID(clientUuid?: string): void {
    if (clientUuid && !isValidUuid(clientUuid)) throw new Error('UUID cannot begin with string "PDID"');
    
    this.#uuid = clientUuid || generateValidatedUuid();
  }
}