import { convertPipedriveTimeToMS, generateValidatedUuid, isValidUuid } from './modelUtils';
import {
  MutableProjectProperties,
  ProjectInterface,
  ProjectSalesStage,
  ProjectStatus,
} from '@improovy/types';
import {
  assembleProjectDetails,
  buildLowLevelProjectDetails,
  fetchProjectInterface,
  fetchProjectInterfaceByPdid,
  hardDeleteProject,
  putProject,
  removeProjectProperty,
  updateProject,
} from '../ddbRepos/projects-repo';
import Client from './Client';
import { PipedriveDeal, PipedriveDealFieldKeys } from '@improovy/types';
import { ChangeSource } from '@improovy/types';

const mapProjectStatus = (pipedriveStatus: string, pipedriveSalesStage: number | string) => {
  switch (pipedriveStatus) {
    case 'open':
      return ProjectStatus.OPEN;
    case 'lost':
    case 'deleted':
      return ProjectStatus.LOST;
    case 'won':
      return `${pipedriveSalesStage}` === '6' ? ProjectStatus.CLOSED : ProjectStatus.OPEN;
    default:
      throw new Error('unknown pipedrive project status');
      
  }
};

const mapProjectSalesStage = (pipedriveSalesStage: number | string) => {
  switch (`${pipedriveSalesStage}`) {
    case '1':
      return ProjectSalesStage.UNATTEMPTED;
    case '14':
      return ProjectSalesStage.ATTEMPTED;
    case '2':
      return ProjectSalesStage.ESTABLISHED_CONTACT;
    case '3':
      return ProjectSalesStage.ESTIMATED;
    case '6':
      return ProjectSalesStage.BOOKED;
    default:
      throw new Error('unknown pipedrive project stage');
  }
};

export default class Project implements ProjectInterface {
  #gclid?: string;

  #createdAt: number;

  #clientUuid?: string;

  #gaTransactionID?: string;

  #isDeleted: boolean;

  #uuid!: string;

  #projectStatus: ProjectStatus;

  #salesStage: ProjectSalesStage;

  #clientCost?: number;

  #pdid?: string;

  #lastUpdatedField?: MutableProjectProperties;

  #client?: Client | null;

  #isTest?: boolean;

  #changeSource: ChangeSource;

  static async fetch(uuid:string) {
    return new Project(await fetchProjectInterface(uuid));
  }

  static async fetchByPdid(pdid:string) {
    const projectInterface = await fetchProjectInterfaceByPdid(pdid);
    if (!projectInterface) return null;
    return new Project(projectInterface);
  }

  static differingFields(project1: Project, project2: Project) {
    const diff: MutableProjectProperties[] = [];
    const projectInterface1 = assembleProjectDetails(buildLowLevelProjectDetails(project1));
    const projectInterface2 = assembleProjectDetails(buildLowLevelProjectDetails(project2));

    for (let key of Object.values(MutableProjectProperties)) {
      if (projectInterface1[key] !== projectInterface2[key] && key !== MutableProjectProperties.lastUpdatedField && key !== MutableProjectProperties.changeSource) {
        diff.push(key);
      }
    }

    return diff;
  }

  static dealIsProject(deal: PipedriveDeal) {
    const pipelineID = deal[PipedriveDealFieldKeys.pipeline]
    return pipelineID === 1 || pipelineID === '1'
  }

  static async fromPipedrive(pdProject: PipedriveDeal) {
    const client = await Client.fetchByPdid(`${pdProject[PipedriveDealFieldKeys.personPdid]}`);
    
    if (!client) return null;

    const pipelineID = `${pdProject[PipedriveDealFieldKeys.pipeline]}`;

    const projectInterface: ProjectInterface = {
      pdid: `${pdProject.id}`,
      gclid: pdProject[PipedriveDealFieldKeys.gclid],
      createdAt: convertPipedriveTimeToMS(pdProject[PipedriveDealFieldKeys.createdAt]),
      clientUuid: client.uuid,
      gaTransactionID: pdProject[PipedriveDealFieldKeys.gaTransactionID],
      isDeleted: pipelineID === '5' || pdProject[PipedriveDealFieldKeys.projectStatus] === 'deleted',
      uuid: pdProject[PipedriveDealFieldKeys.uuid] || generateValidatedUuid(),
      projectStatus: mapProjectStatus(pdProject[PipedriveDealFieldKeys.projectStatus], pdProject[PipedriveDealFieldKeys.stage]),
      salesStage: mapProjectSalesStage(pdProject[PipedriveDealFieldKeys.stage]),
      clientCost: Number(pdProject[PipedriveDealFieldKeys.clientCost]) || undefined,
      changeSource: ChangeSource.pipedrive,
    };
    
    if (pdProject.isTest) projectInterface.isTest = pdProject.isTest;

    return new Project(projectInterface);
  }

  constructor(projectInterface: ProjectInterface) {
    if (projectInterface.gclid) this.#gclid = projectInterface.gclid;
    this.#isDeleted = projectInterface.isDeleted || false;
    if (projectInterface.pdid) this.#pdid = projectInterface.pdid;
    if (projectInterface.lastUpdatedField) this.#lastUpdatedField = projectInterface.lastUpdatedField;
    if (projectInterface.clientUuid) this.#clientUuid = projectInterface.clientUuid;
    if (projectInterface.isTest) this.#isTest = projectInterface.isTest;

    this.#assignAndValidateUUID(projectInterface.uuid);

    this.#createdAt = projectInterface.createdAt;
    this.#gaTransactionID = projectInterface.gaTransactionID;
    this.#projectStatus = projectInterface.projectStatus;
    this.#salesStage = projectInterface.salesStage;
    this.#clientCost = projectInterface.clientCost;
    this.#changeSource = projectInterface.changeSource;
  }

  async fetchClient() {
    if (!this.#client && this.clientUuid) {
      this.#client = await Client.fetch(this.clientUuid);
    }
    
    return this.#client;
  }

  async put() {
    return putProject(this);
  }

  async hardDelete() {
    return hardDeleteProject(this);
  }

  async updateFields(fields: MutableProjectProperties[]) {
    for (let field of fields) {
      this.#lastUpdatedField = field;
      await updateProject(this);
    }
  }

  async updateGclid(gclid?: string | undefined | null) {
    if (!gclid) return this.removeGclid();
    if (this.#gclid) throw new Error("project already has a gclid");
    
    this.#gclid = gclid;
    this.#lastUpdatedField = MutableProjectProperties.gclid;
    return updateProject(this);
  }

  async removeGclid() {
    this.#gclid = undefined;
    this.#lastUpdatedField = MutableProjectProperties.gclid;
    return removeProjectProperty(this);
  }

  updateClientUuid(uuid: string) {
    if (!uuid) throw new Error('clientUuid must exist');
    
    this.#clientUuid = uuid;
    this.#lastUpdatedField = MutableProjectProperties.clientUuid;
    return updateProject(this);
  }

  updateSalesStage(salesStage: ProjectSalesStage) {
    if (!salesStage) throw new Error('salesStage must exist');
    
    this.#salesStage = salesStage;
    this.#lastUpdatedField = MutableProjectProperties.salesStage;
    return updateProject(this);
  }

  async softDelete() {
    this.#isDeleted = true;
    this.#lastUpdatedField = MutableProjectProperties.isDeleted;
    return updateProject(this);
  }

  get uuid() {
    return this.#uuid;
  }

  get changeSource() {
    return this.#changeSource
  }

  get clientCost() {
    return this.#clientCost;
  }

  get isTest() {
    return this.#isTest || false
  }

  get gaTransactionID() {
    return this.#gaTransactionID;
  }

  get clientUuid() {
    return this.#clientUuid;
  }

  get projectStatus() {
    return this.#projectStatus;
  }

  get salesStage() {
    return this.#salesStage;
  }

  get lastUpdatedField() {
    return this.#lastUpdatedField;
  }

  get createdAt() {
    return this.#createdAt;
  }

  get pdid() {
    return this.#pdid;
  }

  get gclid() {
    return this.#gclid;
  }

  get isDeleted() {
    return this.#isDeleted;
  }
  
  #assignAndValidateUUID(projectUuid?: string): void {
    if (projectUuid && !isValidUuid(projectUuid)) throw new Error('UUID cannot begin with string "PDID"');
    
    this.#uuid = projectUuid || generateValidatedUuid();
  }
}