import { apiGatewayConfig, AWS_REGION, CALLRAIL_SIGNING_KEY, ddb, lambdaConfig } from '@svc/config';
import { handler } from '@svc/handlers/callRail/callRailRouter';
import { ApiGatewayHandlerInvoker, ApiGatewayInvocation } from '@tests/utils/handler-invokers/ApiGatewayHandlerInvoker';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import getCallRailSignature from '@svc/lib/getCallRailSignature';
import uuid from '@svc/lib/uuid';

const testDataGenerator = new TestDataGenerator();

const apiInvoker = new ApiGatewayHandlerInvoker({
  baseUrl: apiGatewayConfig.getBaseUrl(),
  handler,
});

const getInvokerTemplate = () => {
  const invokerTemplate: ApiGatewayInvocation = {
    event : {
      pathTemplate: 'callrail',
      httpMethod: 'POST',
      body: JSON.stringify({msg: uuid()}),
    }
  }
  return invokerTemplate
}

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.activitiesTable,
};

describe('`callRailRouter`', () => {
  // todo: most of these can be unit tests
  it('returns 401 if there is no Signature header', async () => {
    const response = await apiInvoker.invoke(getInvokerTemplate());
    
    expect(response.statusCode).toEqual(401);
  });

  it('returns 401 if the signature header does not match the body', async () => {
    const response = await apiInvoker.invoke({ event: {
      ...getInvokerTemplate().event,
      headers: {
        Signature: 'wrong signature'
      }
    }});
    
    expect(response.statusCode).toEqual(401);
  });

  it('returns 400 if there is no "type" query string param', async () => {
    const event = getInvokerTemplate().event;
    const response = await apiInvoker.invoke({ event: {
      ...event,
      headers: {
        Signature: getCallRailSignature(event.body, CALLRAIL_SIGNING_KEY)
      }}
    });
    
    expect(response.statusCode).toEqual(400);
  });

  it('returns 400 if the "type" query string param does not match any known types', async () => {
    const event = getInvokerTemplate().event;
    const response = await apiInvoker.invoke({ event: {
      ...event,
      queryStringParameters: {
        type: 'unknownType'
      },
      headers: {
        Signature: getCallRailSignature(event.body, CALLRAIL_SIGNING_KEY),
      }}
    });
    
    expect(response.statusCode).toEqual(400);
  });

  it('successfully invokes the addNewCallRail function', async () => {
    const event = getInvokerTemplate().event;
    const testStr = uuid();
    const testBody = JSON.stringify({testStr})
    const Signature = getCallRailSignature(testBody, CALLRAIL_SIGNING_KEY);
    const response = await apiInvoker.invoke({ event: {
      ...event,
      body: testBody,
      queryStringParameters: {
        type: 'postCall'
      },
      headers: {
        Signature,
      }}
    });
    
    expect(response.statusCode).toEqual(200);
    await expect({region: AWS_REGION, function: `${lambdaConfig.functionNamePrefix}APIGWAddNewCallRail`, timeout: 50000}).toHaveLog(testStr)
  });
  
  afterAll(() => {
    return testDataGenerator.deleteAll();
  });
});