import { apiGatewayConfig, AWS_REGION, ddb, pipedriveAuthHeader, TEST_PERSON_PDID, TEST_PERSON_UUID  } from '@svc/config';
import { handler } from '@svc/handlers/pipedrive/projects/addPipedriveDealToProjects';
import { buildLowLevelProjectDetails } from '@svc/lib/ddbRepos/projects-repo';
import { PipedriveDealFieldKeys } from '@improovy/types';
import { ApiGatewayHandlerInvoker } from '@tests/utils/handler-invokers/ApiGatewayHandlerInvoker';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import _ from 'lodash';

const testDataGenerator = new TestDataGenerator();

const apiInvoker = new ApiGatewayHandlerInvoker({
  baseUrl: apiGatewayConfig.getBaseUrl(),
  handler,
});

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.projectsTable,
};

describe('`addPipedriveDealToProjects`', () => {
  it('adds the deal from pipedrive to the projects table', async () => {
    const req = testDataGenerator.getPipedriveDealRequestBody();
    const testProject = req.current;

    const testClient = testDataGenerator.getClient({ overrides: { pdid: `${testProject.person_id}` } });

    await testClient.put();
    await waitForGSIUpdates();
    
    // needed to so the created project is deleted
    const testProject2 = testDataGenerator.getProject({ overrides: { uuid: testProject.e4b1a5fdc67bdbc43f89d319658446f68f24edf6! } });
    const details = buildLowLevelProjectDetails(testProject2);

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/new',
        httpMethod: 'POST',
        body: JSON.stringify(req),
        headers: pipedriveAuthHeader,
      },
    });
    
    expect(response.statusCode).toEqual(200);

    await expect(awsTestOptions).toHaveItem(_.pick(details, ['pk', 'sk']));
  });

  it('adds the deal and its client to the database if the client is not already in there', async () => {
    const req = testDataGenerator.getPipedriveDealRequestBody({ [PipedriveDealFieldKeys.clientPdid]: TEST_PERSON_PDID }); 
    
    const testDeal = req.current;

    // needed so the created project is deleted
    const testProject = testDataGenerator.getProject({ overrides: { uuid: testDeal[PipedriveDealFieldKeys.uuid]! } });

    // needed so the created client is deleted
    testDataGenerator.getClient({overrides: {uuid: TEST_PERSON_UUID}})

    const details = buildLowLevelProjectDetails(testProject);

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/new',
        httpMethod: 'POST',
        body: JSON.stringify(req),
        headers: pipedriveAuthHeader,
      },
    });
    
    expect(response.statusCode).toEqual(200);

    await expect(awsTestOptions).toHaveItem(_.pick(details, ['pk', 'sk']));
  });

  it('ignores deals in a different pipeline', async () => {
    const testDealReq = testDataGenerator.getPipedriveDealRequestBody({[PipedriveDealFieldKeys.pipeline]: 29, [PipedriveDealFieldKeys.stage]: 50});

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/new',
        httpMethod: 'POST',
        body: JSON.stringify(testDealReq),
        headers: pipedriveAuthHeader,
      },
    });

    expect(response.statusCode).toEqual(200);

    await waitForGSIUpdates();

    const keys = {pk: testDealReq.current[PipedriveDealFieldKeys.uuid], sk: 'Details'};

    await expect(awsTestOptions).not.toHaveItem(keys);
  });
  
  afterAll(() => {
    return testDataGenerator.deleteAll();
  });
});