import { apiGatewayConfig, AWS_REGION, ddb, pipedriveAuthHeader  } from '@svc/config';
import { handler } from '@svc/handlers/pipedrive/projects/updateProjectFromPipedriveDeal';
import Project from '@svc/lib/models/Project';
import { buildLowLevelProjectDetails } from '@svc/lib/ddbRepos/projects-repo';
import { PipedriveDealFieldKeys } from '@improovy/types';
import { ApiGatewayHandlerInvoker } from '@tests/utils/handler-invokers/ApiGatewayHandlerInvoker';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import _ from 'lodash';

const testDataGenerator = new TestDataGenerator();

const apiInvoker = new ApiGatewayHandlerInvoker({
  baseUrl: apiGatewayConfig.getBaseUrl(),
  handler,
});

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.projectsTable,
};

describe('`updateProjectFromPipedriveDeal`', () => {
  it('updates the deal from pipedrive in the projects table', async () => {
    // add a new person to the database
    const testPerson = testDataGenerator.getClient();
    await testPerson.put();

    await waitForGSIUpdates();

    // add a project belonging to that person
    const testDealReqOld = testDataGenerator.getPipedriveDealRequestBody({ [PipedriveDealFieldKeys.clientPdid]: testPerson.pdid });
    const testProjectOld = await Project.fromPipedrive(testDealReqOld.current);
    await testProjectOld!.put();

    await waitForGSIUpdates();

    const newGclid = 'test_gclid';
    
    // create an update request with the new gclid
    const testDealReqNew = testDataGenerator.getPipedriveDealRequestBody({ [PipedriveDealFieldKeys.uuid]: testProjectOld!.uuid, [PipedriveDealFieldKeys.gclid]: newGclid, [PipedriveDealFieldKeys.pdid]: testProjectOld!.pdid, [PipedriveDealFieldKeys.clientPdid]: testPerson.pdid });
    
    const newTestProject = await Project.fromPipedrive(testDealReqNew.current);
    const lowLevelTestProject = buildLowLevelProjectDetails(newTestProject!);

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/update',
        httpMethod: 'POST',
        body: JSON.stringify(testDealReqNew),
        headers: pipedriveAuthHeader,
      },
    });

    expect(response.statusCode).toEqual(200);

    await waitForGSIUpdates();

    await expect(awsTestOptions).toHaveItem(_.pick(lowLevelTestProject, ['pk', 'sk']), { ...lowLevelTestProject, gclid: newGclid, lastUpdatedField: 'gclid' });
  });

  it('if the project is not yet in the projects table, it adds it', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();

    const testDealReq = testDataGenerator.getPipedriveDealRequestBody({ [PipedriveDealFieldKeys.clientPdid]: testClient.pdid });
    
    await waitForGSIUpdates();
    const testProject = await Project.fromPipedrive(testDealReq.current);
    
    const lowLevelTestProject = buildLowLevelProjectDetails(testProject!);

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/update',
        httpMethod: 'POST',
        body: JSON.stringify(testDealReq),
        headers: pipedriveAuthHeader,
      },
    });

    expect(response.statusCode).toEqual(200);

    await waitForGSIUpdates();

    await expect(awsTestOptions).toHaveItem(_.pick(lowLevelTestProject, ['pk', 'sk']), lowLevelTestProject);
  });

  it('ignores deals in a different pipeline', async () => {
    const testDealReq = testDataGenerator.getPipedriveDealRequestBody({[PipedriveDealFieldKeys.pipeline]: 29, [PipedriveDealFieldKeys.stage]: 50, }); 

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/update',
        httpMethod: 'POST',
        body: JSON.stringify(testDealReq),
        headers: pipedriveAuthHeader,
      },
    });

    expect(response.statusCode).toEqual(200);

    await waitForGSIUpdates();

    const keys = {pk: testDealReq.current[PipedriveDealFieldKeys.uuid], sk: 'Details'};

    await expect(awsTestOptions).not.toHaveItem(keys);
  });

  it('removes deals updated to a different pipeline', async () => {
    const testClient = testDataGenerator.getClient();
    await testClient.put();
    const testDealReq = testDataGenerator.getPipedriveDealRequestBody({[PipedriveDealFieldKeys.personPdid]: testClient.pdid});
    const testDeal = await Project.fromPipedrive(testDealReq.current);

    await testDeal!.put();
    testDealReq.previous = { ...testDealReq.current }
    testDealReq.current[PipedriveDealFieldKeys.pipeline] = 29;
    testDealReq.current[PipedriveDealFieldKeys.stage] = 50;

    const response = await apiInvoker.invoke({
      event: {
        pathTemplate: 'pipedrive/projects/update',
        httpMethod: 'POST',
        body: JSON.stringify(testDealReq),
        headers: pipedriveAuthHeader,
      },
    });

    expect(response.statusCode).toEqual(200);

    await waitForGSIUpdates();

    const keys = {pk: testDealReq.current[PipedriveDealFieldKeys.uuid], sk: 'Details'};

    const lowLevelTestProject = buildLowLevelProjectDetails(testDeal!)

    await expect(awsTestOptions).toHaveItem(keys, {...lowLevelTestProject, isDeleted: true, lastUpdatedField: 'isDeleted'});
  });

  afterAll(() => {
    return testDataGenerator.deleteAll();
  });
});