import { AWS_REGION, ddb } from '@svc/config';
import Activity from '@svc/lib/models/Activity';
import { buildActivityIndex1Keys, buildLowLevelActivity } from '@svc/lib/ddbRepos/activities-repo';
import _ from 'lodash';
import TestDataGenerator, { generateCallSidBody } from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import { MutableActivityProperties } from '@improovy/types';

const testDataGenerator = new TestDataGenerator();

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.activitiesTable,
};

describe('`Activity`', () => {
  describe('`Activity::fetch', () => {
    it('fetches an activity via pdid', async () => {
      const testActivity = testDataGenerator.getActivity();
      await testActivity.put();
      const testPdid = testActivity.pdid;
  
      const returnedActivity = await Activity.fetch(testPdid);
  
      expect(returnedActivity!.pdid).toBe(testPdid);
      expect(returnedActivity).toEqual(expect.objectContaining(testActivity));
      expect(testActivity).toEqual(expect.objectContaining(returnedActivity));
    });

    it('returns null if the activity is not found', async () => {
      const returnedActivity = await Activity.fetch('nonexistentPdid')
  
      expect(returnedActivity).toBe(null);
    });
  });

  describe('`Activity::fetchByCallSid', () => {
    it('fetches an activity via callSid', async () => {
      const testActivity = testDataGenerator.getActivity({fieldsToGenerate: ['callSid']});
      await testActivity.put();
      const testCallSid = testActivity.callSid!;
  
      const returnedActivity = await Activity.fetchByCallSid(testCallSid);

      await waitForGSIUpdates();
  
      expect(returnedActivity!.pdid).toBe(testActivity.pdid);
      expect(returnedActivity).toEqual(expect.objectContaining(testActivity));
      expect(testActivity).toEqual(expect.objectContaining(returnedActivity));
    });

    it('returns null if the activity is not found', async () => {
      const returnedActivity = await Activity.fetchByCallSid('nonexistentCallSid')
  
      expect(returnedActivity).toBe(null);
    });
  });

  describe('`Activity#put`', () => {
    it('adds the activity to the activity DynamoDB table', async () => {
      const testActivity = testDataGenerator.getActivity();
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await testActivity.put();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelActivity, ['pk', 'sk']), {
        ...lowLevelActivity,
      });
    });
  });

  describe('`Activity#hardDelete`', () => {
    it('removes all items associated with the activity', async () => {
      const testActivity = testDataGenerator.getActivity();
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await testActivity.put();

      await waitForGSIUpdates();

      await testActivity.hardDelete();

      await waitForGSIUpdates();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelActivity, ['pk', 'sk']));
    });
  });

  describe('`Activity#updateFields`', () => {
    it('updates the fields passed in and only those passed in', async () => {
      const testActivityInterface1 = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivityInterface2 = { ...testActivityInterface1, body: generateCallSidBody('testCallSid') };
      const testActivity1 = new Activity(testActivityInterface1);
      const testActivity2 = new Activity(testActivityInterface2);
      const lowLevelAcivity = buildLowLevelActivity(testActivity2);
      const expectedActivity = new Activity({ ...testActivityInterface1, callSid: 'testCallSid', body: generateCallSidBody('testCallSid'), lastUpdatedField: MutableActivityProperties.callSid });
      const expectedLowLevel = buildLowLevelActivity(expectedActivity);

      await testActivity1.put();

      await waitForGSIUpdates();

      await testActivity2.updateFields([MutableActivityProperties.body, MutableActivityProperties.callSid]);

      await waitForGSIUpdates();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelAcivity, ['pk', 'sk']), expectedLowLevel);
    });
  });
  
  describe('`Activity#updateBody`', () => {
    it('updates the DDB activity item with the new body and lastUpdatedField', async () => {
      const testActivity = testDataGenerator.getActivity({overrides: {lastUpdatedField: MutableActivityProperties.isDeleted}});
      const lowLevelAcivity = buildLowLevelActivity(testActivity)

      await testActivity.put();

      await waitForGSIUpdates();
      
      await testActivity.updateBody('test body');

      const keys = _.pick(lowLevelAcivity, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelAcivity, lastUpdatedField: 'body', body: 'test body' });
    });

    it('updates the callSid if the body`s callSid changes', async () => {
      const testActivity = testDataGenerator.getActivity({overrides: {lastUpdatedField: MutableActivityProperties.isDeleted}});
      const lowLevelAcivity = buildLowLevelActivity(testActivity)

      await testActivity.put();

      await waitForGSIUpdates();

      const testCallSid = 'test_call_sid';

      const testBody = generateCallSidBody(testCallSid);
      
      await testActivity.updateBody(testBody);

      await waitForGSIUpdates();

      const keys = _.pick(lowLevelAcivity, ['pk', 'sk']);

      const gsiKey = buildActivityIndex1Keys(testCallSid);

      const newProperties = {
        lastUpdatedField: 'callSid',
        body: testBody,
        callSid: testCallSid,
        gsi1_pk: gsiKey,
        gsi1_sk: gsiKey,
      }

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelAcivity, ...newProperties});
    });

    it('does not update the callSid if the body`s callSid does not change', async () => {
      const testActivity = testDataGenerator.getActivity({overrides: {lastUpdatedField: MutableActivityProperties.isDeleted}, fieldsToGenerate: ['callSid']});
      const lowLevelAcivity = buildLowLevelActivity(testActivity)

      await testActivity.put();

      await waitForGSIUpdates();

      const testBody = testActivity.body + " new addition to body"
      
      await testActivity.updateBody(testBody);

      await waitForGSIUpdates();

      const keys = _.pick(lowLevelAcivity, ['pk', 'sk']);

      const newProperties = {
        lastUpdatedField: 'body',
        body: testBody,
      }

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelAcivity, ...newProperties});
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});