import { AWS_REGION, ddb } from '@svc/config';
import Client from '@svc/lib/models/Client';
import Project from '@svc/lib/models/Project';
import { buildLowLevelProjectDetails } from '@svc/lib/ddbRepos/projects-repo';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import _ from 'lodash';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import { PipedriveDealFieldKeys, PipedriveDealStatus } from '@improovy/types';
import { MutableProjectProperties, ProjectSalesStage } from '@improovy/types';
import { waitForGSIUpdates } from '@tests/utils/wait';

const testDataGenerator = new TestDataGenerator();

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.projectsTable,
};

describe('`Project`', () => {
  describe('`Project#put`', () => {
    it('adds the project to the projects DynamoDB table', async () => {
      const testProject = testDataGenerator.getProject();
      const lowLevelProject = buildLowLevelProjectDetails(testProject);

      testProject.put();

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelProject, ['pk', 'sk']), {
        ...lowLevelProject,
      });
    });
  });

  describe('`Project::fetch', () => {
    it('fetches a project via uuid', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.put();
      const testUuid = testProject.uuid;
  
      const returnedProject = await Project.fetch(testUuid);
  
      expect(returnedProject.uuid).toBe(testUuid);
      expect(returnedProject).toEqual(expect.objectContaining(testProject));
      expect(testProject).toEqual(expect.objectContaining(returnedProject));
    });
  });

  describe('`Project::fetchByPdid', () => {
    it('fetches a project via Pdid', async () => {
      const testProject = testDataGenerator.getProject();
      await testProject.put();
      const testPdid = testProject.pdid;
  
      const returnedProject = await Project.fetchByPdid(testPdid!);
  
      expect(returnedProject!.uuid).toBe(testProject.uuid);
      expect(returnedProject).toEqual(expect.objectContaining(testProject));
      expect(testProject).toEqual(expect.objectContaining(returnedProject));
    });

    it('returns null if the project does not exist', async () => {
      const returnedProject = await Project.fetchByPdid('nonexistent pdid');
  
      expect(returnedProject).toBeNull();
    });
  });

  describe('`Project#updateFields`', () => {
    it('updates the fields passed in and only those passed in', async () => {
      const testProjectInterface1 = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProjectInterface2 = { ...testProjectInterface1, gclid: 'testing2', isDeleted: true, pdid: '5' };
      const testProject1 = new Project(testProjectInterface1);
      const testProject2 = new Project(testProjectInterface2);
      const lowLevelProject = buildLowLevelProjectDetails(testProject2);
      const expectedProject = new Project({ ...testProjectInterface2, pdid: testProjectInterface1.pdid, lastUpdatedField: MutableProjectProperties.isDeleted });
      const expectedLowLevel = buildLowLevelProjectDetails(expectedProject);

      await testProject1.put();

      await testProject2.updateFields([MutableProjectProperties.gclid, MutableProjectProperties.isDeleted]);

      expect(awsTestOptions).toHaveItem(_.pick(lowLevelProject, ['pk', 'sk']), expectedLowLevel);
    });
  });

  describe('`Project::fromPipedrive`', () => {
    it('correctly maps in a person from the pipeDrive object', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();

      await waitForGSIUpdates();

      const gclid = 'test-gclid';
      const uuid = 'test-uuid';
      const gaTransactionID = 'test-ga-transaction-id';
      const pdid = 1234;
      const date = '2022-04-20 22:27:36';
      const createdAt = 1650493656000; // convert "date" to milliseconds since epoch
      const clientUuid = testClient.uuid;
      const clientPdid = testClient.pdid!;
      const clientCostStr = '1000';
      const clientCostNum = 1000;
      
      const deal = testDataGenerator.getPipedriveDeal({
        [PipedriveDealFieldKeys.pdid]: pdid,
        [PipedriveDealFieldKeys.gclid]: gclid,
        [PipedriveDealFieldKeys.uuid]: uuid,
        [PipedriveDealFieldKeys.createdAt]: date,
        [PipedriveDealFieldKeys.gaTransactionID]: gaTransactionID,
        [PipedriveDealFieldKeys.clientPdid]: clientPdid,
        [PipedriveDealFieldKeys.pipeline]: 1,
        [PipedriveDealFieldKeys.stage]: 6,
        [PipedriveDealFieldKeys.clientCost]: clientCostStr,
        [PipedriveDealFieldKeys.projectStatus]: PipedriveDealStatus.open,
      });

      const project = await Project.fromPipedrive(deal);

      if (!project) {
        expect(project).toBeTruthy();
        return;
      }

      expect(project.gclid).toBe(gclid);
      expect(project.createdAt).toBe(createdAt);
      expect(project.clientUuid).toBe(clientUuid);
      expect(project.gaTransactionID).toBe(gaTransactionID);
      expect(project.isDeleted).toBe(false);
      expect(project.uuid).toBe(uuid);
      expect(project.projectStatus).toBe('OPEN');
      expect(project.salesStage).toBe('BOOKED');
      expect(project.clientCost).toBe(clientCostNum);
      expect(project.pdid).toBe(`${pdid}`); // pdid is converted to a string
    });
  });

  describe('`Project#hardDelete`', () => {
    it('removes all items associated with the project', async () => {
      const testProject = testDataGenerator.getProject();
      const lowLevelProject = buildLowLevelProjectDetails(testProject);

      await testProject.put();

      await testProject.hardDelete();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelProject, ['pk', 'sk']));

      const fetchedItems = await queryByPK(ddb.projectsTable, testProject.uuid);
      
      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`Project#updateGclid`', () => {
    it('updates the DDB project details item with the new gclid and lastUpdatedField', async () => {
      const testProject = testDataGenerator.getProject();
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.updateGclid('test_gclid');

      const keys = _.pick(details, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'gclid', gclid: 'test_gclid' });
    });

    it('removes the gclid property if it receives a falsy value', async () => {
      const testProject = testDataGenerator.getProject({ fieldsToGenerate: ['gclid'] });
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.updateGclid(undefined);

      const keys = _.pick(details, ['pk', 'sk']);

      delete details.gclid;

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'gclid' });
    });
  });

  describe('`Project#updateClientUuid`', () => {
    it('updates the DDB project details item with the new clientUuid and lastUpdatedField', async () => {
      const testProject = testDataGenerator.getProject();
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.updateClientUuid('test_clientUuid');

      const keys = _.pick(details, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'clientUuid', clientUuid: 'test_clientUuid' });
    });

    it('throws an error if it receives a falsy value', async () => {
      const testProject = testDataGenerator.getProject();

      expect(() => testProject.updateClientUuid('')).toThrowError();
    });
  });

  describe('`Project#updateSalesStage`', () => {
    it('updates the DDB project details item with the new salesStage and lastUpdatedField', async () => {
      const testProject = testDataGenerator.getProject();
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.updateSalesStage(ProjectSalesStage.ATTEMPTED);

      const keys = _.pick(details, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'salesStage', salesStage: 'ATTEMPTED'});
    });

    it('throws an error if it receives a falsy value', async () => {
      const testProject = testDataGenerator.getProject();

      // @ts-expect-error
      expect(() => testProject.updateSalesStage(null)).toThrowError();
    });
  });

  describe('`Project#removeGclid`', () => {
    it('removes the gclid property from the ddb item and updates lastUpdatedField', async () => {
      const testProject = testDataGenerator.getProject({ fieldsToGenerate: ['gclid'] });
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.removeGclid();

      const keys = _.pick(details, ['pk', 'sk']);

      delete details.gclid;

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'gclid' });
    });
  });

  describe('`Project#softDelete`', () => {
    it('updates the DDB project details item with isDeleted set to true and lastUpdatedField', async () => {
      const testProject = testDataGenerator.getProject();
      const details = buildLowLevelProjectDetails(testProject);

      await testProject.put();
      
      await testProject.softDelete();

      const keys = _.pick(details, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.projectsTable }).toHaveItem(keys, { ...details, lastUpdatedField: 'isDeleted', isDeleted: true });
    });
  });

  describe('`Project#fetchClient`', () => {
    it('fetches the client the project belongs to', async () => {
      const testClientInterface = testDataGenerator.getClientInterface();
      const testClient = new Client(testClientInterface);
      const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });

      await testClient.put();
      await testProject.put();
      
      const receivedClient = await testProject.fetchClient();

      if (!receivedClient) {
        expect(receivedClient).toBeTruthy();
        return;
      }

      expect(receivedClient.uuid).toBe(testClient.uuid);
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});