import { AWS_REGION, ddb } from '@svc/config';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import _ from 'lodash';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';
import {
  buildLowLevelCallRail,
  putCallRail,
  hardDeleteCallRail,
  updateCallRail,
  fetchCallRailInterface,
} from '@svc/lib/ddbRepos/callRail-repo';
import CallRail from '@svc/lib/models/CallRail';
import { ChangeSource } from '@improovy/types';
import { MutableCallRailProperties } from '@improovy/types';

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.callRailTable,
};

const testDataGenerator = new TestDataGenerator();

describe('`CallRail` repo', () => {
  describe('`putCallRail`', () => {
    it('adds the callRail instance to the callRail DynamoDB table', async () => {
      const testCallRail = testDataGenerator.getCallRail();
      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);

      await putCallRail(testCallRail);

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelCallRail, ['pk', 'sk']), {
        ...lowLevelCallRail,
      });
    });
  });

  describe('`fetchCallRailInterface`', () => {
    it('returns the highlevel callRail interface matching the requested phoneNumber', async () => {
      const testCallRailInterface1 = testDataGenerator.getCallRailInterface({ fieldsToGenerate: ['all'], overrides: { phoneNumber: '2625555286' } });
      const testCallRail1 = new CallRail(testCallRailInterface1);

      await testCallRail1.put();

      const receivedInterface = await fetchCallRailInterface(testCallRail1.phoneNumber);

      expect(receivedInterface).toStrictEqual(testCallRailInterface1);
    });
    
    it('returns null if the interface is not found', async () => {
      const receivedInterface = await fetchCallRailInterface('nonexistentPhoneNumber');

      expect(receivedInterface).toBeNull();
    });
  });

  describe('`hardDeleteCallRail`', () => {
    it('removes all items associated with the callRail', async () => {
      const testCallRail = testDataGenerator.getCallRail();
      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);

      await putCallRail(testCallRail);

      await hardDeleteCallRail(testCallRail);

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelCallRail, ['pk', 'sk']));

      const fetchedItems = await queryByPK(ddb.callRailTable, testCallRail.phoneNumber);

      expect(fetchedItems).toHaveLength(0);
    });
  });

  describe('`updateCallRail`', () => {
    it('updates the activity lastUpdatedField, field referenced by lastUpdatedField, and changeSource', async () => {
      const testCallRailInterface = testDataGenerator.getCallRailInterface({ fieldsToGenerate: ['all'] });
      const testCallRail = new CallRail(testCallRailInterface);
      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);

      await testCallRail.put();
      
      testCallRailInterface.changeSource = ChangeSource.aws;
      testCallRailInterface.lastUpdatedField= MutableCallRailProperties.isAttributed;
      testCallRailInterface.isAttributed = true;

      // @ts-expect-error
      await updateCallRail(testCallRailInterface);

      const keys = _.pick(lowLevelCallRail, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelCallRail, lastUpdatedField: MutableCallRailProperties.isAttributed, isAttributed: true, changeSource: ChangeSource.aws });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});