import { AWS_REGION, ddb } from '@svc/config';
import Client from '@svc/lib/models/Client';
import { buildLowLevelClient, buildLowLevelContactMethods } from '@svc/lib/ddbRepos/clients-repo-utils';
import { buildLowLevelProjectDetails, buildProjectGSI2_PK } from '@svc/lib/ddbRepos/projects-repo';
import { queryByPK, removeProperty, updateItem, queryByGSI } from '@svc/lib/ddbRepos/repo-utils';
import { ContactType, MutableClientProperties } from '@improovy/types';
import _ from 'lodash';
import TestDataGenerator, { generateTestEmail } from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import { buildLowLevelActivity } from '@svc/lib/ddbRepos/activities-repo';
import { MutableActivityProperties } from '@improovy/types';
import Activity from '@svc/lib/models/Activity';
import { ChangeSource } from '@improovy/types';
import generateNewPhoneNumber from '@tests/utils/generateNewPhoneNumber';

const testDataGenerator = new TestDataGenerator();

describe('`repo-utils`', () => {
  describe('`queryByPK`', () => {
    it.concurrent('returns the low level interfaces for the queried entity', async () => {
      const testClientInterface = testDataGenerator.getClientInterface();

      const testClient = new Client(testClientInterface);
      await testClient.put();
      const lowLevelInterface = buildLowLevelClient(testClient);

      const results = await queryByPK(ddb.clientsTable, testClient.uuid);
      
      const expected = [...lowLevelInterface.contactMethods, lowLevelInterface.profile];

      expect(results).toEqual(expect.arrayContaining(expected));
      expect(expected).toEqual(expect.arrayContaining(results));
    });

    it.concurrent('optionally queries by a sortkey', async () => {
      const contactEmail = {
        contactType: ContactType.EMAIL,
        contactAddress: generateTestEmail(),
      }
      const contactPhone = {
        contactType: ContactType.PHONE,
        contactAddress: generateNewPhoneNumber(),
      }
      const testClientInterface = testDataGenerator.getClientInterface({overrides: {
        contactMethods: [
          contactEmail,
          contactPhone,
        ]
      }});

      const testClient = new Client(testClientInterface);
      await testClient.put();
      
      const lowLevelInterface = buildLowLevelClient(testClient);

      const results = await queryByPK(ddb.clientsTable, testClient.uuid, 'Contact');
      const results2 = await queryByPK(ddb.clientsTable, testClient.uuid, 'Contact#PHONE')
      
      expect(results).toHaveLength(2)
      expect(results).toContainEqual(expect.objectContaining(contactEmail));
      expect(results).toContainEqual(expect.objectContaining(contactPhone));

      expect(results2).toHaveLength(1)
      expect(results2[0]).toStrictEqual(expect.objectContaining(contactPhone))
    });

    it.concurrent('returns an empty array if no results are found', async () => {
      const results = await queryByPK(ddb.clientsTable, 'fake_uuid');
  
      expect(results).toEqual([]);
    });
  });

  describe('`updateItem`', () => {
    it.concurrent('updates the item in the respective table based on the item`s lastUpdatedField', async () => {
      const testClient = testDataGenerator.getClient();
      const { profile } = buildLowLevelClient(testClient);
    
      await testClient.put();
      
      profile.gclid = 'test_gclid';
      profile.lastUpdatedField = MutableClientProperties.gclid;

      await updateItem(profile, ddb.clientsTable);

      const keys = _.pick(profile, ['pk', 'sk']);
      await expect({ region: AWS_REGION, table: ddb.clientsTable }).toHaveItem(keys, { ...profile, lastUpdatedField: 'gclid', gclid: 'test_gclid' });
    });

    it.concurrent('updates the entities changeSource if it exists', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({fieldsToGenerate: ['callSid']});
      const testActivity = new Activity(testActivityInterface);

      await testActivity.put();

      await waitForGSIUpdates();

      const updatedActivityInterface = {...testActivityInterface, isDeleted: true, lastUpdatedField: MutableActivityProperties.isDeleted, ChangeSource: ChangeSource.pipedrive }

      const updatedActivity = new Activity(updatedActivityInterface);

      const updatedLowLevelActivity = buildLowLevelActivity(updatedActivity);

      await updateItem(updatedLowLevelActivity, ddb.activitiesTable);

      await waitForGSIUpdates();

      const keys = _.pick(updatedLowLevelActivity, ['pk', 'sk']);

      await expect({ region: AWS_REGION, table: ddb.activitiesTable }).toHaveItem(keys, updatedLowLevelActivity);
    });

    it.concurrent('removes the property from the item if the value is undefined or null', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({fieldsToGenerate: ['callSid']});
      const testActivity = new Activity(testActivityInterface);

      await testActivity.put();

      await waitForGSIUpdates();

      const updatedActivityInterface = {...testActivityInterface, lastUpdatedField: MutableActivityProperties.callSid }
      delete updatedActivityInterface.callSid;
      delete updatedActivityInterface.body;

      const updatedActivity = new Activity(updatedActivityInterface);

      const updatedLowLevelActivity = buildLowLevelActivity(updatedActivity);

      const secondaryIndices = {
        indexNum: 1
      }

      await updateItem(updatedLowLevelActivity, ddb.activitiesTable, secondaryIndices);

      await waitForGSIUpdates();

      const keys = _.pick(updatedLowLevelActivity, ['pk', 'sk']);
      await expect({ region: AWS_REGION, table: ddb.activitiesTable }).toHaveItem(keys, { ...updatedLowLevelActivity, body: testActivity.body});
    });

    it.concurrent('optionally takes in a secondaryIndices to update object and updates those indices', async () => {
      const testActivity = testDataGenerator.getActivity();
      const testLowLevelActivity = buildLowLevelActivity(testActivity);
    
      await testActivity.put();
      
      testLowLevelActivity.callSid = 'testCallSid';
      testLowLevelActivity.lastUpdatedField = MutableActivityProperties.callSid;

      const key1 = 'testKey1';
      const key2 = 'testKey2';

      const secondaryIndices = {
        indexNum: 1,
        pk: key1,
        sk: key2,
      }

      await updateItem(testLowLevelActivity, ddb.activitiesTable, secondaryIndices);

      await waitForGSIUpdates();

      const keys = _.pick(testLowLevelActivity, ['pk', 'sk']);
      await expect({ region: AWS_REGION, table: ddb.activitiesTable }).toHaveItem(keys, { ...testLowLevelActivity, lastUpdatedField: 'callSid', callSid: 'testCallSid', gsi1_pk: key1, gsi1_sk: key2 });
    });
  });

  describe('`removeProperty`', () => {
    it.concurrent('removes the property of the respective item based on the item`s lastUpdatedField', async () => {
      const testClient = testDataGenerator.getClient({ fieldsToGenerate: ['gclid'] });
      const { profile } = buildLowLevelClient(testClient);
      await testClient.put();
      profile.lastUpdatedField = MutableClientProperties.gclid;

      // object reassignment needed to reassure TypeScript that the lastUpdatedField exists
      await removeProperty({...profile, lastUpdatedField: profile.lastUpdatedField}, ddb.clientsTable);

      const keys = _.pick(profile, ['pk', 'sk']);
      await expect({ region: AWS_REGION, table: ddb.clientsTable }).not.toHaveItem(keys, { ...profile });

      delete profile.gclid;
      await expect({ region: AWS_REGION, table: ddb.clientsTable }).toHaveItem(keys, { ...profile });

    });
  });

  describe('queryByGSI', () => {
    it.concurrent('fetches all items under a given partition key for the given GSI number for a table', async () => {
      const testClient = testDataGenerator.getClient();
      await testClient.put();
      
      const testProjects = [];
      for (let i = 0; i < 2; i++) {
        const testProject = testDataGenerator.getProject({ overrides: { clientUuid: testClient.uuid } });
        await testProject.put();
        testProjects.push(buildLowLevelProjectDetails(testProject));
      }
      
      await waitForGSIUpdates();

      const receivedProjects = await queryByGSI(ddb.projectsTable, 2, buildProjectGSI2_PK({ clientUuid: testClient.uuid }));

      expect(receivedProjects).toHaveLength(testProjects.length);
      expect(receivedProjects).toEqual(expect.arrayContaining(testProjects));
    });
    it.todo('optionally takes in a string and returns all items under the partition key whose sort key starts with the string');
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});