import Activity from '@svc/lib/models/Activity';
import TestDataGenerator, { generateCallSidBody } from '@tests/utils/testDataGenerator';
import { ActivityInterface, MutableActivityProperties } from '@improovy/types';
import { PipedriveActivityFieldKeys } from '@improovy/types';
import { ChangeSource } from '@improovy/types';

const testDataGenerator = new TestDataGenerator(false);

describe('`Activity` model', () => {
  it('correctly maps the fields when created with a ActivityInterface', () => {
    const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
    // if this assertion fails, it's likely because the ActivityInterface was changed.
    // Update the expected number of properties and add tests to make sure they are
    // mapped correctly
    expect(Object.keys(testActivityInterface)).toHaveLength(9);

    const testActivity = new Activity(testActivityInterface);

    for (let key in testActivityInterface) {
      expect(testActivity[key as keyof Activity]).toBe(testActivityInterface[key as keyof ActivityInterface]);
    }
  });

  it('correctly assigns default values when values are missing', () => {
    const testActivityInterface = testDataGenerator.getActivityInterface({fieldsToGenerate: ['all']});

    const callSid = testActivityInterface.callSid;

    delete testActivityInterface.callSid;
    
    //todo: add variations to test for other versions of callSid format

    const testActivity = new Activity(testActivityInterface)

    expect(testActivity.callSid).toBeTruthy();
    expect(testActivity.callSid).toBe(callSid);
  });

  describe('`Activity::fromPipedrive`', () => {
    it('correctly maps in an activity from a pipedrive activity', () => {
      const callSid = "sampleCallSid"
      const testPipedriveActivity = testDataGenerator.getPipedriveActivity({bodyOptions: {type: 'callSid', override: callSid}})

      const pdid = testPipedriveActivity[PipedriveActivityFieldKeys.pdid];
      const body = testPipedriveActivity[PipedriveActivityFieldKeys.body];
      const projectPdid = testPipedriveActivity[PipedriveActivityFieldKeys.projectPdid];
      const activityType = testPipedriveActivity[PipedriveActivityFieldKeys.activityType];

      const testActivity = Activity.fromPipedrive(testPipedriveActivity);

      expect(testActivity.pdid).toBe(pdid);
      expect(testActivity.callSid).toBe(callSid);
      expect(testActivity.body).toBe(body);
      expect(testActivity.projectPdid).toBe(projectPdid);
      expect(testActivity.activityType).toBe(activityType);
    })
  })

  describe('`Activity::differingFields`', () => {
    it('returns the fields that are different between two activities', async () => {
      const testActivityInterface1 = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivityInterface2 = { ...testActivityInterface1, body: generateCallSidBody(), isDeleted: true };
      const testActivity1 = new Activity(testActivityInterface1);
      const testActivity2 = new Activity(testActivityInterface2);

      const diff = Activity.differingFields(testActivity1, testActivity2);

      expect(diff).toHaveLength(3);
      expect(diff).toEqual(expect.arrayContaining(['body', 'callSid', 'isDeleted']));
    });

    it('does not return `lastUpdatedField` even if it is different', async () => {
      const testActivityInterface1 = testDataGenerator.getActivityInterface({ overrides: { lastUpdatedField: MutableActivityProperties.callSid} });
      const testActivityInterface2 = { ...testActivityInterface1, lastUpdatedField: MutableActivityProperties.body };
      const testProject1 = new Activity(testActivityInterface1);
      const testProject2 = new Activity(testActivityInterface2);

      const diff = Activity.differingFields(testProject1, testProject2);

      expect(diff).toHaveLength(0);
    });

    it('does not return `changeSource` even if it is different', async () => {
      const testActivityInterface1 = testDataGenerator.getActivityInterface();
      const testActivityInterface2 = { ...testActivityInterface1, changeSource: ChangeSource.aws };
      const testProject1 = new Activity(testActivityInterface1);
      const testProject2 = new Activity(testActivityInterface2);

      const diff = Activity.differingFields(testProject1, testProject2);

      expect(diff).toHaveLength(0);
    });
  });

  describe('`Activity#parseBody`', () => {
    it.todo('correctly identifies the callSid and assigns it to the client (see above test)')
  })
});