import CallRail from '@svc/lib/models/CallRail';
import {
  assembleLowLevelCallRailInterface,
  assembleCallRail,
  buildLowLevelCallRail,
  buildCallRailPutTransaction,
} from '@svc/lib/ddbRepos/callRail-repo';
import { LowLevelCallRail, CallRailInterface } from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/config', () => {
  const original = jest.requireActual('@svc/config');

  return {
    __esModule: true,
    ddb: {
      ...original.ddb,
      documentClient: {
        send: jest.fn(),
      },
    },
  };
});

jest.mock('@svc/lib/ddbRepos/repo-utils', () => {
  return {
    __esModule: true,
    queryByPK: jest.fn().mockImplementation(() => ([])),
  };
});

describe('`CallRail` repo', () => {
  beforeEach(() => jest.clearAllMocks());

  
  describe('`buildLowLevelCallRail`', () => {
    it('has 9 keys when supplied with no optional fields', () => {
      const testCallRail = testDataGenerator.getCallRail();
      const lowLevel = buildLowLevelCallRail(testCallRail);
      expect(Object.keys(lowLevel)).toHaveLength(9);
    });
    
    it('has 9 keys when supplied with all optional fields', () => {
      const testCallRail = testDataGenerator.getCallRail({ fieldsToGenerate: ['all'] });
      const lowLevel = buildLowLevelCallRail(testCallRail);
      expect(Object.keys(lowLevel)).toHaveLength(10);
    });
    
    it('correctly inherits the properties from the high-level interface', () => {
      const testCallRailInterface = testDataGenerator.getCallRailInterface({ fieldsToGenerate: ['all'] });
      const testCallRail = testDataGenerator.getCallRail({ overrides: testCallRailInterface });
      const details = buildLowLevelCallRail(testCallRail);

      for (let key in testCallRailInterface) {
        expect(details[key as keyof LowLevelCallRail]).toBe(testCallRailInterface[key as keyof CallRailInterface]);
      }
    });

    it('correctly assigns the new properties specific to the low-level interface', () => {
      const testCallRailInterface = testDataGenerator.getCallRailInterface({ fieldsToGenerate: ['all'] });
      const testCallRail = testDataGenerator.getCallRail({ overrides: testCallRailInterface });
      const lowLevel = buildLowLevelCallRail(testCallRail);

      expect(lowLevel.pk).toBe(testCallRail.phoneNumber);
      expect(lowLevel.sk).toBe(`CallRail#gclid#${testCallRail.gclid}`);
      expect(lowLevel.type).toBe('CallRailGclid');
    });
  });

  describe('`buildCallRailPutTransaction`', () => {
    const testCallRail = testDataGenerator.getCallRail();
    const lowLevelCallRail = buildLowLevelCallRail(testCallRail);
    const transaction = buildCallRailPutTransaction(lowLevelCallRail);
    const TransactItems = transaction.TransactItems!;

    it('contains only put items', () => {
      TransactItems.every(item => expect(item).toHaveProperty('Put'));
    });
    describe('The CallRail gclid put item', () => {
      it('is the first item', () => {
        expect(TransactItems[0].Put?.Item?.type).toBe('CallRailGclid');
      });

      it('refers to the callRail table', () => {
        expect(TransactItems[0].Put?.TableName).toContain('callRail');
      });

      it('contains the callRail gclid as the item', () => {
        expect(TransactItems[0].Put?.Item).toBe(lowLevelCallRail);
      });
    });
  });

  describe('`assembleCallRail`', () => {
    it('creates a high level callRail profile from a low level one', () => {
      const testCallRailInterface = testDataGenerator.getCallRailInterface({ fieldsToGenerate: ['all'] });
      const testCallRail = new CallRail(testCallRailInterface);

      const lowLevelCallRail = buildLowLevelCallRail(testCallRail);
      const receivedResult = assembleCallRail(lowLevelCallRail);

      expect(receivedResult).toEqual(expect.objectContaining(testCallRailInterface));
      expect(testCallRailInterface).toEqual(expect.objectContaining(receivedResult));
    });
  });

  describe('`assembleLowLevelCallRailInterface`', () => {
    it('creates a LowLevelCallRail object from its respective DDB item', () => {
      const testCallRail = testDataGenerator.getCallRail({ fieldsToGenerate: ['all'] });
      const testLowLevelCallRail = buildLowLevelCallRail(testCallRail);

      const resultLowLevelCallRail = assembleLowLevelCallRailInterface(testLowLevelCallRail);

      expect(resultLowLevelCallRail).toEqual(expect.objectContaining(testLowLevelCallRail));
      expect(testLowLevelCallRail).toEqual(expect.objectContaining(resultLowLevelCallRail));
    });
  });
});