import { ddb } from '@svc/config';
import { TransactWriteCommand } from '@aws-sdk/lib-dynamodb';
import { fetchClientInterface, putClient } from '@svc/lib/ddbRepos/clients-repo';
import {
  ClientContactMethod,
  ClientInterface,
  LowLevelClientContactMethod,
  LowLevelClientProfile,
} from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';
import {
  assembleClientContactMethod,
  assembleClientProfile,
  assembleLowLevelClientInterface,
  buildClientPutTransaction,
  buildLowLevelClient,
  buildLowLevelClientProfile,
  buildLowLevelContactMethods,
} from '@svc/lib/ddbRepos/clients-repo-utils';
import Client from '@svc/lib/models/Client';
import { queryByPK } from '@svc/lib/ddbRepos/repo-utils';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/config', () => {
  const original = jest.requireActual('@svc/config');

  return {
    __esModule: true,
    ddb: {
      ...original.ddb,
      documentClient: {
        send: jest.fn(),
      },
    },
  };
});

jest.mock('@svc/lib/ddbRepos/repo-utils', () => {
  return {
    __esModule: true,
    queryByPK: jest.fn().mockImplementation(() => ([])),
  };
});

describe('`Client` repo', () => {
  beforeEach(() => jest.clearAllMocks());

  
  describe('`buildLowLevelClient`', () => {
    const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
    const testClient = testDataGenerator.getClient({ overrides: testClientInterface });
    const lowLevelClient = buildLowLevelClient(testClient);
    
    it('it contains two objects, `profile` and `contactMethods`', () => {
      expect(Object.keys(lowLevelClient)).toHaveLength(2);
      expect(lowLevelClient).toHaveProperty('profile');
      expect(lowLevelClient).toHaveProperty('contactMethods');
    });

    const { profile, contactMethods } = lowLevelClient;

    describe('`lowLevelClient.profile`', () => {
      it('has 15 keys', () => {
        expect(Object.keys(profile)).toHaveLength(15);
      });

      it('correctly inherits the properties from the high-level interface', () => {
        for (let key in testClientInterface) {
          if (key === 'contactMethods') continue;
          expect(profile[key as keyof LowLevelClientProfile]).toBe(testClientInterface[key as keyof ClientInterface]);
        }
      });

      it('correctly assigns the new properties specific to the low-level interface', () => {
        expect(profile.pk).toBe(testClient.uuid);
        expect(profile.sk).toBe('Profile');
        expect(profile.gsi2_pk).toBe(`PDID#${testClient.pdid}`);
        expect(profile.gsi2_sk).toBe('Profile');
        expect(profile.type).toBe('ClientProfile');
        expect(profile.pipedriveName).toBe(testClientInterface.pipedriveName);
        expect(profile.changeSource).toBe(testClientInterface.changeSource);
      });
    });

    describe('`lowLevelClient.contactMethods`', () => {
      it('has the correct number of contact methods', () => {
        expect(contactMethods).toHaveLength(testClient.contactMethods.length);
      });
      describe('`for each contact method:`', () => {
        it('correctly inherits the properties from the high-level interface', () => {
          for (let i = 0; i < testClient.contactMethods.length; i++) {
            const testContactMethod = testClient.contactMethods[i];
            const contactMethod = contactMethods[i];

            for (let key in testContactMethod) {
              expect(contactMethod[key as keyof LowLevelClientContactMethod]).toBe(testContactMethod[key as keyof ClientContactMethod]);
            }
          }
        });

        it('correctly assigns the new properties specific to the low-level interface', () => {
          for (let i = 0; i < testClient.contactMethods.length; i++) {
            const testContactMethod = testClient.contactMethods[i];
            const contactMethod = contactMethods[i];

            const compositeKey = `Contact#${testContactMethod.contactType}#${testContactMethod.contactAddress}`;

            expect(contactMethod.pk).toBe(testClient.uuid);
            expect(contactMethod.sk).toBe(compositeKey);
            expect(contactMethod.gsi1_pk).toBe(compositeKey);
            expect(contactMethod.gsi1_sk).toBe(testClient.uuid);
            expect(contactMethod.gsi2_pk).toBe(`PDID#${testClient.pdid}`);
            expect(contactMethod.gsi2_sk).toBe(compositeKey);
            expect(contactMethod.type).toBe('ClientContactMethod');
            expect(contactMethod.isDeleted).toBe(false);
          }
        });
      });
    });
  });

  describe('`buildClientPutTransaction`', () => {
    const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
    const testClient = testDataGenerator.getClient({ overrides: testClientInterface });
    const lowLevelClient = buildLowLevelClient(testClient);
    
    const transaction = buildClientPutTransaction(lowLevelClient);

    const TransactItems = transaction.TransactItems!;

    it('contains only put items', () => {
      TransactItems.every(item => expect(item).toHaveProperty('Put'));
    });
    describe('The client profile put item', () => {
      it('is the first item', () => {
        expect(TransactItems[0].Put?.Item?.type).toBe('ClientProfile');
      });

      it('refers to the clients table', () => {
        expect(TransactItems[0].Put?.TableName).toContain('clients');
      });

      it('contains the client profile as the item', () => {
        expect(TransactItems[0].Put?.Item).toBe(lowLevelClient.profile);
      });
    });

    describe('The client contact method put items', () => {
      it('has the correct amount', () => {
        expect(TransactItems).toHaveLength(testClient.contactMethods.length + 1);
      });

      describe('for each item', () => {
        it('refers to the clients table', () => {
          for (let i = 1; i < TransactItems.length; i++) {
            expect(TransactItems[i].Put?.TableName).toContain('clients');
          }
        });

        it('contains the contact method as the item', () => {
          for (let i = 1; i < TransactItems.length; i++) {
            expect(TransactItems[i].Put?.Item).toBe(lowLevelClient.contactMethods[i - 1]);
          }
        });
      });
    });
  });

  describe('`putClient`', () => {
    it('invokes ddb.send with a generated TransactWriteCommand', () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = testDataGenerator.getClient({ overrides: testClientInterface });
      const lowLevelClient = buildLowLevelClient(testClient);
      
      const transaction = buildClientPutTransaction(lowLevelClient);
      const mockSend = ddb.documentClient.send;
      
      expect(mockSend).not.toHaveBeenCalled();      

      putClient(testClient);

      expect(mockSend).toHaveBeenCalledTimes(1);      
      
      // @ts-ignore
      const mockSendInput = mockSend.mock.calls[0][0];
      const expectedInput = new TransactWriteCommand(transaction);

      // there's probably a better way of doing this but everything I've tried causes the test
      // to fail with a "Received: serializes to the same string" error
      expect(JSON.stringify(mockSendInput)).toBe(JSON.stringify(expectedInput));
    });
  });

  describe('`assembleClientContactMethod`', () => {
    it('creates a high level client contact method interface from a low level one', () => {
      const testClientInterface = testDataGenerator.getClientInterface();
      const expectedResults = testClientInterface.contactMethods;
      const testClient = new Client(testClientInterface);
      const lowLevelInterfaces = buildLowLevelContactMethods(testClient);
      const receivedResults = lowLevelInterfaces.map(assembleClientContactMethod);

      expect(receivedResults).toHaveLength(expectedResults.length);
      for (let i = 0; i < receivedResults.length; i++) {
        expect(JSON.stringify(receivedResults[i])).toBe(JSON.stringify(expectedResults[i]));
      }
    });
  });

  describe('`assembleClientProfile`', () => {
    it('creates a high level client profile from a low level one', () => {
      const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
      const testClient = new Client(testClientInterface);

      const lowLevelProfile = buildLowLevelClientProfile(testClient);
      const receivedResult = assembleClientProfile(lowLevelProfile, testClient.contactMethods);

      expect(receivedResult).toEqual(expect.objectContaining(testClientInterface));
      expect(testClientInterface).toEqual(expect.objectContaining(receivedResult));
    });
  });

  describe('`fetchClientInterface`', () => {
    const testClientInterface = testDataGenerator.getClientInterface({ fieldsToGenerate: ['all'] });
    const testClient = new Client(testClientInterface);
    const lowLevelInterface = buildLowLevelClient(testClient);
    
    // @ts-ignore
    queryByPK.mockReturnValue([...lowLevelInterface.contactMethods, lowLevelInterface.profile]);
    
    it('invokes the queryByPK function with the uuid and table name', async () => {
      expect(queryByPK).not.toHaveBeenCalled();     
      fetchClientInterface(testClient.uuid);
      expect(queryByPK).toHaveBeenCalledTimes(1);      
      
      expect(queryByPK).toHaveBeenCalledWith(ddb.clientsTable, testClient.uuid);
    });

    it('returns a highlevel client interface', async () => {
      const fetchedClientInterface = await fetchClientInterface(testClient.uuid);
      expect(fetchedClientInterface).toEqual(expect.objectContaining(testClientInterface));
      expect(testClientInterface).toEqual(expect.objectContaining(fetchedClientInterface));
    });
  });

  describe('`assembleLowLevelClient`', () => {
    it('creates a LowLevelClient object from an array of dynamoDB items', () => {
      const testClient = testDataGenerator.getClient({ fieldsToGenerate: ['all'] });
      const testLowLevelClient = buildLowLevelClient(testClient);
      const resultLowLevelClient = assembleLowLevelClientInterface([testLowLevelClient.profile, ...testLowLevelClient.contactMethods]);
      expect(resultLowLevelClient).toEqual(expect.objectContaining(testLowLevelClient));
      expect(testLowLevelClient).toEqual(expect.objectContaining(resultLowLevelClient));
    });
  });
});