import Project from '@svc/lib/models/Project';
import {
  assembleLowLevelProjectInterface,
  assembleProjectDetails,
  buildLowLevelProjectDetails,
  buildProjectPutTransaction,
} from '@svc/lib/ddbRepos/projects-repo';
import { LowLevelProjectDetails, ProjectInterface } from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/config', () => {
  const original = jest.requireActual('@svc/config');

  return {
    __esModule: true,
    ddb: {
      ...original.ddb,
      documentClient: {
        send: jest.fn(),
      },
    },
  };
});

jest.mock('@svc/lib/ddbRepos/repo-utils', () => {
  return {
    __esModule: true,
    queryByPK: jest.fn().mockImplementation(() => ([])),
  };
});

describe('`Project` repo', () => {
  beforeEach(() => jest.clearAllMocks());

  
  describe('`buildLowLevelProjectDetails`', () => {
    it('has 10 keys when supplied with no optional fields', () => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['pdid', 'isDeleted', 'clientCost', 'clientUuid']});
      const lowLevel = buildLowLevelProjectDetails(testProject);
      expect(Object.keys(lowLevel)).toHaveLength(10)
    });
    
    it('has 20 keys when supplied with all optional fields', () => {
      const testProject = testDataGenerator.getProject({ fieldsToGenerate: ['all'] });
      const lowLevel = buildLowLevelProjectDetails(testProject);
      expect(Object.keys(lowLevel)).toHaveLength(20);
    });
    
    it('correctly inherits the properties from the high-level interface', () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = testDataGenerator.getProject({ overrides: testProjectInterface });
      const details = buildLowLevelProjectDetails(testProject);

      for (let key in testProjectInterface) {
        expect(details[key as keyof LowLevelProjectDetails]).toBe(testProjectInterface[key as keyof ProjectInterface]);
      }
    });

    it('does not create a gsi1 keys if there is no PDID', () => {
      const testProject = testDataGenerator.getProject({fieldsToExclude: ['pdid']});
      const details = buildLowLevelProjectDetails(testProject);

      expect(details.gsi1_pk).toBeUndefined();
      expect(details.gsi1_sk).toBeUndefined();
    });

    it('correctly assigns the new properties specific to the low-level interface', () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = testDataGenerator.getProject({ overrides: testProjectInterface });
      const details = buildLowLevelProjectDetails(testProject);

      expect(details.pk).toBe(testProject.uuid);
      expect(details.sk).toBe('Details');
      expect(details.gsi1_pk).toBe(`PDID#${testProject.pdid}`);
      expect(details.gsi1_sk).toBe(`PDID#${testProject.pdid}`);
      expect(details.gsi2_pk).toBe(`Project#clientUuid#${testProject.clientUuid}`);
      expect(details.gsi2_sk).toBe(testProject.uuid);
      expect(details.type).toBe('ProjectDetails');
    });
  });

  describe('`buildProjectPutTransaction`', () => {
    const testProject = testDataGenerator.getProject();
    const lowLevelProject = buildLowLevelProjectDetails(testProject);
    const transaction = buildProjectPutTransaction(lowLevelProject);
    const TransactItems = transaction.TransactItems!;

    it('contains only put items', () => {
      TransactItems.every(item => expect(item).toHaveProperty('Put'));
    });
    describe('The Project Details put item', () => {
      it('is the first item', () => {
        expect(TransactItems[0].Put?.Item?.type).toBe('ProjectDetails');
      });

      it('refers to the projects table', () => {
        expect(TransactItems[0].Put?.TableName).toContain('projects');
      });

      it('contains the project profile as the item', () => {
        expect(TransactItems[0].Put?.Item).toBe(lowLevelProject);
      });
    });
  });

  describe('`assembleProjectDetails`', () => {
    it('creates a high level project profile from a low level one', () => {
      const testProjectInterface = testDataGenerator.getProjectInterface({ fieldsToGenerate: ['all'] });
      const testProject = new Project(testProjectInterface);

      const lowLevelProject = buildLowLevelProjectDetails(testProject);
      const receivedResult = assembleProjectDetails(lowLevelProject);

      expect(receivedResult).toEqual(expect.objectContaining(testProjectInterface));
      expect(testProjectInterface).toEqual(expect.objectContaining(receivedResult));
    });
  });

  describe('`assembleLowLevelProjectInterface`', () => {
    it('creates a LowLevelProject object from an array of dynamoDB items', () => {
      // this test is currently basically worthless but will be helpful when other items are
      // added to projects
      const testProject = testDataGenerator.getProject({ fieldsToGenerate: ['all'] });
      const testLowLevelProject = buildLowLevelProjectDetails(testProject);

      const resultLowLevelProject = assembleLowLevelProjectInterface([testLowLevelProject]);

      expect(resultLowLevelProject).toEqual(expect.objectContaining(testLowLevelProject));
      expect(testLowLevelProject).toEqual(expect.objectContaining(resultLowLevelProject));
    });
  });
});