import { Context, S3Event } from 'aws-lambda';
import log from '@dazn/lambda-powertools-logger';
import { addSeenS3Record, wasS3RecordSeen } from '@svc/lib/ddbRepos/idempotency-repo';
import { s3Config } from '@svc/config';
import Activity from '@svc/lib/models/Activity';
import { GetObjectCommand, GetObjectCommandInput } from '@aws-sdk/client-s3';
import { Readable } from 'stream';
import parseCallTranscript from '@svc/lib/parseCallTranscript';
import CallTranscript from '@svc/lib/types/CallTranscript'
import { ChangeSource } from '@improovy/types';

export const handler = async (event: S3Event, ctx: Context) => {
  log.debug('Received event', { event });
  
  for (let record of event.Records) {
    const recordWasSeen = await wasS3RecordSeen(record, ctx.functionName);
    if (recordWasSeen) continue;

    const s3Key = record.s3.object.key
    const callSid = s3Key.split('/')[1].replace('.json', '');

    if (callSid.slice(callSid.length - 13) === '-testing.json') continue;

    const activity = await Activity.fetchByCallSid(callSid);

    if (!activity) throw new Error("activity not found");

    const getObject: GetObjectCommandInput = {
      Bucket: s3Config.callTranscriptsBucket,
      Key: s3Key
    }

    const s3Transcript = await new Promise(async (resolve, reject) => {
      try {
        const resp = await s3Config.s3Client.send(new GetObjectCommand(getObject))

        const data: string[] = [];

        (resp.Body! as Readable).on('data', chunk => data.push(chunk));

        (resp.Body! as Readable).once("end", () => {
          const transcriptObj = JSON.parse(data.join(''));
          resolve(transcriptObj)
        });
      } catch(e) {
        return reject(e)
      }
    }) as CallTranscript;
    
    const parsedTranscript = parseCallTranscript(s3Transcript)

    activity.changeSource = ChangeSource.aws
    await activity.updateBody(activity.body + '<br>' + parsedTranscript)

    await addSeenS3Record(record, ctx.functionName);
  }
  return;
};