import { DeleteObjectCommand } from '@aws-sdk/client-s3';
import { Upload } from '@aws-sdk/lib-storage';
import { AWS_REGION, lambdaConfig, s3Config } from '@svc/config';
import Activity from '@svc/lib/models/Activity';
import parseCallTranscript from '@svc/lib/parseCallTranscript';
import TestDataGenerator, { generateCallSid } from '@tests/utils/testDataGenerator';
import { waitForGSIUpdates } from '@tests/utils/wait';
import * as sampleTranscript from '../../sampleData/sampleTranscript.json'

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}S3AddTranscriptToActivity`;

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`addTranscriptToActivity` Lambda function', () => {
  const testCallSid = generateCallSid();
  let testActivity;

  beforeAll(async () => {
    testActivity = testDataGenerator.getActivity({overrides: {callSid: testCallSid}, fieldsToGenerate: ['callSid']})
    await testActivity.put();
  })

  it('is triggered when a new transcript is added to the call transcripts bucket', async () => {
    const uploadParams = {
      Bucket: s3Config.callTranscriptsBucket,
      Key: `analytics/${testCallSid}.json`,
      Body: JSON.stringify(sampleTranscript),
      ContentType: 'application/json; charset=utf-8'
    }

    const upload = new Upload({
      client: s3Config.s3Client,
      params: uploadParams
    })

    await upload.done();
    
    await expect({...awsTestOptions, timeout: 50000}).toHaveLog(testCallSid)
  });

  it('updates the activity body with the parsed call transcript and sets the changeSource to aws', async () => {
    await waitForGSIUpdates();
    const testActivity = await Activity.fetchByCallSid(testCallSid);
    const testBody = parseCallTranscript(sampleTranscript);

    if (testActivity === null) {
      expect(testActivity).not.toBeNull();
      return
    }

    expect(testActivity.body).toContain(testBody)
  });
  
  afterAll(async () => {
    await testDataGenerator.deleteAll()
    const bucketParams = { 
      Bucket: s3Config.callTranscriptsBucket,
      Key: `analytics/${testCallSid}.json`
    }

    await s3Config.s3Client.send(new DeleteObjectCommand(bucketParams))
  });
});
