import { AWS_REGION, lambdaConfig, s3Config } from '@svc/config';
import Activity from '@svc/lib/models/Activity';
import uuid from '@svc/lib/uuid';
import TestDataGenerator, { generateCallSid, generateCallSidBody } from '@tests/utils/testDataGenerator';
import { lambdaLogsWarmWait, waitForGSIUpdates, waitForLambdaLogs } from '@tests/utils/wait';
import { DeleteObjectCommand } from '@aws-sdk/client-s3'

const testDataGenerator = new TestDataGenerator();

const lambdaFunctionName = `${lambdaConfig.functionNamePrefix}DDBDownloadCallRecording`;

const awsTestOptions = {
  region: AWS_REGION,
  function: lambdaFunctionName,
};

describe('`downloadCallRecording` Lambda function', () => {
  it('is triggered when a new activity with a callSid attribute is inserted into the Activities DDB table', async () => {
    const testActivity = testDataGenerator.getActivity({fieldsToGenerate: ['callSid']});
    await testActivity.put();
    const expectedLog = testActivity.pdid;
    return expect({
      ...awsTestOptions,
      timeout: 50000,
    }).toHaveLog(expectedLog);

  });

  it.concurrent('is not triggered when a new activity without a callSid attribute is inserted into the Activities DDB table', async () => {
    const testActivity = testDataGenerator.getActivity();
    await testActivity.put();
    const unexpectedLog = testActivity.pdid;

    await waitForLambdaLogs();

    return expect({
      ...awsTestOptions,
    }).not.toHaveLog(unexpectedLog);

  });

  it.concurrent('is triggered when an activity is updated with a new callSid attribute', async () => {
    const testActivityInterface = testDataGenerator.getActivityInterface();
    const testActivity = new Activity(testActivityInterface);

    await testActivity.put();

    await waitForGSIUpdates();

    const newTestActivityInterface = {...testActivityInterface, body: generateCallSidBody()};

    const newTestActivity = new Activity(newTestActivityInterface);

    const diffs = Activity.differingFields(newTestActivity, testActivity);

    await newTestActivity.updateFields(diffs);

    const expectedLog = testActivity.pdid;

    return expect({
      ...awsTestOptions,
      timeout: lambdaLogsWarmWait,
    }).toHaveLog(expectedLog);

  });

  it.concurrent('is not triggered when an activity is updated without a new callSid attribute or when a callSid attribute is removed', async () => {
    const unexpectedLog = uuid()

    const testActivityInterface = testDataGenerator.getActivityInterface({overrides: {body: unexpectedLog}});
    const testActivity = new Activity(testActivityInterface);

    const oldTestActivityInterface = {...testActivityInterface, body: generateCallSidBody()};

    const oldTestActivity = new Activity(oldTestActivityInterface);
    
    await oldTestActivity.put();

    const diffs = Activity.differingFields(oldTestActivity, testActivity);

    await testActivity.updateFields(diffs);

    await waitForLambdaLogs();

    return expect({
      ...awsTestOptions,
      timeout: lambdaLogsWarmWait,
    }).not.toHaveLog(unexpectedLog);
  });

  it.concurrent('uploads the referenced call to s3', async () => {
    const testRecordingKey = generateCallSid();

    const bucketParams = { 
      Bucket: s3Config.callRecordingsBucket,
      Key: testRecordingKey
    }

    await s3Config.s3Client.send(new DeleteObjectCommand(bucketParams))

    const testActivity = testDataGenerator.getActivity({ overrides: { body: generateCallSidBody(testRecordingKey), isTest: false} });
    
    await testActivity.put();

    await waitForLambdaLogs();
    
    return await expect({region: AWS_REGION, bucket: s3Config.callRecordingsBucket, timeout: 60000}).toHaveObject(testRecordingKey)
  });

  afterAll(async () => testDataGenerator.deleteAll());
});