import { AWS_REGION, ddb } from '@svc/config';
import TestDataGenerator, { generateCallSid } from '@tests/utils/testDataGenerator';
import _ from 'lodash';
import {
  buildLowLevelActivity,
  putActivity,
  hardDeleteActivity,
  fetchActivityInterface,
  updateActivity,
  fetchActivityInterfaceByCallSid,
} from '@svc/lib/ddbRepos/activities-repo';
import Activity from '@svc/lib/models/Activity';
import { waitForGSIUpdates } from '@tests/utils/wait';
import { MutableActivityProperties } from '@improovy/types';

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.activitiesTable,
};

const testDataGenerator = new TestDataGenerator();

describe('`Activity` repo', () => {
  describe('`putActivity`', () => {
    it('adds the activity instance to the activity DynamoDB table', async () => {
      const testActivity = testDataGenerator.getActivity();
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await putActivity(testActivity);

      await expect(awsTestOptions).toHaveItem(_.pick(lowLevelActivity, ['pk', 'sk']), {
        ...lowLevelActivity,
      });
    });
  });

  describe('`fetchActivityInterface`', () => {
    it('returns the high level interface of the activity matching the activity pdid', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all']});
      const testActivity = new Activity(testActivityInterface);
  
      await testActivity.put();
  
      await waitForGSIUpdates();
  
      const receivedInterface = await fetchActivityInterface(testActivity.pdid);
  
      expect(receivedInterface).toStrictEqual(testActivityInterface);
    });
    
    it('returns null if no results are found', async () => {
      const receivedInterface = await fetchActivityInterface('nonexistent pdid');
  
      expect(receivedInterface).toBeNull();
    });
  });

  describe('`fetchActivityInterfaceByCallSid`', () => {
    it('returns the high level interface of the activity matching the activity callSid', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all']});
      const testActivity = new Activity(testActivityInterface);
  
      await testActivity.put();
  
      await waitForGSIUpdates();
  
      const receivedInterface = await fetchActivityInterfaceByCallSid(testActivity.callSid!);
  
      expect(receivedInterface).toStrictEqual(testActivityInterface);
    });
    
    it('returns null if no results are found', async () => {
      const receivedInterface = await fetchActivityInterfaceByCallSid('nonexistent callSid');
  
      expect(receivedInterface).toBeNull();
    });
  });

  describe('`hardDeleteActivity`', () => {
    it('removes all items associated with the activity', async () => {
      const testActivity = testDataGenerator.getActivity();
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await putActivity(testActivity);

      await waitForGSIUpdates();

      await hardDeleteActivity(testActivity);

      await waitForGSIUpdates();

      await expect(awsTestOptions).not.toHaveItem(_.pick(lowLevelActivity, ['pk', 'sk']));
    });
  });

  describe('`updateActivity`', () => {
    it('updates the activity lastUpdatedField, field referenced by lastUpdatedField, and changeSource', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivity = new Activity(testActivityInterface);
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await testActivity.put();
      
      testActivityInterface.lastUpdatedField = MutableActivityProperties.body;
      testActivityInterface.body = 'test body';

      // @ts-expect-error
      await updateActivity(testActivityInterface);

      const keys = _.pick(lowLevelActivity, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelActivity, lastUpdatedField: MutableActivityProperties.body, body: 'test body' });
    });

    it('updates the gsi1_pk and gsi1_sk fields if the callSid is updated', async () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivity = new Activity(testActivityInterface);
      const lowLevelActivity = buildLowLevelActivity(testActivity);

      await testActivity.put();
      
      testActivityInterface.lastUpdatedField = MutableActivityProperties.callSid;
      const newCallSid = generateCallSid();
      testActivityInterface.callSid = newCallSid;

      const secondaryKey = `Sid#${newCallSid}`;

      await updateActivity(testActivityInterface as Activity);

      const keys = _.pick(lowLevelActivity, ['pk', 'sk']);

      await expect(awsTestOptions).toHaveItem(keys, { ...lowLevelActivity, lastUpdatedField: MutableActivityProperties.callSid, callSid: newCallSid, gsi1_pk: secondaryKey, gsi1_sk: secondaryKey });
    });
  });

  afterAll(async () => {
    return testDataGenerator.deleteAll();
  });
});