import { DeleteCommand, DeleteCommandInput } from '@aws-sdk/lib-dynamodb';
import { AWS_REGION, ddb } from '@svc/config';
import { addSeenCallSid, addSeenToken, wasCallSidSeen, wasTokenSeen } from '@svc/lib/ddbRepos/idempotency-repo';
import uuid from '@svc/lib/uuid';

const awsTestOptions = {
  region: AWS_REGION,
  table: ddb.idempotencyTable,
};

const testFunc = 'testFunc';

describe('`Idempotency` repo', () => {
  describe('addSeenToken', () => {
    it('takes in the full pk adds it to the database and adds a default TTL of 3 days from now, rounded up to the nearest 1000 seconds', async () => {
      // note, AWS may take up to 48 hours after TTL to delete the item
      const testToken = uuid();
      await addSeenToken(testToken);
      const savedToken = `${testToken}#`

      const TTL = Math.ceil(Date.now() / 1000000) * 1000 + 3 * 24 * 60 * 60

      await expect(awsTestOptions).toHaveItem({ pk: savedToken },{pk: savedToken, TTL });
    });

    it('does not set the TTL if the daysToExpiry is null', async () => {
      const testToken = uuid();
      await addSeenToken(testToken, '', null);
      const savedToken = `${testToken}#`

      await expect(awsTestOptions).toHaveItem({ pk: savedToken }, {pk: savedToken, TTL: undefined});
      
      const deleteCommand: DeleteCommandInput = {
        TableName: ddb.idempotencyTable,
        Key: {
          pk: savedToken
        }
      };

      await ddb.documentClient.send(new DeleteCommand(deleteCommand))
    });

    it('takes in an optional seenBy string and appends it to the end of the token', async () => {
      const testToken = uuid();
      await addSeenToken(testToken, testFunc);

      await expect(awsTestOptions).toHaveItem({ pk: `${testToken}#${testFunc}` });
    });
  });

  describe('wasTokenSeen', () => {
    it('takes in the full pk and returns true if the token exists in the database', async () => {
      const testToken = uuid();
      await addSeenToken(testToken, testFunc);
      
      const wasSeen = await wasTokenSeen(testToken, testFunc);

      expect(wasSeen).toBe(true);
    });

    it('returns false if the token does not exist in the database', async () => {
      const testToken = uuid();
      
      const wasSeen = await wasTokenSeen(testToken);

      expect(wasSeen).toBe(false);
    });
  });

  // describe('addSeenPipedriveRequest', () => {
  //   it('takes in activity request and adds its unique identifier to the database', async () => {
  //     const ts = uuid();
  //     const entityId = uuid();
  //     const eventAction = PipedriveEventActions.updated;
  //     const eventEntity = PipedriveEventEntities.activity;
  //     const request = {
  //       'meta': {
  //         'id': entityId,
  //         'timestamp_micro': ts,
  //         'action': eventAction,
  //         'object': eventEntity,
  //       },
  //     };

  //     await addSeenPipedriveRequest(request);

  //     await expect(awsTestOptions).toHaveItem({ pk: `${eventEntity}#${entityId}#${eventAction}#${ts}#` });
  //   });

  //   it('takes in an optional seenBy string and appends it to the end of the token', async () => {
  //     const ts = uuid();
  //     const entityId = uuid();
  //     const eventAction = PipedriveEventActions.updated;
  //     const eventEntity = PipedriveEventEntities.activity;
  //     const request = {
  //       'meta': {
  //         'id': entityId,
  //         'timestamp_micro': ts,
  //         'action': eventAction,
  //         'object': eventEntity,
  //       },
  //     };

  //     await addSeenPipedriveRequest(request, testFunc);

  //     await expect(awsTestOptions).toHaveItem({ pk: `${eventEntity}#${entityId}#${eventAction}#${ts}#${testFunc}` });
  //   });
  // });

  // describe('wasPipedriveRequestSeen', () => {
  //   it('takes in activity request and returns true if it was already seen', async () => {
  //     const ts = uuid();
  //     const eventId = uuid();
  //     const eventAction = PipedriveEventActions.updated;
  //     const eventEntity = PipedriveEventEntities.activity;
  //     const request = {
  //       'meta': {
  //         'id': eventId,
  //         'timestamp_micro': ts,
  //         'action': eventAction,
  //         'object': eventEntity,
  //       },
  //     };

  //     await addSeenPipedriveRequest(request, testFunc);

  //     const wasSeen = await wasPipedriveRequestSeen(request, testFunc);

  //     expect(wasSeen).toBe(true);
  //   });

  //   it('returns false if the activity request was not already seen', async () => {
  //     const ts = uuid();
  //     const eventId = uuid();
  //     const eventAction = PipedriveEventActions.updated;
  //     const eventEntity = PipedriveEventEntities.activity;
  //     const request = {
  //       'meta': {
  //         'id': eventId,
  //         'timestamp_micro': ts,
  //         'action': eventAction,
  //         'object': eventEntity,
  //       },
  //     };

  //     const wasSeen = await wasPipedriveRequestSeen(request);

  //     expect(wasSeen).toBe(false);
  //   });
  // });

  describe('addSeenCallSid', () => {
    it('takes in a call SID and adds its unique identifier to the database', async () => {
      const callSid = uuid();

      await addSeenCallSid(callSid);

      await expect(awsTestOptions).toHaveItem({ pk: `callSid#${callSid}#` });
    });

    it('takes in an optional seenBy argument and appends it to the end of the unique token', async () => {
      const callSid = uuid();

      await addSeenCallSid(callSid, testFunc);

      await expect(awsTestOptions).toHaveItem({ pk: `callSid#${callSid}#${testFunc}` });
    });
  });

  describe('wasCallSidSeen', () => {
    it('takes in a call SID and returns true if it was already seen', async () => {
      const callSid = uuid();

      await addSeenCallSid(callSid, testFunc);

      const wasSeen = await wasCallSidSeen(callSid, testFunc);

      expect(wasSeen).toBe(true);
    });

    it('returns false if the call SID was not already seen', async () => {
      const callSid = uuid();

      const wasSeen = await wasCallSidSeen(callSid);

      await expect(wasSeen).toBe(false);
    });

    describe('the optional seenBy argument', () => {
      it('returns true if the full token exists already in the database', async () => {
        const callSid = uuid();
  
        await addSeenCallSid(callSid, testFunc);
  
        const wasSeen = await wasCallSidSeen(callSid, testFunc);
  
        expect(wasSeen).toBe(true);
      });
  
      it('returns false if the combination was not already seen', async () => {
        const callSid = uuid();

        await addSeenCallSid(callSid);
  
        const wasSeen = await wasCallSidSeen(callSid, testFunc);
  
        await expect(wasSeen).toBe(false);
      });

    })
  });

  // describe('addSeenCallRailSignature', () => {
  //   it('takes in a CallRail signature and adds its unique identifier to the database', async () => {
  //     const signature = uuid();

  //     await addSeenCallRailSignature(signature);

  //     await expect(awsTestOptions).toHaveItem({ pk: `callRail#reqSignature#${signature}#` });
  //   });

  //   it('takes in an optional seenBy argument and appends it to the end of the unique token', async () => {
  //     const signature = uuid();

  //     await addSeenCallRailSignature(signature, testFunc);

  //     await expect(awsTestOptions).toHaveItem({ pk: `callRail#reqSignature#${signature}#${testFunc}` });
  //   });
  // });

  // describe('wasCallRailSignatureSeen', () => {
  //   it('takes in a CallRail signature and returns true if it was already seen', async () => {
  //     const signature = uuid();
  //     await addSeenCallRailSignature(signature, testFunc);

  //     const wasSeen = await wasCallRailSignatureSeen(signature, testFunc);

  //     expect(wasSeen).toBe(true);
  //   });

  //   it('returns false if the signature was not already seen', async () => {
  //     const wasSeen = await wasCallRailSignatureSeen('non-existent signature');

  //     expect(wasSeen).toBe(false);
  //   });
  // });

  // describe('addSeenGclid', () => {
  //   it('takes in a gclid and adds its unique identifier to the database', async () => {
  //     const gclid = uuid();

  //     await addSeenGclid(gclid);

  //     await expect(awsTestOptions).toHaveItem({ pk: `gclid#${gclid}#` });
  //   });

  //   it('takes in an optional seenBy argument and appends it to the end of the unique token', async () => {
  //     const gclid = uuid();

  //     await addSeenGclid(gclid, testFunc);

  //     await expect(awsTestOptions).toHaveItem({ pk: `gclid#${gclid}#${testFunc}` });
  //   });
  // });

  // describe('wasGclidSeen', () => {
  //   it('takes in a gclid and returns true if it was already seen', async () => {
  //     const gclid = uuid();

  //     await addSeenGclid(gclid);

  //     const wasSeen = await wasGclidSeen(gclid);

  //     expect(wasSeen).toBe(true);
  //   });

  //   it('returns false if the gclid was not already seen', async () => {
  //     const gclid = uuid();

  //     const wasSeen = await wasGclidSeen(gclid);

  //     await expect(wasSeen).toBe(false);
  //   });

  //   describe('the optional seenBy argument', () => {
  //     it('returns true if the full token exists already in the database', async () => {
  //       const gclid = uuid();
  
  //       await addSeenGclid(gclid, 'testFunc');
  
  //       const wasSeen = await wasGclidSeen(gclid, 'testFunc');
  
  //       expect(wasSeen).toBe(true);
  //     });
  
  //     it('returns false if the combination was not already seen', async () => {
  //       const gclid = uuid();

  //       await addSeenGclid(gclid);
  
  //       const wasSeen = await wasGclidSeen(gclid, 'testFunc');
  
  //       await expect(wasSeen).toBe(false);
  //     });

  //   })
  // });
});