import Activity from '@svc/lib/models/Activity';
import {
  assembleLowLevelActivityInterface,
  assembleActivity,
  buildLowLevelActivity,
  buildActivityPutTransaction,
  assembleActivityInterfaceFromDDBItems,
} from '@svc/lib/ddbRepos/activities-repo';
import { LowLevelActivity, ActivityInterface } from '@improovy/types';
import TestDataGenerator from '@tests/utils/testDataGenerator';

const testDataGenerator = new TestDataGenerator(false);

jest.mock('@svc/config', () => {
  const original = jest.requireActual('@svc/config');

  return {
    __esModule: true,
    ddb: {
      ...original.ddb,
      documentClient: {
        send: jest.fn(),
      },
    },
  };
});

jest.mock('@svc/lib/ddbRepos/repo-utils', () => {
  return {
    __esModule: true,
    queryByPK: jest.fn().mockImplementation(() => ([])),
  };
});

describe('`Activity` repo', () => {
  beforeEach(() => jest.clearAllMocks());

  
  describe('`buildLowLevelActivity`', () => {
    it('has 8 keys when supplied with no optional fields', () => {
      const testActivity = testDataGenerator.getActivity({fieldsToExclude: ['isTest', 'body', 'lastUpdatedField', 'isDeleted']});
      const lowLevel = buildLowLevelActivity(testActivity);
      expect(Object.keys(lowLevel)).toHaveLength(8);
    });
    
    it('has 14 keys when supplied with all optional fields', () => {
      const testActivity = testDataGenerator.getActivity({ fieldsToGenerate: ['all'] });
      const lowLevel = buildLowLevelActivity(testActivity);
      expect(Object.keys(lowLevel)).toHaveLength(14);
    });
    
    it('correctly inherits the properties from the high-level interface', () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivity = testDataGenerator.getActivity({ overrides: testActivityInterface });
      const details = buildLowLevelActivity(testActivity);

      for (let key in testActivityInterface) {
        expect(details[key as keyof LowLevelActivity]).toBe(testActivityInterface[key as keyof ActivityInterface]);
      }
    });

    it('correctly assigns the new properties specific to the low-level interface', () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivity = testDataGenerator.getActivity({ overrides: testActivityInterface });
      const lowLevel = buildLowLevelActivity(testActivity);

      expect(lowLevel.pk).toBe(testActivity.pdid);
      expect(lowLevel.sk).toBe(`${testActivity.pdid}`);
      expect(lowLevel.gsi1_pk).toBe(`Sid#${testActivity.callSid}`);
      expect(lowLevel.gsi1_sk).toBe(`Sid#${testActivity.callSid}`);
    });
  });

  describe('`buildActivityPutTransaction`', () => {
    const testActivity = testDataGenerator.getActivity();
    const lowLevelActivity = buildLowLevelActivity(testActivity);
    const transaction = buildActivityPutTransaction(lowLevelActivity);
    const TransactItems = transaction.TransactItems!;

    it('contains only put items', () => {
      TransactItems.every(item => expect(item).toHaveProperty('Put'));
    });

    it('contains only one item', () => {
      expect(TransactItems).toHaveLength(1);
    });

    it('refers to the activity table', () => {
      expect(TransactItems[0].Put?.TableName).toContain('activities');
    });

    it('contains the activity gclid as the item', () => {
      expect(TransactItems[0].Put?.Item).toBe(lowLevelActivity);
    });
  });

  describe('`assembleActivity`', () => {
    it('creates a high level activity profile from a low level one', () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });
      const testActivity = new Activity(testActivityInterface);

      const lowLevelActivity = buildLowLevelActivity(testActivity);
      const receivedResult = assembleActivity(lowLevelActivity);

      expect(receivedResult).toEqual(expect.objectContaining(testActivityInterface));
      expect(testActivityInterface).toEqual(expect.objectContaining(receivedResult));
    });
  });

  describe('`assembleLowLevelActivityInterface`', () => {
    it('creates a LowLevelActivity object from its respective DDB item', () => {
      const testActivity = testDataGenerator.getActivity({ fieldsToGenerate: ['all'] });
      const testLowLevelActivity = buildLowLevelActivity(testActivity);

      const resultLowLevelActivity = assembleLowLevelActivityInterface(testLowLevelActivity);

      expect(resultLowLevelActivity).toEqual(expect.objectContaining(testLowLevelActivity));
      expect(testLowLevelActivity).toEqual(expect.objectContaining(resultLowLevelActivity));
    });
  });

  describe('`assembleActivityInterfaceFromDDBItems`', () => {
    it('creates a LowLevelActivity object from its respective DDB item', () => {
      const testActivityInterface = testDataGenerator.getActivityInterface({ fieldsToGenerate: ['all'] });

      const testActivity = new Activity(testActivityInterface)

      const testLowLevelActivity = buildLowLevelActivity(testActivity);

      const resultActivity = assembleActivityInterfaceFromDDBItems([testLowLevelActivity as { [key: string] : any }]);

      expect(resultActivity).toEqual(expect.objectContaining(testActivityInterface));
      expect(testActivityInterface).toEqual(expect.objectContaining(resultActivity));
    });
  });
});